/* This program sorts the elements of bibliographic entries, placing them    */
/* into lexicographic order.  For each entry, the related elements are       */
/* temporarily arranged into their proper order through the use of a linked  */
/* list, and then written to an outfile.  For the sake of clarity it should  */
/* be noted that, for reasons not worth going into here, the evolution of    */
/* this program included setting up the linked list with a dummy node at the */
/* front, and traversing it through the use of a counter.  It functions      */
/* properly in its current form, and time did not allow for revising it      */
/* prior to the initial release.  However, it would be more straightforward  */
/* to remove the dummy node and the counter, and traverse the linked list by */
/* looping until a null pointer is found.  If this code needs to be revised  */
/* in the future, it may be worthwhile to incorporate these suggested        */
/* changes.                                                                  */


#include <errno.h>
#include <stdio.h>

#define TRUE 1
#define FALSE 0
#define EXCEPTION -1
#define BUFFLENGTH 100    /* length of buffer */

/*GLOBAL VARIABLES */
FILE *fid_in;  /* input file */
FILE *fid_out; /* output file */

struct tag{            /* for storing info on various tag types & their text */
  char *begin_element; /* element begin tag */
  char *end_element;   /* element end tag */
  int count;   /* count of chars in text (for reallocating memory if needed) */
  char *textptr;       /* text field between begin and end tags */
  struct tag *nexttag; /* points to next tag in list */
};

usage()
/* print usage message to screen */
{
  printf("usage: bibsort unique_infilename unique_outfilename \n");
  exit(1);
}  /* end usage */


openfile(infile, outfile)
/* open the specified bib infile and outfile */
     char *infile, *outfile;    /* names of input and output files */
{
  if ((fid_in = fopen (infile, "r")) == 0)
    {
      printf ("Bibsort: unable to open infile %s for reading\n", infile);
      exit(1);
    }
  if ((fid_out = fopen (outfile, "w")) == 0)
    {
      printf ("Bibsort: unable to open outfile %s for writing\n", outfile);
      exit(1);
    }
}  /* end openfile */


closefile(infile, outfile)
/* close the bib infile and outfile */
     char *infile, *outfile;    /* names of input and output files */
{
  if (fclose(fid_in) == EXCEPTION)
    {
      printf ("Bibsort: unable to close infile %s\n", infile);
      return FALSE;
    }
  if (fclose(fid_out) == EXCEPTION)
    {
      printf ("Bibsort: unable to close outfile %s\n", outfile);
      return FALSE;
    }
}  /* end closefile */


gettag(tagname, outfile)
/* get the next tag in the bib file */
  char tagname[],
       *outfile;        /* name of outfile    */
{
  char ch; /* next char in infile */
  int i;   /* counter */


  while ((ch = ((char) (getc(fid_in)))) != ('<'))
    /* write character to outfile and loop (takes care of things like */
    /* carriage returns which may appear before a tag)                */
    if (fprintf (fid_out, "%c", ch) == EOF)  
      {
	printf("Bibsort: unable to write to outfile %s\n", outfile);
	exit(1);
      }

  ungetc(ch, fid_in);  /* put the '<' back */


  i = 0;

  while (((ch = ((char) (getc(fid_in)))) != ('>')) && (i < (BUFFLENGTH - 3)))
/* BUFFLENGTH - 3 leaves room for the ">", and the "/" in the coming end tag */
    {
      tagname[i] = ch;
      i++;
    }
  tagname[i] = ch;
  i++;
  tagname[i] = 0;    /* add a NULL */
  ch = ((char) (getc(fid_in)));  /* check for end of file */
  ungetc(ch, fid_in);
}  /* end gettag */


get_non_pi_tag(non_pi_tagname, outfile)
/* Call gettag, and check if it returns a processing instruction tag.        */
/* If so, write the processing instruction tag to the outfile and loop,      */
/* continuing to call gettag until a non-processing instruction tag is found */

     char *non_pi_tagname, /* pointer to the tag */
          *outfile;        /* name of outfile    */

{
  gettag(non_pi_tagname, outfile);
  while (non_pi_tagname[1] == '?')
    {
      /* write processing instruction tag to outfile and loop */
      if (fprintf (fid_out, "%s", non_pi_tagname) == EOF)  
	{
	  printf("Bibsort: unable to write to outfile %s\n", outfile);
	  exit(1);
	}
      gettag(non_pi_tagname);
    }
}  /* end get_non_pi_tag */


endtag(begin_tag, end_tag)
/* produce an end tag for the corresponding begin tag, for comparison */
  char *begin_tag, *end_tag;
{

  *end_tag = *begin_tag;
  begin_tag++;
  end_tag++;
  *end_tag = '/';
  end_tag++;
  for(; *begin_tag != 0; begin_tag++)
    {
      *end_tag = *begin_tag;
      end_tag++;
    }
      *end_tag = *begin_tag;  
}  /* end endtag */


sortlist(tagcntptr, temp, front, tagptr)
/* place an element into linked list lexicographically */
     int *tagcntptr;   /* points to count of tags in list */
     struct tag *temp, *front, *tagptr;  /* for manipulating linked list */

{
  int i;    /* counter */
  int looping;  /* flag to end a loop */
  char *readtag,  /* for reading contents of tags between the "<" and       */
                  /*    the ">"  into temptag buffers for strcmp &          */
                  /*    lexicographic insertion of tag into the linked list */
       temptag1[BUFFLENGTH], *ttag1ptr,  /* see comment in *readtag above   */
       temptag2[BUFFLENGTH], *ttag2ptr;  /* see comment in *readtag above   */

  temp = front;  /* initializations */
  i = 0;
  ttag1ptr = temptag1;
  ttag2ptr = temptag2;

/* place internal tag contents of tagptr->begin_element into temptag1 */
  readtag = tagptr->begin_element; 
  readtag++;    /* skip the '<' */
  for(; ((*readtag != ('>')) && (*readtag != 0)); readtag++)
    {
      *ttag1ptr = *readtag;
      ttag1ptr++;
    }
  *ttag1ptr = 0;
  ttag1ptr = &temptag1[0];

  looping = TRUE;

  while(looping)  /* move through list to appropriate lexicographic position */
    {
      if (i < *tagcntptr)
	{
/* place internal tag contents of temp->nexttag->begin_element into temptag2 */
	  readtag = temp->nexttag->begin_element; 
	  readtag++;
	  for(; ((*readtag != ('>')) && (*readtag != 0)); readtag++)
	    {
	      *ttag2ptr = *readtag;
	      ttag2ptr++;
	    }
	  *ttag2ptr = 0;
	  ttag2ptr = &temptag2[0];

	if (strcmp(ttag1ptr, ttag2ptr) >= 0)
	  {
	    temp = temp->nexttag;
	    i++;
	  }

        else
	  looping = FALSE;
      }
      else
	looping = FALSE;
    }
  tagptr->nexttag = temp->nexttag;  /* insert into list */
  temp->nexttag = tagptr;
}  /* end sortlist */


sortfile(infile, outfile)
/* sort the specified bib infile */
     char *infile, *outfile;    /* names of input and output files */
{
  char begin_filetype[BUFFLENGTH], *b_fptr,  /* begin tag for filetype */
       end_filetype[BUFFLENGTH], *e_fptr,    /* end tag for filetype   */
       begin_entry[BUFFLENGTH], *b_eptr,     /* begin tag for entry    */
       end_entry[BUFFLENGTH], *e_eptr,       /* end tag for entry      */
       tempbuf[BUFFLENGTH], *t_ptr,          /* temporary buffer       */
       ch;                                   /* character from infile  */
  int match,       /* flag used to find a match for an element endtag  */
      looping;     /* flag used to end a loop */
  unsigned alloc_size;  /* size allocated for text storage within an element */



  struct tag *front, *temp, *tagptr;  /* for manipulating linked list */
  int tagcnt;  /* count of nodes in linked list */
  int length;   /* length of a string */
  int i;        /* counter */
  int j;   /* counter */


  b_fptr = begin_filetype;    /* initializations */
  e_fptr = end_filetype;
  b_eptr = begin_entry;
  e_eptr = end_entry;
  t_ptr = tempbuf;
  tempbuf[0] = 0;  /* initialize to null for later comparisons */
  
  ch = ((char) (getc(fid_in)));  /* be sure file isn't empty */
  ungetc(ch, fid_in);

  if  (feof(fid_in))
/* print message and exit if infile is empty */
    {
      printf("Bibsort: file %s is empty\n", infile);
      exit(1);
    }

  get_non_pi_tag(b_fptr, outfile);  /* get tag for filetype */
  if (fprintf (fid_out, "%s", b_fptr) == EOF)
    {                         /* write begin_filetype to outfile */
      printf("Bibsort: unable to write to outfile %s\n", outfile);
      exit(1);
    }
  endtag(b_fptr, e_fptr);  /*create corresponding endtag */

  tagptr = ((struct tag *) malloc(sizeof(struct tag))); /*dummy node */

  tagptr->begin_element  = (char *) malloc(1);
  tagptr->begin_element[0] = 0;  /* initialize to null for later comparisons */
  tagptr->nexttag = 0;  /*initialize as a null pointer */

  front = tagptr;  /* initializations */
  temp = tagptr;
  tagcnt = 0;

  alloc_size = BUFFLENGTH;

  while (strcmp(t_ptr, e_fptr))
    {
/* get entry tag */
      get_non_pi_tag(t_ptr, outfile);

      if (strcmp(t_ptr, e_fptr))
	{     /* $$ */
	  strcpy(b_eptr, t_ptr);  /* store current tag as a begin entry tag */
	  endtag(b_eptr, e_eptr); /* store corresponding end entry tag      */
	  
/* continue until a new tag matches end_entry */
	  get_non_pi_tag(t_ptr, outfile);
	  while (strcmp(t_ptr, e_eptr))
	    {
/* set up linked list of tag data */
/* store element begin tag        */
	      tagptr = ((struct tag *) malloc(sizeof(struct tag)));
	      length = strlen(t_ptr);   /* get/store length of element begin tag */
	      tagptr->begin_element  = (char *) malloc(length + 1); /* add the null */
	      strcpy(tagptr->begin_element, t_ptr);

/* make & store a copy of the coming element end tag */
	      tagptr->end_element  = (char *) malloc(length + 2); /* add the '/' and the null */
	      endtag(tagptr->begin_element, tagptr->end_element);

	      ch = ((char) (getc(fid_in)));
                    /* get space for text */
	      tagptr->textptr = (char *) malloc(alloc_size); 
	      tagptr->count = 0;
	      i = 0;

	      match = FALSE;
	      while (!(match))
		{
		  while (ch != ('<')) 
		    /* add to text & get next char */
		    {
		      tagptr->textptr[i] = ch;
		      i++;
		      tagptr->count++;
		      ch = ((char) (getc(fid_in)));
		      if (tagptr->count == alloc_size)
			{       /* reallocate if text exceeds space */
			  alloc_size = (alloc_size * 2);
			  realloc(tagptr->textptr, alloc_size);
			}
		    } /* end while (ch != ('<')) */

		  looping = TRUE;
		  match = TRUE;
		  t_ptr[0] = ch;
		  j = 1;

		  while (looping)
/* check if chars following '<' are the element end tag or just more text */
		    { 
		      if (tagptr->end_element[j] != 0)
			if ((ch = ((char) (getc(fid_in)))) == (tagptr->end_element[j]))
			  {
			    t_ptr[j] = ch;
			    j++;
			  }
			else
			  {
			    match = FALSE;
			    looping = FALSE;
			  }
		      else
			looping = FALSE;
		    } /* end while (looping) */
	       

		  if (!(match))
/* If a match wasn't found for the element end tag, place the chars from the */
/* temporary buffer into the linked list text field and keep searching       */
		    {		      
		      tagptr->count = ((tagptr->count) + j);
		      if (tagptr->count >= alloc_size)
			{           /* reallocate if text exceeds space */
			  alloc_size = (alloc_size * 2);
			  realloc(tagptr->textptr, alloc_size);
			}
		      length = j;
		      for (j = 0; j < length; j++)
			{
			  tagptr->textptr[i] = t_ptr[j];
			  i++;
			}

		    }  /* end if(!match) */

		}  /* end while (!match) */


	      tagptr->textptr[i] = 0;   /* put null at end of text */

	      sortlist(&tagcnt, temp, front, tagptr);
	      tagcnt++;

	      get_non_pi_tag(t_ptr, outfile);

	      ch = ((char) (getc(fid_in)));
	      ungetc(ch, fid_in);
	    }    /* end while (strcmp(t_ptr, e_ptr)) */

/* write begin_entry to outfile */
	  if (fprintf (fid_out, "%s", b_eptr) == EOF)  
	    {
	      printf("Bibsort: unable to write to outfile %s\n", outfile);
	      exit(1);
	    }

/* write entry details to outfile */
	  temp = front->nexttag;
	  for (i = 1; i <= tagcnt; i++)
	    {
	      if (fprintf(fid_out, "%s", temp->begin_element) == EOF)
		{              /* write begin_element tag to outfile */
		  printf("Bibsort: unable to write to outfile %s\n", outfile);
		  exit(1);
		}
	      if (fprintf(fid_out, "%s", temp->textptr) == EOF)
		{                      /* write element text to outfile */
		  printf("Bibsort: unable to write to outfile %s\n", outfile);
		  exit(1);
		}
	      if (fprintf(fid_out, "%s", temp->end_element) == EOF)
		{                      /* write end_element tag to outfile */
		  printf("Bibsort: unable to write to outfile %s\n", outfile);
		  exit(1);
		}

	      if (i < tagcnt)
		temp = temp->nexttag;
	    }

/* write end_entry to outfile */
	  if (fprintf (fid_out, "%s", e_eptr) == EOF)  
	    {
	      printf("Bibsort: unable to write to outfile %s\n", outfile);
	      exit(1);
	    }


/* debugging code to traverse list */
/*  printf("\n Tested linked list: \n\n");  
  temp = front->nexttag;
  for (i = 1; i <= tagcnt; i++)
    {
      printf("\nnode %d: begin_element is %s \n", i, temp->begin_element);
      printf("node %d: end_element is %s \n", i, temp->end_element);
      printf("node %d: count is %d \n", i, temp->count);
      printf("node %d: text is: %s \n\n", i, temp->textptr);
      if (i < tagcnt)
	temp = temp->nexttag;
    } */


/* release memory used for linked list */
	  temp = front->nexttag;
	  for (i = 1; i <= tagcnt; i++)
	    {
	      free(temp->begin_element);
	      free(temp->end_element);
	      free(temp->textptr);
	      if (i < tagcnt)
		{
		  tagptr = temp->nexttag;
		  free(temp);
		  temp = tagptr;
		}
	    }
	  free(temp);


/* reset linked list */
	  tagcnt = 0;
	  tagptr = front;
	  temp = front;


	}   /* end if (strcmp(t_ptr, e_fptr))*/
    }    /* end while (strcmp(t_ptr, e_fptr))*/

/* write end_filetype to outfile */
  if (fprintf (fid_out, "%s", e_fptr) == EOF)  
    {
      printf("Bibsort: unable to write to outfile %s\n", outfile);
      exit(1);
    }

/* write a newline at the end of the outfile */
  if (fprintf (fid_out, "\n") == EOF)  
    {
      printf("Bibsort: unable to write to outfile %s\n", outfile);
      exit(1);
    }
} /* end sortfile */


main (argc, argv)
     int argc;
     char *argv[];

{
  int  i;           /* loopcounter */

  if (argc == 3)    /* get file names from argv */
/* input and output filenames must be unique */
    if (strcmp (argv[1], argv[2]) == 0) 
      usage();
    else
      {
      openfile(argv[1], argv[2]);
      sortfile(argv[1], argv[2]);
      closefile(argv[1], argv[2]);
      printf("Sorting process completed successfully\n");
      }
  else
    usage();
}  /* end main */


