/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: deflist.c,v $
   Purpose:
       To handle the available symbol list.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.9 $
   Date:        $Date: 92/05/29 14:45:52 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   EXPORT  RemoveDefItem

   EXPORT  InsertDefItem
                                           MakeUpper
   HIDDEN  DefSelect
                                           AddSymbol
					   GetDefinition
					   GetNextNode
					   IsLastNode
					   IsTerminal
					   NewNode
					   PutDefined
					   UpdateInterface
   EXPORT  CreateDefSymbolList
                               
*/

/* HISTORY
   $Log:	deflist.c,v $
 * Revision 1.9  92/05/29  14:45:52  boyd
 * Changes for improved resizing, consisting mostly of replacing Form widgets
 * with Group, Table, and Viewform widgets.
 * 
 * 
 * Revision 1.8  92/03/18  14:15:03  ware
 * Working on consistent interface.
 * 
 * Revision 1.7  1992/02/20  04:46:24  cso
 * Made the insertion routines case sensitive although the
 * alphabetization is case insensitive, i.e., find the spot where the
 * symbol should go case-insensitively (there's a new term...), but check
 * to see if it is there in a case-sensitive manner.
 *
 * Revision 1.6  1992/01/28  14:45:13  cso
 * Corrected a problem in RemoveDefItem.  I was not freeing the bottom of
 * the list properly.  Simple reordering of index decrement fixed it.
 *
 * Revision 1.5  1992/01/14  17:49:20  cso
 *  - Reorganizing header files.
 *  - Fixed a problem about freeing up the next available spot before it
 * had been malloc'd.
 *
 * Revision 1.4  1992/01/03  21:50:23  cso
 * Using the new attribute popup panel and removed the use of Xlwh
 * widgets.
 *
 * Revision 1.3  1991/11/20  00:02:24  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.2  1991/11/14  14:31:12  cso
 * types.h utils.c utils.h CVS: CVS: Enter Log.  Lines beginning with
 * 'CVS: ' are removed automatically CVS:
 * ----------------------------------------------------------------------
 * Made the necessary additions and changes to provide a mechanism to
 * change symbol types between nonterminal types Nonterminal & Grouping.
 * In SGML terms, therefore, intermediate form terms, this is the
 * difference between having element tags and not.
 *
 * Revision 1.1.1.3  1991/10/04  20:15:36  jbm
 * First HaL mods to ICA
 *
 * Revision 2.0  91/03/30  13:47:18  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.13  91/02/21  09:58:02  cso
 * Completed documentation.
 * Reworked the use of Defndx so that it always indicates
 * 	1 - Number of symbols in the list
 * 	2 - The next position to place a symbol (similar to Ndx)
 * 
 * Revision 1.12  90/11/20  15:22:20  cso
 * Changes from lint and to AddSymbol.
 * 
 * Revision 1.11  90/11/05  18:01:58  cso
 * Added function template to all functions, but still need to fill it in.
 * Changed the RemoveDefItem so that the list would be NULL and not show 
 * anything unsightful without sensitizing.
 * Cosmetics.  Also, fixed the Insert so that all items get inserted.
 * 
 * Revision 1.10  90/10/24  15:33:43  cso
 * Changed the order of UpdateInterface.
 * 
 * Revision 1.9  90/10/04  15:52:47  cso
 * Removed unnecessary resource declaration from deflist.
 * 
 * Revision 1.8  90/10/02  13:30:26  cso
 * Made the list shorter as a default.
 * 
 * Revision 1.7  90/09/25  15:37:43  cso
 * Changed DefSelect to only reuse a symbol if a symbol position is 
 * available.  Otherwise, it goes to the selected symbol.
 * 
 * Revision 1.6  90/09/07  09:38:14  cso
 * Used some of the new NODE routines.
 * 
 * Revision 1.5  90/09/03  18:23:42  cso
 * Function name changes.
 * 
 * Revision 1.4  90/09/03  14:30:02  cso
 * Changed the names to be more consistent.
 * 
 * Revision 1.3  90/08/29  17:46:58  cso
 * Changed some of the spacing and look.  Have to be careful with the
 * viewport arguments because they are right at 10 now.
 * 
 * Revision 1.2  90/08/26  18:43:59  cso
 * changed include file name - was list.h needed to be deflist.h
 * 
 * Revision 1.1  90/08/26  17:39:44  cso
 * Initial revision
 * 
 * Revision 1.2  90/08/26  14:13:22  cso
 * added module template
 * 
*/

/* #INCLUDES */

#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "att.h"
#include "get.h"
#include "main.h"
#include "node.h"
#include "put.h"
#include "status.h"
#include "symtab.h"
#include "utils.h"



/*
   FUNCTION: 		RemoveDefItem

   Synopsis: 
       Remove an item from the list of defined symbols.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above, except that the list includes all symbols not just 
       defined ones.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Frame             To access other widgets
       Defsymlist        The list of symbols
       Defndx            Index into the next available spot on the list

   Globals Modified:
       Defsymlist        Possibly one item less
       Defndx            Possibly one less

   Return value:	void

   Range:		none
*/

void
RemoveDefItem(str)

String
   str;   /* The symbol to remove from the list */

{
   int
      i,  /* Loop and index counter */
      j;  /* Loop and index counter */


   /*
    * Find the spot in the list where the string exists, if at all
    */
   for (i = 0; ((i < Defndx) && (strcmp(str, Defsymlist[i]) != 0)); i++)
   {
   }

   /*
    * If the spot is before an open spot in the list, remove the symbol
    */
   if (i < Defndx)
   {

      /*
       * From that spot to the end of the list, move each symbol up
       */
      for (j = i; (j < (Defndx - 1)); j++)
      {
	 XtFree(Defsymlist[j]);
	 Defsymlist[j] = XtNewString(Defsymlist[j + 1]);
      }

      /*
       *  Decrement the index, free up the last spot, and assign NULL
       */
      Defndx--;
      XtFree(Defsymlist[Defndx]);
      Defsymlist[Defndx] = NULL;

      /*
       * Update the displayed list
       */
      XawListChange(XtNameToWidget(Frame, "listframe*deflist"),
		    Defsymlist, Defndx, 0, True);

      /*
       * If that was the last symbol in the list, make the list insensitive
       */
      if (Defndx == 0)
      {
	 XtSetSensitive(XtNameToWidget(Frame, "listframe*defview"), False);
      }
   }
   return;
}




/*
   FUNCTION: 		InsertDefItem

   Synopsis: 
       Insert a symbol into the defined symbol list.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above, except that all symbols are in this list, not just
       defined symbols.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Frame             To access the list widget
       Defsymlist        The list of symbols
       Defndx            Index into the list

   Globals Modified:
       Defsymlist        May contain one more symbol
       Defndx            May be incremented

   Return value:	void

   Range:		none
*/

void
InsertDefItem(str)

String
   str;         /* The symbol to inserted in the list */

{
   int
      i,        /* Loop and index counter */
      j;        /* Loop and index counter */
   String
      upperstr; /* Symbol converted to uppercase, so that alphabetization 
		   can be done without regard for case */


   /*
    * Convert the symbol to uppercase
    */
   upperstr = MakeUpper(str);

   /*
    * Find a spot to insert the symbol
    */
   for (i = 0;
	((i < Defndx) &&
	 (strcmp(upperstr, MakeUpper(Defsymlist[i])) > 0));
	i++)
   {
   }

   /*
    * If there is a match str == Defsymlist[i], just return
    * If there are no spots (i == Defndx) or a spot exists 
    * Yet another possibility is the uppercase (or mixed case version)
    * of the same words matching, but not in their normal case, so insert
    * the symbol
    */
   if ((i == Defndx) ||
       (strcmp(upperstr, MakeUpper(Defsymlist[i])) < 0) ||
       (strcmp(str, Defsymlist[i]) != 0))
   {

      /*
       * Move the symbols from the end of the list to the next spot
       */
      for (j = Defndx; j > i; j--)
      {
	 if (j != Defndx)
	 {
	    XtFree(Defsymlist[j]);
	 }
	 Defsymlist[j] = XtNewString(Defsymlist[j - 1]);
      }

      /*
       * Insert the symbol and increment Defndx
       */
      Defsymlist[i] = XtNewString(str);
      Defndx++;

      /*

       * Update the displayed list widget and make sure it is sensitive
       */
      XawListChange(XtNameToWidget(Frame, "listframe*deflist"),
		    Defsymlist, Defndx, 0, True);
      XtSetSensitive(XtNameToWidget(Frame, "listframe*defview"), True);

      /*
       * If we are at our limit, warn the user
       */
      if (Defndx == MAXSYMBOLS)
      {
	 XicaStatus(Statw,
		    "WARNING: InsertDefItem: Defndx is maxed out [%d]\n",
		    Defndx);
	 XicaStatus(Statw, "Increase MAXSYMBOLS in defs.h and recompile\n");
      }
   }
   return;
}




/*
   FUNCTION: 		DefSelect

   Synopsis: 
       If the user selects an itme in the list, reuse that symbol or
       move to that symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Tree              The current focus of the interface

   Globals Modified:
       Tree              May be moved to a different symbol

   Return value:	void

   Range:		none
*/

static void
DefSelect(w, client_data, current)

Widget
   w;           /* The list widget calling this function */
caddr_t
   client_data; /* Unused */
XawListReturnStruct
   *current; /* The symbol and its index selected by the user */

{
   Node
      *oldptr;  /* The position of the focus before the selection occurred */

   /*
    * Unhighlight the selection
    */
   XawListUnhighlight(w);
#ifdef DEBUG
   fprintf(stderr, "selection = [%s], position = [%d]\n",
	   current->string, current->list_index);
#endif

   /*
    * Save the current focus
    */
   oldptr = Tree;

   /*
    * If the current focus is not a NULL symbol,
    * Then jump to the selected symbol
    * Else reuse the selected symbol in the NULL symbol's place
    */
   if (*Tree->symbol != '\0')
   {
      /*
       * Find the definition that uses the selected symbol and
       * Move Tree to point to it
       * Update the interface
       */
      Tree = GetDefinition(current->string);
      UpdateInterface(oldptr, Tree);
   }
   else
   {
      /*
       * Add the selected symbol
       */
      AddSymbol(Tree, current->string, GetSymbolType(current->string),
		False);

      /*
       * If the selected symbol is a nonterminal,
       * puts its defined value in the tree and get an occurrence
       */
      if (IsTerminal(Tree->symbol) == TriF)
      {
	 PutDefined(Tree, IsDefined(Tree->symbol));
	 switch (GetSymbolType(Tree->symbol))
	 {
	 default:
	    break;
	    
	 case Part:
	    if (Tree->typew != NULL)
	    {
	       PutLabel(Tree->typew, "N");
	       XtSetSensitive(Tree->typew, True);
	       PopupAttPanel(Attnon, False);
	    }
	    break;
	    
	 case Group:
	    if (Tree->typew != NULL)
	    {
	       PutLabel(Tree->typew, "G");
	       XtSetSensitive(Tree->typew, True);
	       PopupAttPanel(Attgrp, False);
	    }
	    break;
	    
	 case File:
	    if (Tree->typew != NULL)
	    {
	       PutLabel(Tree->typew, "F");
	       XtSetSensitive(Tree->typew, True);
	       PopupAttPanel(Attfile, False);
	    }
	    break;
	 }
      }
      else
      {
	 /*
	  * If the selected symbol is a Terminal
	  * it is definitely defined without an occurrence or typew
	  */
	 PutDefined(Tree, True);
	 if (Tree->occurw != NULL)
	 {
	    XtDestroyWidget(Tree->occurw);
	    Tree->occurw = NULL;
	 }
	 if (Tree->typew != NULL)
	 {
	    XtDestroyWidget(Tree->typew);
	    Tree->typew = NULL;
	 }
	 Tree->occurrence = "^";

	 /*
	  * If the position where I added the selected symbol is the
	  * end of a production, add a new spot for a symbol
	  */
	 if (IsLastNode(Tree) == True)
	 {
	    Tree->next = NewNode(Tree->parent, Tree);
	    Tree = GetNextNode(Tree);
	 }

	 /*
	  * Update the interface
	  */
	 UpdateInterface(oldptr, Tree);
      }
   }
   return;
}




/*
   FUNCTION: 		CreateDefSymbolList

   Synopsis: 
       Create the interface for the defined symbol list.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above, except that the list is for all symbols.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void CreateDefSymbolList( parent, col, top, bottom )
Widget
   parent,
   col,             /* column gadget */
   top, bottom;     /* row gadgets */
{
   Widget
      deflabel,     /* Label widget with title of list */
      deflist,      /* List widget */
      defview;      /* Viewport so that list can grow beyond screen */

   deflabel = XtVaCreateManagedWidget( "deflabel",
       labelWidgetClass, parent,
       XtNlabel,       "Available Symbols:",
       XtNborderWidth, 0,
       XtNtableRow,    top,
       XtNtableCol,    col,
       NULL );

   defview = XtVaCreateManagedWidget( "defview",
       viewportWidgetClass, parent,
       XtNallowHoriz,    True,
       XtNallowVert,     True,
       XtNuseBottom,     False,
       XtNuseRight,      False,
       XtNforceBars,     True,
       XtNpreference,    False,
/*
       XtNprefWidth,     (FRAMEWIDTH/2),
       XtNminWidth,      100,
       XtNprefHeight,    300,
       XtNminHeight,     50,
*/
       XtNtableRow,      bottom,
       XtNtableCol,      col,
       XtNhorzAlignment, taFill,
       XtNvertAlignment, taFill,
       NULL );

   deflist = XtVaCreateManagedWidget( "deflist",
       listWidgetClass, defview,
       XtNdefaultColumns, 1,
       XtNforceColumns,   True,
       XtNlist,           Defsymlist,
       XtNnumberStrings,  0,
       NULL );
   XtAddCallback( deflist, XtNcallback, DefSelect, NULL );

   return;
}
