/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: fileio.c,v $
   Purpose:
       Handle all of the I/O between DeveGram and the user.  
       Loading a file is done in a different module, 
       but saving is done here.

   Programmer:  Con O'Connell
   Reviewer:    
   Revision:    $Revision: 1.8 $
   Date:        $Date: 92/03/18 14:15:22 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  SetFilePrompt

   HIDDEN  GetFilename
                           SetFilePrompt
			                   XlwhChangeFileAndLabel
					   XlwhCheckFile
					   XlwhFileDialogGetButton
					   XlwhFileFreeString
                                           XlwhPopup
   EXPORT  OpenFile
                           GetFileName
   EXPORT  SaveFile
			   SaveFile
					   GetSymbolType
                                           IsFirstNode
   EXPORT  CreateFileIOShell
                                           XlwhMakeFileShell
*/

/* HISTORY
   $Log:	fileio.c,v $
 * Revision 1.8  92/03/18  14:15:22  ware
 * Working on consistent interface.
 * 
 * Revision 1.9  92/03/07  18:39:43  boyd
 * Changed (hardcoded) prompt for FileDlg from "" to NULL, so that it doesn't
 * take up display space (it gets unmanaged when NULL, not when "").
 * 
 * Revision 1.8  92/03/07  18:38:02  boyd
 * As received from cso, before my changes.
 * 
 * Revision 1.1  92/03/07  18:36:39  boyd
 * Initial revision
 * 
 * Revision 1.7  1992/02/06  04:43:23  cso
 * Modified titles and names for the file dialog
 *
 * Revision 1.6  1992/01/30  15:00:36  cso
 * The file dialog name changed
 *
 * Revision 1.5  1992/01/28  15:18:17  cso
 * Added a cast to the file dialog widget to silence lint
 *
 * Revision 1.4  1992/01/14  19:56:25  cso
 *  - Reorganizing header files.
 *  - Added the intermediate form parser extension.
 *
 * Revision 1.3  1992/01/03  21:48:13  cso
 * Removed all Xlwh uses.
 * Using new FileDlg widgets.
 *
 * Revision 1.2  1991/11/20  00:02:53  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.1.1.3  1991/10/04  20:15:46  jbm
 * First HaL mods to ICA
 *
 * Revision 2.0  91/03/30  13:47:53  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.18  91/03/28  14:42:11  cso
 * Fixed SaveFile not to output a NULL symbol.
 * Removed the New and Abandon buttons from the shell.
 * 
 * Revision 1.17  91/02/22  09:20:57  cso
 * Completed documentation.
 * Moved the routines dealing with loading a file to another module
 * 	loadfile.c
 * 
 * Revision 1.16  90/11/29  14:22:54  cso
 * Added a warning message (eventual use of a message routine) for informing
 * the user that a symbol never got an occurrence assigned to it (perhaps
 * because the program terminated prematurely before an occurrence was
 * received by the program).
 * Played with the resources for the FileDialog shell.
 * 
 * Revision 1.15  90/11/26  13:11:59  cso
 * Cosmetic and lint modifications.
 * 
 * Revision 1.14  90/11/20  15:23:37  cso
 * Changes from lint, AddSymbol, and a check for an undefined occurrence value.
 * 
 * Revision 1.13  90/11/05  18:03:57  cso
 * Added function template to all functions, but still need to fill it in.
 * Changed OpenFile to actually open the file and return a file pointer.
 * The old OpenFile became GetFilename.  Changed some proc names, and added
 * switch statements for cosmetics.  Also broke LoadFile down a bit into
 * two more procedures.
 * 
 * Revision 1.12  90/10/24  15:33:49  cso
 * Changed mallocs to XtMallocs
 * 
 * Revision 1.11  90/10/04  15:54:36  cso
 * Added XtUnmanage/ManageChild routines to keep Treeshell from looking
 * too silly as it loads in a file.
 * 
 * Revision 1.10  90/10/02  13:31:06  cso
 * Fixed a switch statement.  Added XtDestroyWidget on the Root nodew
 * because it was not getting the symbol value added to it properly.
 * 
 * Revision 1.9  90/09/25  15:39:25  cso
 * Changed default path.
 * 
 * Revision 1.8  90/09/07  09:39:03  cso
 * cosmetics.
 * 
 * Revision 1.7  90/09/03  18:24:02  cso
 * Function name changes.
 * 
 * Revision 1.6  90/09/03  14:32:05  cso
 * Changed function names to be more consistent, and added #ifdef XLWH
 * so that this file can be used without the library.
 * 
 * Revision 1.5  90/08/26  17:40:38  cso
 * Comment all calls RE: auxiliary information.
 * 
 * Revision 1.4  90/08/26  14:13:10  cso
 * added module template
 * 
*/

/* #INCLUDES */

#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "get.h"
#include "main.h"
#include "node.h"
#include "status.h"



/*
   FUNCTION: 		SetFilePrompt

   Synopsis: 
       Create the Prompt that is displayed when asking the user
       to specify a file whether it be for loading or saving.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	String

   Range:		"" or extension-specific string
*/

static void
SetFilePrompt(loadflag, exten, filter, prompt, title)

Boolean
   loadflag;  /* Indicating a load or save operation */
String
   exten,     /* The extension of the file being loaded or saved */
   *filter,
   *prompt,
   *title;

{
   /*
    * Initialize the prompt in case an unsupported extension is passed in
    */
   *prompt = "";
   *title  = "";
   *filter = "*";

   /*
    * If we are loading a file, it could only be a .dgm file,
    * else it is a file being saved, .dgm, or generated, one of the 
    * supported notations
    */
   if (loadflag == True)
   {
      *title  = "Edit a file [.dgm]";
      *prompt = "Edit";
      *filter = "*.dgm";
   }
   else if (strcmp(exten, "dgm") == 0)
   {
      *title  = "Save a file [.dgm]";
      *prompt = "Save";
      *filter = "*.dgm";
   }
   else if (strcmp(exten, "bnf") == 0)
   {
      *title  = "Generate a Backus-Naur Form file [.bnf]";
      *prompt = "Generate";
      *filter = "*.bnf";
   }
   else if (strcmp(exten, "dtd") == 0)
   {
      *title  = "Generate an SGML DTD file [.dtd]";
      *prompt = "Generate";
      *filter = "*.dtd";
   }
   else if (strcmp(exten, "ica") == 0)
   {
      *title  = "Generate an ICA file [.ica]";
      *prompt = "Generate";
      *filter = "*.ica";
   }
   else if (strcmp(exten, "yacc") == 0)
   {
      *title  = "Generate a Yacc file [.yacc]";
      *prompt = "Generate";
      *filter = "*.yacc";
   }
   else if (strcmp(exten, "ifx") == 0)
   {
      *title  = "Generate an Intermediate Form parser file [.ifx]";
      *prompt = "Generate";
      *filter = "*.ifx";
   }
   else
   {
      /*
       * This is an error in some sense, but it would only be a 
       * generation problem, thus it is not fatal
       */
      XicaStatus(Statw, "SetFilePrompt passed an unsupported extension\n");
      *title  = "Open a file [*]";
      *prompt = "Open";
      *filter = "*";
   }

   /*
    * Return the prompt created
    */
   return;
}



/*
   FUNCTION: 		GetFilename

   Synopsis: 
       Get a filename from the user by using Loyde's XlwhPopup, which will 
       not popdown until a valid filename has been specified.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	String

   Range:		"" - no filename, e.g., the user changed his mind
                        a valid filename including path and extension
*/

static String
GetFilename(exten, loadflag)

String
   exten;             /* The extension to be used with the filename */
Boolean
   loadflag;          /* Specifies the operation */

{
   Arg
      args[MAXARGS];  /* List of arguments */
   int
      cnt;            /* Number of arguments */
   String
      filestr,        /* The filename to be returned */
      filter,         /* The filter of files */
      name,
      name_ext,
      path,
      prompt,         /* The prompt created by SetFilePrompt */
      title;
   Widget
      dialog;         /* The dialog widget requiring modification */

   filestr = XtNewString("./test.dgm");
   SetFilePrompt(loadflag, exten, &filter, &prompt, &title);
   
   dialog = XtNameToWidget(Toplevel, "*filedlg");

   if (dialog == NULL)
   {
      XicaStatus(Statw, "Error GetFilename: Unable to find filedlg\n");
      return("");
   }
   
   /*
    * Build a filename as a default, using any previous name and
    * the new extension.
    * If a previous name is present, use it with the
    * new extension.
    * Set up the dialog for this invocation.
    * Dialog title text
    * file spec that must be matched to exit
    * prompt for file name input field
    * label for open button
    * must file exist to be opened?
    * read access expecte? 
    * write access expected?
    * confirm overwrite
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNname, &name); cnt++;
   XtGetValues(dialog, args, cnt);

   if (strlen(name) > 0)
   {
      name_ext = (String) XtMalloc(strlen(name) + strlen(exten) + 2);
      strcpy(name_ext, name);
      strcat(name_ext, ".");
      strcat(name_ext, exten);
   }
   else
   {
      name_ext = NULL;
   }

   cnt = 0;
   if ((name != NULL) && (strlen(name) > 0))
   {
      XtSetArg(args[cnt], XtNname, name_ext); cnt++;
   }
   else
   {
      XtSetArg(args[cnt], XtNname, ""); cnt++;
   }
   XtSetArg(args[cnt], XtNtitle,      "File Dialog");     cnt++;
   XtSetArg(args[cnt], XtNdlgTitle,   title);     cnt++;
   XtSetArg(args[cnt], XtNspec,       filter);    cnt++;
   XtSetArg(args[cnt], XtNprompt,     NULL);      cnt++;
   XtSetArg(args[cnt], XtNbutton,     prompt);    cnt++;
   XtSetArg(args[cnt], XtNmustExist,  loadflag);  cnt++;
   XtSetArg(args[cnt], XtNdirsFirst,  False);     cnt++;
   XtSetArg(args[cnt], XtNfilesFirst, True);      cnt++;
   XtSetArg(args[cnt], XtNwillRead,   True);      cnt++;
   XtSetArg(args[cnt], XtNwillWrite,  !loadflag); cnt++;
   XtSetArg(args[cnt], XtNconfirm,    True);      cnt++;
   XtSetValues(dialog, args, cnt);
   
   /*
    * Pop up the dialog and wait for a response from the user
    */
   DoFileDialog((FileDlgWidget) dialog);

   /*
    * If a file was selected, the name will have contents.
    * Concatenate the returned path and name.
    * In all cases, the name will also have an extension; strip the
    * extension from the name so a new one can be appended next time.
    * Also, if the dialog was cancelled, we want to keep the name,
    * but it will have been cleared, so use the one that was passed in.
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNname, &name); cnt++;
   XtSetArg(args[cnt], XtNpath, &path); cnt++;
   XtGetValues(dialog, args, cnt);

   XtFree(filestr);
   filestr = (String) XtMalloc(strlen(path) + strlen(name) + 1);
   if (strlen(name) == 0)
   {
      strcpy(filestr, "");
   }
   else
   {
      strcpy(filestr, path);
      strcat(filestr, name);
      if (name_ext != NULL)
      {
	 XtFree(name_ext);
      }
      name_ext = (String) XtMalloc(strlen(name) + 1);
      strcpy(name_ext, name);
   }
   if (name_ext != NULL)
   {
      /*
       * Cut off the extension
       * Set the name (for next time) to be the nonextended filename
       */
      name_ext[strlen(name_ext) - strlen(exten) - 1] = '\0';
      cnt = 0;
      XtSetArg(args[cnt], XtNname, name_ext); cnt++;
      XtSetArg(args[cnt], XtNpath, path);     cnt++;
      XtSetValues(dialog, args, cnt);
      XtFree(name_ext);
   }
   
#ifdef DEBUG
   fprintf(stderr,"filestr = [%s]\n", filestr);
#endif
   return(filestr);
}




/*
   FUNCTION: 		OpenFile

   Synopsis: 
       Get a filename from the user and open that file and return the 
       file pointer.

   Initial purpose:
       Get a filename from the user..

   Present purpose:
       Get the name and perform the opening of the file as the function
       name suggests.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	FILE *

   Range:		NULL or a valid file pointer
*/

FILE
*OpenFile(exten, loadflag)

String
   exten;    /* The extension for the file */
Boolean
   loadflag; /* Determines the operation load/save - read/write */

{
   String
      filestr;  /* The name of the file including explicit path */
   FILE
      *fptr;    /* File pointer to return */

   /*
    * Get a filename from the user
    */
   filestr = GetFilename(exten, loadflag);

   /*
    * If the filename is not "", perform the open based on the loadflag
    * Else return NULL, no file pointer for a nonexistent file.
    *
    * The pointer returned is NULL because of two reasons - 
    *   1) a nonexistent file was chosen by the user, i.e., "", or 
    *   2) opening the file failed
    * Or it is a valid file pointer to a file opened per loadflag
    */
   if (*filestr != '\0')
   {
      if (loadflag == True)
      {
	 fptr = fopen(filestr, "r");
      }
      else
      {
	 fptr = fopen(filestr, "w");
      }

      /*
       * If we were unsuccessful in opening a file, inform the user
       */
      if (fptr == NULL)
      {
	 XicaStatus(Statw, "Unable to open file\n\t[%s]\n", filestr);
      }
      return(fptr);
   }
   else
   {
      return(NULL);
   }
}




/*
   FUNCTION: 		SaveFile

   Synopsis: 
       Save the grammar description in a file using DeveGram's syntax.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/22   cso    Only output a symbol if it is nonNULL

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
SaveFile(nptr, fptr)

Node
   *nptr;  /* The node currently being saved */
FILE
   *fptr;  /* The file pointer to the file being saved */

{
   Alt
      *aptr; /* Pointer to the definitions of nptr */
   Aux
      *uptr; /* Pointer to the auxiliary information associated with
		each alternative definition of nptr */
   Node
      *qptr; /* Pointer to access each child in the 
		alternative definitions of nptr */

   /*
    * If nptr has a definition, i.e., not a leaf
    */
   if ((nptr->alts != NULL) && (*nptr->symbol != '\0'))
   {
      /*
       * Output the symbol being defined
       * For each alternative definition, output a production
       */
      fprintf(fptr, "!%s\n", nptr->symbol);
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 /*
	  * For each node in a definition, output its type and occurrence
	  * Only if it is a nonNULL symbol
	  */
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       /*
		* The first node gets an additional marker
		*/
	       if (IsFirstNode(qptr) == True)
	       {
		  fprintf(fptr, "&");
	       }
	       
	       /*
		* Output the symbol type
		*/
	       switch (GetSymbolType(qptr->symbol))
	       {
	       default:
		  XicaStatus(Statw, "Error: Bad SymbolType in SaveFile\n");
		  break;
		  
	       case File:
		  fprintf(fptr, "F");
		  break;
		  
	       case Group:
		  fprintf(fptr, "G");
		  break;
		  
	       case Part:
		  fprintf(fptr, "P");
		  break;
	       }
	       
	       /*
		* Output the occurrence, including a check for an 
		* unspecified occurrence
		*/
	       if (*qptr->occurrence == ' ')
	       {
		  fprintf(fptr, "$");
	       }
	       else
	       {
		  fprintf(fptr, "%s",  qptr->occurrence);
	       }
	       
	       /*
		* Output the symbol string
		*/
	       fprintf(fptr, "%s ", qptr->symbol);
	    } /* IF it is not a NULL symbol */
	 } /* FOR each symbol in the definition */

	 /*
	  * Output the end of that production
	  */
	 fprintf(fptr, "\n");
#ifdef AUX

	 /*
	  * If any auxiliary information for that definition exists,
	  * For each A.I. for this definition, output the type of A.I.,
	  * length of A.I., and actual string
	  */
	 if (aptr->auxinf != NULL)
	 {
	    for (uptr = aptr->auxinf; uptr != NULL; uptr = uptr->next)
	    {
	       fprintf(fptr, "#%s %d %s\n",
		       uptr->type, strlen(uptr->info), uptr->info);
	    }
	 }
#endif
      } /* FOR each definition of nptr */

      /*
       * Output a separator between symbols being defined
       */
      fprintf(fptr, "\n");

      /*
       * For each definition of nptr, for each node, call SaveFile
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    SaveFile(qptr, fptr);
	 }
      }
   } /* IF nptr has a definition */
   return;
}




/*
   FUNCTION: 		CreateFileIOShell

   Synopsis: 
       Create the shell that gets a filename from the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/22   cso    Turned off some buttons in the file shell

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
CreateFileIOShell(parent)

Widget
   parent;  /* The parent of this widget */

{
   SetupFileDialog(parent);

   return;
}
