/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: filemenu.c,v $
   Purpose:     
       To create the menu choices for the FILE menu, and their functions.

   Programmer:  Con O'Connell
   Reviewer:    
   Revision:    $Revision: 1.7 $
   Date:        $Date: 92/05/29 14:46:03 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  SaveChanges
                                           Alert
					   OpenFile
					   SaveFile
   HIDDEN  Reset
                           SaveChanges
					   Initialize
   HIDDEN  Edit
                           SaveChanges
					   Initialize
					   EditFile
					   OpenFile
					   PrepareDisplayTree
					   PrettyPrintTree
					   UpdateInterface
					   WipeWindow
   HIDDEN  Save
                                           Alert
					   OpenFile
					   SaveFile
   HIDDEN  Quit
                           SaveChanges
                                           CheckTerminates
					   InitTerminate
   EXPORT  CreateFileMenu
                                           
*/

/* HISTORY
   $Log:	filemenu.c,v $
 * Revision 1.7  92/05/29  14:46:03  boyd
 * Changes for improved resizing, consisting mostly of replacing Form widgets
 * with Group, Table, and Viewform widgets.
 * 
 * 
 * Revision 1.6  92/03/18  14:15:29  ware
 * Working on consistent interface.
 * 
 * Revision 1.5  1992/02/21  05:45:29  cso
 * Added a cancel option to the save changes dialog
 *
 * Revision 1.4  1992/01/14  17:56:30  cso
 *  - Reorganizing header files.
 *
 * Revision 1.3  1992/01/03  21:46:06  cso
 * Use of the message box widgets.
 *
 * Revision 1.2  1991/11/20  00:03:03  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.1.1.3  1991/10/04  20:15:49  jbm
 * First HaL mods to ICA
 *
 * Revision 2.0  91/03/30  13:48:07  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.7  91/03/28  14:43:12  cso
 * Added a confirmation to reset the interface.
 * 
 * Revision 1.6  91/02/22  09:25:05  cso
 * Completed documentation.
 * Added the function SaveChanges to check with the user, if changes 
 * should be saved or not.
 * Reset was simplified to just calling Initialize.
 * Still working on Load cooperating with the tree display.
 * Quit was combined with conclude (no longer exists).
 * 
 * Revision 1.5  90/11/26  13:13:30  cso
 * Cosmetic and lint modifications.
 * 
 * Revision 1.4  90/11/20  15:25:09  cso
 * Changes to Load to eliminate the time Prettyprinting.
 * Added blank lines in the menu.
 * 
 * Revision 1.3  90/11/05  18:12:01  cso
 * Added function template to all functions, but still need to fill it in.
 * Taking advantage of the new OpenFile.  Modified Quit to check Dirty_Flag.
 * Else, cosmetics.
 * 
 * Revision 1.2  90/10/29  10:19:55  cso
 * Added the appropriate decls to make it work.
 * 
 * Revision 1.1  90/10/25  13:33:59  cso
 * Initial revision
 * 
*/

/* #INCLUDES */

#include "std.h"
#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "display.h"
#include "fileio.h"
#include "loadfile.h"
#include "main.h"
#include "status.h"
#include "symtab.h"
#include "tree.h"
#include "utils.h"



/*
   FUNCTION: 		SaveChanges

   Synopsis: 
       Check if the  current session has changed a grammar description.
       Offer the user to save or forget changes AND if unsuccessful
       in saving let them know and let them leave the function
       invoking this one without losing the changes.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Globals used:
       Dirtyflag  

   Globals Modified:
       none

   Return value:	Boolean

   Range:		True or False
*/

static Boolean
SaveChanges()

{
   Boolean
      continue_oper,
      response;
   FILE
      *fptr;

   /*
    * If the current session has changed, give the user a chance to
    * save the file or at least realize that changes have occurred.
    */
   continue_oper = True;
   if (Dirtyflag == True)
   {
      /*
       * Ask the user if they wish to save the changes, if Continue, 
       * then get a file name and save the file
       */
      continue_oper = YesNoCancelMessage(Toplevel, "Save changes?",
					 True, &response);
      
      if ((continue_oper == True) && (response == True))
      {
	 /*
	  * Get the file to save the current session
	  */
	 fptr = OpenFile("dgm", False);
	 
	 /*
	  * If successful in getting a file, save out the current session
	  * Else alert the user that changes will be lost
	  */
	 if (fptr != NULL)
	 {
	    SaveFile(Root, fptr);
	    fclose(fptr);
	    Dirtyflag = False;
	 }
	 else
	 {
	    /*
	     * If the user does not wish to lose changes,
	     * then return False, preventing whatever action would lose
	     * the changes from occurring.
	     */
	    return(YesNoMessage(Frame, "Lose changes?", False));
	 }
      }
   }
   return(continue_oper);
}




/*
   FUNCTION: 		Reset

   Synopsis: 
       Put DeveGram into an initialized state after ensuring that a 
       previous session has been saved (or at least the option to do so).

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/08   cso    Added a confirmation on resetting the interface

   Globals used:
       Dirtyflag Need to check if a current session needs saving
       Ndx       Need to know how much to initialize
       Root      The initial node to start saving

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
Reset(w, client_data, event)

Widget
   w;           /* Unused */
caddr_t
   client_data; /* Unused */
XEvent
   *event;      /* Unused */

{
   if (SaveChanges() == True)
   {
      /*
       * Initialize devegram, after confirming with user
       */
      if (YesNoMessage(Frame, "Reset interface?", True) == True)
      {
	 XtSetKeyboardFocus(Frame, None);
	 Initialize(Ndx);
	 Dirtyflag = False;
      }
   }
   return;
}




/*
   FUNCTION: 		Edit

   Synopsis: 
       Call the functions to Edit a file and update the interface.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Dirtyflag  To give the user a chance to save the current session
       Ndx        To initialize the interface
       Root       Needed for some of the recursive procedures

   Globals Modified:
       Dirtyflag  Loading a file will cause the dirtyflag to be set even
                  though no changes have occurred from the user
       Tree       Needs to be updated to point to the Root

   Return value:	void

   Range:		none
*/

static void
Edit(w, client_data, event)

Widget
   w;           /* Unused */
caddr_t
   client_data; /* Unused */
XEvent
   *event;      /* Unused */

{
   FILE
      *fptr;   /* The file to load */

   if (SaveChanges() == True)
   {
      /*
       * Get the file to load
       */
      fptr = OpenFile("dgm", True);
      if (fptr != NULL)
      {
	 /*
	  * Initialize the interface
	  */
	 XtSetKeyboardFocus(Frame, None);
	 Initialize(Ndx);
	 
	 /*
	  * Load in a file without doing layout all of the time 
	  * in the tree view
	  */
	 XawFormDoLayout(XtNameToWidget(Treeshell, "*treeframe"), False);
	 EditFile(fptr);
	 XawFormDoLayout(XtNameToWidget(Treeshell, "*treeframe"), True);
	 fclose(fptr);
	 
	 /*
	  * Clear off the tree view and pretty print the tree, 
	  * update the interface and tree pointer
	  * reset the dirtyflag from loading in the file
	  */
	 (void) PrettyPrintTree(Root, (Widget) NULL, True);
#ifdef ThreeD
	 
	 /*
	  * Display the initial two-dimensional version of the tree
	  */
	 PrepareDisplayTree(Root, True);
	 WipeWindow();
#endif
	 Tree = Root;
	 UpdateInterface(Root, Root);
	 Dirtyflag = False;
      }
   }
   return;
}




/*
   FUNCTION: 		Save

   Synopsis: 
       Save the current session into a file.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/08   cso    Removed the RemoveNullSymbols call, it caused
		          Tree to get out of date.

   Globals used:
       Dirtyflag  Checked to give the user a chance to skip
       Root       Used as a reference in recursive calls

   Globals Modified:
       Dirtyflag  May get reset

   Return value:	void

   Range:		none
*/

static void
Save(w, client_data, event)

Widget
   w;           /* Unused */
caddr_t
   client_data; /* Unused */
XEvent
   *event;      /* Unused */

{
   FILE
      *fptr; /* File to save */

   /*
    * If nothing has changed, give the user a chance to skip this
    */
   if (Dirtyflag == False)
   {
      if (YesNoMessage(Frame,
		       "No changes have been made.\nSave anyway?",
		       False) == False)
      {
	 return;
      }
   }

   /*
    * Get and save a file, and reset Dirtyflag
    */
   fptr = OpenFile("dgm", False);
   if (fptr != NULL)
   {
      SaveFile(Root, fptr);
      fclose(fptr);
      Dirtyflag = False;
   }
   return;
}




/*
   FUNCTION: 		Quit

   Synopsis: 
       Check for changes in current session and offer to save them.  
       Check termination of the grammar and inform the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Root  To see if there is a grammar to check for termination.

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
Quit(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   /*
    * Check if changes need to be saved.  If the user does not wish to lose
    * changes, but was unsuccessful in saving them, just return.
    */
   if (SaveChanges() == True)
   {
      /*
       * If the grammar is not NULL, check termination of it.
       */
      if (*Root->symbol != '\0')
      {
	 /*
	  * Initialize the termination values in the symbol table
	  */
	 InitTerminate();

	 /*
	  * Depending on the return of CheckTerminates inform the user.
	  */
	 if (CheckTerminates(Root) == False)
	 {
	    XicaStatus(Statw, "WARNING: Grammar does not terminate\n");
	 }
	 else
	 {
	    XicaStatus(Statw, "Grammar terminates\n");
	 }
      }

      /*
       * Quit DeveGram
       */
#ifdef DEBUG
      monitor(0);
#endif
      exit(0);
   }
   return;
}




/*
   FUNCTION: 		CreateFileMenu

   Synopsis: 
       Create the list of options for the FILE menu.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
CreateFileMenu(parent)

Widget
   parent;  /* The parent of this list of menu options */

{
   Arg
      args[MAXARGS];/* List of arguments */
   int
      cnt;      /* Number of arguments */
   Widget
      loadsel,  /* Edit  option widget */
      quitsel,  /* Quit  option widget */
      resetsel, /* Reset option widget */
      savesel;  /* Save  option widget */

   /*
    * Offer RESET first, a line to separate, LOAD and SAVE,
    * another line separator, and lastly QUIT
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNlabel, "Reset    "); cnt++;
   resetsel = XtCreateManagedWidget("resetsel", smeBSBObjectClass,
				    parent, args, cnt);
   XtAddCallback(resetsel, XtNcallback, Reset, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlineWidth, 3); cnt++;
   (void) XtCreateManagedWidget("linesep", smeLineObjectClass,
				parent, args, cnt);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel, "Edit     "); cnt++;
   loadsel = XtCreateManagedWidget("loadsel", smeBSBObjectClass,
				   parent, args, cnt);
   XtAddCallback(loadsel, XtNcallback, Edit, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel, "Save     "); cnt++;
   savesel = XtCreateManagedWidget("savesel", smeBSBObjectClass,
				   parent, args, cnt);
   XtAddCallback(savesel, XtNcallback, Save, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlineWidth, 3); cnt++;
   (void) XtCreateManagedWidget("linesep", smeLineObjectClass,
				parent, args, cnt);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel, "Quit     "); cnt++;
   quitsel = XtCreateManagedWidget("quitsel", smeBSBObjectClass,
				   parent, args, cnt);
   XtAddCallback(quitsel, XtNcallback, Quit, NULL);

   return;
}
