/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: generate.c,v $
   Purpose:
       To provide the functions to generate the different grammar notations.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.8 $
   Date:        $Date: 92/03/18 14:15:38 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   EXPORT  PrintOccur

   HIDDEN  TabGen

   EXPORT  BareGen
                           BareGen
			                   GetDefinition
					   IsTerminal
					   Visited
   EXPORT  ExpandGen
                           ExpandGen
			                   GetDefinition
					   IsTerminal
					   Visited
   EXPORT  LinearGen
                           LinearGen
			   PrintOccur
			                   GetDefinition
					   Visited
   EXPORT  BNF_Gen
                           BNF_Gen
			   PrintOccur
                                           IsTerminal
   EXPORT  ICA_Gen
                           ICA_Gen
			   PrintOccur
                                           GetSymbolType
					   MakeUpper
   EXPORT  ICA_RTT
                                           GetIndex
   EXPORT  ICA_ITT
                           PrintContent
                           PrintGroupParents
   EXPORT  YaccPre
   
   EXPORT  YaccBNF

   EXPORT  YaccGen
                           PrintOccur
			   YaccGen
   EXPORT  IFYaccPre
					   MakeUpper
   EXPORT  IFYaccBNF

   EXPORT  IFYaccGen
                           IFYaccGen
                           PrintOccur
                                           IsFirstAlt
					   IsTerminal
					   GetSymbolType
*/

/* HISTORY
   $Log:	generate.c,v $
 * Revision 1.8  92/03/18  14:15:38  ware
 * Working on consistent interface.
 * 
 * Revision 1.7  1992/02/20  04:47:25  cso
 * Fixed an aesthetic problem in the ica file generation
 *
 * Revision 1.6  1992/02/06  04:44:15  cso
 * Fixed the ICA file generation
 *
 * Revision 1.5  1992/01/14  20:04:28  cso
 *  - Reorganizing header files.
 *  - Updated the ica language generated.
 *  - Added a see above comment for certain views.
 *  - Fixed a problem with BareGen.  It was expanding under the
 * Chosennode even if it was not the right thing to do.  Now it checks
 * for the level as well.
 *  - Initialized flags in DetermineContentType.
 *  - Fixed an open comment.
 *
 * Revision 1.4  1991/11/25  14:36:47  cso
 * Made lint modifications to ensure that function and variable types are
 * being used and declared correctly and consistently.
 *
 * Revision 1.3  1991/11/20  00:03:12  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.2  1991/11/12  23:18:58  cso
 * Added modifications to generate intermediate form yacc parsers.  Still
 * need to generate the lex side of things and to provide the auxiliary
 * information extensions.
 *
 * Revision 1.1.1.3  1991/10/04  20:15:53  jbm
 * First HaL mods to ICA
 *
 * Revision 2.2  91/05/28  12:21:22  cso
 * Added tests so that NULL symbols are not printed in any of the generation
 * routines.
 * Added some linefeeds to the extra production routines (yacc stuff).
 * 
 * Revision 2.1  91/05/28  09:59:59  cso
 * Changed the ICA generation routines in a couple of ways:
 *  - The list of segmenting tags was made uppercase
 *  - A complete ITT spec was generated.
 * 
 * Revision 2.0  91/03/30  13:48:19  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.12  91/03/28  14:43:33  cso
 * Added another notation to generate called DBM (Database mapping).
 * 
 * Revision 1.11  91/02/26  15:23:45  cso
 * Check for a node not having children(terminal symbols).
 * 
 * Revision 1.10  91/02/22  09:28:59  cso
 * Completed documentation.
 * The order of functions in the file changed.  Also some of the names
 * changed.
 * The generation functions for MDL and STRUCT were removed.
 * RightSide was moved to a new module along with SGML generation
 * functions to a new module sgml.c
 * For generating a Yacc file, new functions were created and called
 * so that only one file pointer is open at a time and no system
 * call is used to concatenate the files together.
 * Added more information in the ICA spec file.  Info for RTT and ITT
 * 
 * Revision 1.9  90/11/20  15:26:21  cso
 * Changes from lint.
 * 
 * Revision 1.8  90/11/05  18:13:47  cso
 * Added function template to all functions, but still need to fill it in.
 * Added PrintOccur.  Also used /tmp for temp files.  Changed STR to upperstr
 * to avoid the possibility of case-insensitivity.  Else, cosmetics.
 * 
 * Revision 1.7  90/10/24  15:34:11  cso
 * Changed mallocs to XtMallocs
 * 
 * Revision 1.6  90/10/02  13:32:49  cso
 * Added #ifdef AUX instead of comments.
 * 
 * Revision 1.5  90/09/03  14:32:51  cso
 * Changed function names to be more consistent.
 * 
 * Revision 1.4  90/08/26  17:41:14  cso
 * Comment all calls RE: auxiliary information.
 * 
 * Revision 1.3  90/08/26  14:13:15  cso
 * added module template
 * 
*/

/* #INCLUDES */

#include "widgets.h"
#include "types.h"

#include "get.h"
#include "main.h"
#include "node.h"
#include "status.h"
#include "symtab.h"
#include "utils.h"



/*
   FUNCTION: 		PrintOccur

   Synopsis: 
       Based on the occurrence character for a symbol, output the appropriate
       string, e.g., * in Yacc is STAR, but in BNF it is *.  It depends on
       who is calling this function.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
PrintOccur(fptr, occurchar, termstr, oncestr, optlstr, starstr, plusstr)

FILE
   *fptr;     /* Pointer to file that is being written */
char
   occurchar; /* The occurrence character under consideration */
String
   termstr,   /* The terminal representation to be printed */
   oncestr,   /* The only once representation to be printed */
   optlstr,   /* The optional representation to be printed */
   starstr,   /* The zero or more representation to be printed */
   plusstr;   /* The one or more representation to be printed */

{
   /*
    * Depending on the occurrence character,
    * output the appropriate string
    */
   switch (occurchar)
   {
   default:
      XicaStatus(Statw, "Error: PrintOccur passed bad occurchar [%c]\n",
	      occurchar);
      break;

   case '\0':
      break;
      
   case '^':
      fprintf(fptr, "%s", termstr);
      break;
      
   case '_':
      fprintf(fptr, "%s", oncestr);
      break;
      
   case '?':
      fprintf(fptr, "%s", optlstr);
      break;
      
   case '*':
      fprintf(fptr, "%s", starstr);
      break;
      
   case '+':
      fprintf(fptr, "%s", plusstr);
      break;
   }
   return;
}




/*
   FUNCTION: 		TabGen

   Synopsis: 
       Create indentation string to show nesting.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	String

   Range:		"" or a longer string of spaces
*/

static String
TabGen(n)

int
   n;      /* The nesting level */

{
   String
      str; /* The indentation string */
   int
      i;   /* Loop counter */
   
   /*
    * Allocate space for the indentation string
    */
   str = (char *) XtMalloc((unsigned) (n * 2) + 1);

   /*
    * Initialize the indentation string
    */
   str = strcpy(str, "");

   /*
    * For each indentation level concatenate 2 spaces
    */
   for (i = 0; i < n; i++)
   {
      str = strcat(str, "| ");
   }

   /*
    * Return the string created
    */
   return(str);
}




/*
   FUNCTION: 		BareGen

   Synopsis: 
       Generate the Barest string which is the minimal (shortest?) string
       that represents a legal string in the language.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added a test so that NULL symbols are not printed.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
BareGen(nptr, fptr, depth)

Node
   *nptr;  /* The node under consideration */
FILE
   *fptr;  /* The file being generated */
int
   depth;  /* The nesting level */

{
   Alt
      *aptr;   /* Loop pointer to look at each definition of a symbol */
   Node
      *qptr;   /* Loop pointer to check each node of a production */
   int
      level;   /* Nesting level counter */
   String
      tabstr,  /* The indentation string for the nesting level */
      namestr; /* The symbol string */

   /*
    * Increment the nesting level
    */
   level   = depth + 1;

   /*
    * Assign the symbol string to namestr
    */
   namestr = XtNewString(nptr->symbol);

   /*
    * Get the indentation string
    */
   tabstr  = XtNewString(TabGen(level));

   /*
    * If a symbol's occurrence cannot be zero,
    * generate its production
    * Else do not generate anything, unless it was the initial node
    */
   if ((*nptr->occurrence != '?') && (*nptr->occurrence != '*'))
   {
      /*
       * If the node is a terminal symbol, output it
       * Else generate the barest encapulation
       */
      if (IsTerminal(namestr) == TriT)
      {
	 fprintf(fptr, "%s%s\n", tabstr, namestr);
      }
      else
      {
	 /*
	  * Generate the bracketing BEG[], interior, and END[]
	  */
	 fprintf(fptr, "%sBEG[%s]\n", tabstr, namestr);

	 /*
	  * If we haven't visited a symbol yet then generate the interior
	  * of the barest string.
	  * In other words, we don't want to have the same symbol's
	  * definition expanded more than once
	  */
	 if (Visited(namestr) == 1)
	 {
	    /*
	     * Find the position where the definition of nptr exists
	     */
	    nptr = GetDefinition(namestr);

	    /*
	     * For each of the definitions, separate them with OR, 
	     * and expand the nodes in the productions by the same rules
	     */
	    for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
	    {
	       if (aptr->prev != NULL)
	       {
		  fprintf(fptr, "%sOR\n", tabstr);
	       }
	       for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	       {
		  if (*qptr->symbol != '\0')
		  {
		     BareGen(qptr, fptr, level);
		  }
	       }
	    } /***** FOR my children, keep generating */
	 } /***** IF already visited */
	 else
	 {
	    fprintf(fptr, "%s| /* See above */\n", tabstr);
	 }
	 fprintf(fptr, "%sEND[%s]\n", tabstr, namestr);
      }
   }
   else if ((nptr == Chosennode) && (level == 0))
   {
      /*
       * In case the initial node had an occurrence of * or ?,
       * still use it for a bracket and expand it
       * Otherwise the algorithm is the same as above.
       */
      fprintf(fptr, "%sBEG[%s]\n", tabstr, namestr);

      /*
       * Find the position where the definition of nptr exists
       */
      nptr = GetDefinition(namestr);

      /*
       * Expand each of its definitions
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 if (aptr->prev != NULL)
	 {
	    fprintf(fptr, "%sOR\n", tabstr);
	 }
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       BareGen(qptr, fptr, level);
	    }
	 }
      }
      fprintf(fptr, "%sEND[%s]\n", tabstr, namestr);
   }
   return;
}




/*
   FUNCTION: 		ExpandGen

   Synopsis: 
       Generate a string that shows what symbols are reachable from the 
       initial symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added a test so that NULL symbols are not printed.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
ExpandGen(nptr, fptr, depth)

Node
   *nptr;  /* The node under consideration */
FILE
   *fptr;  /* The file being generated */
int
   depth;  /* The nesting level */

{
   Alt
      *aptr;   /* Loop pointer to look at each definition of a symbol */
   Node
      *qptr;   /* Loop pointer to check each node of a production */
   int
      level;   /* Nesting level counter */
   String
      tabstr,  /* The indentation string for the nesting level */
      namestr; /* The symbol string */
   
   /*
    * Increment the nesting level
    */
   level   = depth + 1;

   /*
    * Assign the symbol string to namestr
    */
   namestr = XtNewString(nptr->symbol);

   /*
    * Get the indentation string
    */
   tabstr  = XtNewString(TabGen(level));

   /*
    * If the node is a terminal symbol, output it
    * Else generate the expanded possibilities
    */
   if (IsTerminal(namestr) == TriT)
   {
      fprintf(fptr, "%s%s\n", tabstr, namestr);
   }
   else
   {
      /*
       * Generate the bracketing BEG[], interior, and END[]
       */
      fprintf(fptr, "%sBEG[%s]\n", tabstr, namestr);

      /*
       * If we haven't visited a symbol yet then generate the interior
       * of the expanded string.
       * In other words, we don't want to have the same symbol's
       * definition expanded more than once
       */
      if (Visited(namestr) == 1)
      {
	 /*
	  * Find the position where the definition of nptr exists
	  */
	 nptr = GetDefinition(namestr);

	 /*
	  * For each of the definitions, separate them with OR, 
	  * and expand the nodes in the productions by the same rules
	  */
	 for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
	 {
	    if (aptr->prev != NULL)
	    {
	       fprintf(fptr, "%sOR\n", tabstr);
	    }
	    for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	    {
	       if (*qptr->symbol != '\0')
	       {
		  ExpandGen(qptr, fptr, level);
	       }
	    }
	 } /***** FOR my children, keep generating */
      } /***** IF already visited */
      else
      {
	 fprintf(fptr, "%s| /* See above */\n", tabstr);
      }
      fprintf(fptr, "%sEND[%s]\n", tabstr, namestr);
   }
   return;
}




/*
   FUNCTION: 		LinearGen

   Synopsis: 
       Generate a linear form of the grammar in extended-BNF-like notation.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/02/23   cso    Additional check to prevent generating a 
                         production for a terminal symbol.
       91/05/28   cso    Added a test so that NULL symbols are not printed.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
LinearGen(nptr, fptr)

Node
   *nptr; /* Node under consideration */
FILE
   *fptr; /* File being output */

{
   Alt
      *aptr;  /* Loop pointer for each definition of the node */
   Node
      *qptr;  /* Loop pointer for each symbol in a definition */

   /*
    * If we haven't visited a symbol yet then expand its definition
    * In other words, we don't want to have the same symbol's
    * definition expanded more than once
    */
   if (Visited(nptr->symbol) == 1)
   {
      /*
       * Find the position where the definition of nptr exists
       */
      nptr = GetDefinition(nptr->symbol);

      /*
       * If we have found a node's definition, expand it
       * 91/02/23 - cso - Added additional check for a terminal
       */
      if ((nptr != NULL) && (IsTerminal(nptr->symbol) != TriT))
      {
	 /*
	  * Output the node, then for each of its definitions, 
	  * output the production
	  */
	 fprintf(fptr, "%s\t->\t", nptr->symbol);
	 for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
	 {
	    /*
	     * Output an OR, "|", symbol if more than one definition exists
	     */
	    if (IsFirstAlt(aptr) == False)
	    {
	       fprintf(fptr, "\t|\t");
	    }

	    /*
	     * Output each symbol and its occurrence in a definition
	     */
	    for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	    {
	       if (*qptr->symbol != '\0')
	       {
		  fprintf(fptr, "%s", qptr->symbol);
		  PrintOccur(fptr, *qptr->occurrence,
			     " ", " ", "? ", "* ", "+ ");
	       }
	    }

	    /*
	     * Output the end of a production, a linefeed in this case
	     */
	    fprintf(fptr, "\n");
	 }

	 /*
	  * If nptr->alts == NULL, output a linefeed
	  */
	 if (nptr->alts == NULL)
	 {
	    fprintf(fptr, "\n");
	 }

	 /*
	  * For each definition, and for each node, do the same
	  */
	 for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
	 {
	    for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	    {
	       LinearGen(qptr, fptr);
	    }
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		BNF_Gen

   Synopsis: 
       Generate an extended-BNF notation for the grammar.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added a test so that NULL symbols are not printed.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
BNF_Gen(nptr, fptr)

Node
   *nptr;  /* Node under consideration for expansion */
FILE
   *fptr;  /* File that is being output */

{
   Alt
      *aptr;  /* Loop pointer for each definition of a symbol */
   Node
      *qptr;  /* Loop pointer to visit each node in a production */

   /*
    * If a node has a definition then output the definitions
    * and expand the nodes in the productions
    */
   if (nptr->alts != NULL)
   {
      /*
       * Output the left hand side
       */
      fprintf(fptr, "<%s>\t::=\t", nptr->symbol);

      /*
       * For each definition output it
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 /*
	  * If this is not the first definition, output the | symbol
	  */
	 if (IsFirstAlt(aptr) == False)
	 {
	    fprintf(fptr, "\t|\t");
	 }

	 /*
	  * For each symbol in a definition, output it and its occurrence
	  * on a single line, then output a linefeed to separate productions
	  */
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       /*
		* If a symbol is a terminal, don't output the <>'s
		*/
	       if (IsTerminal(qptr->symbol) == TriT)
	       {
		  fprintf(fptr, "%s", qptr->symbol);
	       }
	       else
	       {
		  fprintf(fptr, "<%s>", qptr->symbol);
	       }
	       PrintOccur(fptr, *qptr->occurrence, " ", " ", "? ", "* ", "+ ");
	    }
	 }
	 fprintf(fptr, "\n");
      }

      /*
       * Add a blank line between symbols being defined
       */
      fprintf(fptr, "\n");

      /*
       * Expand each of the symbols in each of the definitions
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    BNF_Gen(qptr, fptr);
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		ICA_Gen

   Synopsis: 
       Generate the file format for the ICA.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added a test so that NULL symbols are not printed.
       91/08/05   cso    Generate the default permutation explicitly

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
ICA_Gen(nptr, fptr)

Node
   *nptr;  /* The node under consideration */
FILE
   *fptr;  /* The file being output */

{
   Alt
      *aptr;    /* Loop pointer to for each definition of nptr */
   int
      pos;
   String
      upperstr; /* A string to store the uppercase version of a symbol */
   Node
      *qptr;    /* Loop pointer for each symbol in a definition */

   /*
    * If a node has a definition
    */
   if (nptr->alts != NULL)
   {
      /*
       * Get an upper case version of the symbol
       * Output the left hand side symbol
       */
      upperstr = MakeUpper(nptr->symbol);
      fprintf(fptr, "%s\t: ", nptr->symbol);

      /*
       * Output the symbols making up a definition
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 /*
	  * Output the OR part if this is not the first definition
	  * If the symbol is a Part, output a begin tag for it
	  * Output the symbols comprising the production, and occurrences
	  * If the symbol is a Part, output an end tag for it
	  */
	 if (IsFirstAlt(aptr) == False)
	 {
	    fprintf(fptr, "\t| ");
	 }

	 pos = 1;
	 
	 if (GetSymbolType(nptr->symbol) == Part)
	 {
	    fprintf(fptr, "<%s>#[%d] ", upperstr, pos);
	    pos++;
	 }
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       fprintf(fptr, "%s", qptr->symbol);
	       if ((*qptr->occurrence != '_') && (*qptr->occurrence != '^'))
	       {
		  fprintf(fptr, "%s", qptr->occurrence);
	       }
	       fprintf(fptr, "#[%d] ", pos);
	       pos++;
	    }
	 }
	 
	 if (GetSymbolType(nptr->symbol) == Part)
	 {
	    fprintf(fptr, "</%s>#[%d] ", upperstr, pos);
	    pos++;
	 }

	 /*
	  * Output the default permutation within "()".
	  */
	 pos = 1;
	 fprintf(fptr, "\n\t(");
	 if (GetSymbolType(nptr->symbol) == Part)
	 {
	    fprintf(fptr, "<%s>#[%d] ", upperstr, pos);
	    pos++;
	 }
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       fprintf(fptr, "%s", qptr->symbol);
	       if ((*qptr->occurrence != '_') && (*qptr->occurrence != '^'))
	       {
		  fprintf(fptr, "%s", qptr->occurrence);
	       }
	       fprintf(fptr, "#[%d] ", pos);
	       pos++;
	    }
	 }
	 
	 if (GetSymbolType(nptr->symbol) == Part)
	 {
	    fprintf(fptr, "</%s>#[%d] ", upperstr, pos);
	    pos++;
	 }
	 fprintf(fptr, ")\n");
      }

      /*
       * Output the end of a symbol and its definitions
       */
      fprintf(fptr, "\t;\n\n");

      /*
       * For each of its definitions and their nodes expand them
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    ICA_Gen(qptr, fptr);
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		ICA_RTT

   Synopsis: 
       Generate an alphabetized list of tokens with a special marker for the 
       Root symbol for teh Replace Tag Tool.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON
       910418   cso      Because the tags in the grammar are
                         uppercase, the tag list here needs to be
			 uppercase as well.        

   Globals used:
       Defndx      To limit access
       Defsymlist  Access an already alphabetized list of symbols
       Info        Need access to the symbol table

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
ICA_RTT(fptr)

FILE
   *fptr; /* File being output */

{
   int
      i, /* Loop counter */
      j; /* Index into the symbol table for the next symbol
	    to be output */

   /*
    * Loop through the alphabetic list of symbols.
    */
   for (i = 0; i < Defndx; i++)
   {
      /*
       * Get the index of each alphabetic symbol in the symbol table
       * If that index is the Root symbol output a marker.
       * If the symbol would require a tag in SGML, i.e., a
       * Part (not Group) type and a nonterminal, output it in uppercase.
       */
      j = GetIndex(Defsymlist[i]);
      if (j == 0)
      {
	 fprintf(fptr, "#");
      }
      if ((Info[j].symtype == Part) && (Info[j].terminal == TriF))
      {
	 fprintf(fptr, "%s\n", MakeUpper(Info[j].symbol));
      }
   }
}




/*
   FUNCTION: 		PrintGroupParents

   Synopsis: 
       Print the locations where a group symbol is used.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
PrintGroupParents(fptr, nptr)

FILE
   *fptr;
Node
   *nptr;

{
   int
      i;   
   Node
      *qptr;
   NodeList
      *nlptr;

   i = GetIndex(nptr->symbol);
   for (nlptr = Info[i].definition; nlptr != NULL; nlptr = nlptr->next)
   {
      qptr = nlptr->nodeptr;
      if (qptr != NULL)
      {
	 if (qptr->parent != NULL)
	 {
	    qptr = qptr->parent;	       
	    /*
	     * If the parent node is not NULL
	     * Output that immediate parent
	     */
	    if (GetSymbolType(qptr->symbol) == Group)
	    {
	       PrintGroupParents(fptr, qptr);
	    }
	    else
	    {
	       fprintf(fptr, "\t<%s>", MakeUpper(qptr->symbol));
	    }
	 }
      }
   }
   return;
}



/*
   FUNCTION: 		DetermineContentType

   Synopsis: 
       Determine the type of content that a tag can have.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
DetermineContentType(nptr, dataflag, tagsflag)

Node
   *nptr;
Boolean
   *dataflag,
   *tagsflag;

{
   Alt
      *aptr;
   int
      i;
   Node
      *qptr;
   
   if (nptr == NULL)
   {
      return;
   }
   for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
   {
      for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
      {
	 i = GetIndex(qptr->symbol);
	 if (Info[i].terminal == TriT)
	 {
	    *dataflag = True;
	 }
	 else if (Info[i].symtype == Part)
	 {
	    *tagsflag = True;
	 }
	 else if (Info[i].symtype == Group)
	 {
	    DetermineContentType(GetDefinition(qptr->symbol),
				 dataflag, tagsflag);
	 }
	 
	 if ((*dataflag == True) && (*tagsflag == True))
	 {
	    return;
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		PrintContent

   Synopsis: 
       Print the type of content this symbol can have.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
PrintContent(fptr, indx)

FILE
   *fptr;
int
   indx;

{
   Boolean
      dataflag,
      tagsflag;
   
   if ((Info[indx].definition != NULL) &&
       (Info[indx].definition->nodeptr != NULL))
   {
      dataflag = False;
      tagsflag = False;

      DetermineContentType(Info[indx].definition->nodeptr,
			   &dataflag, &tagsflag);

      if ((dataflag == True) && (tagsflag == True))
      {
	 fprintf(fptr, ":MIXED:\t");
      }
      else if (tagsflag == True)
      {
	 fprintf(fptr, ":TAGS:\t");
      }
      else
      {
	 fprintf(fptr, ":DATA:\t");
      }
   }
   return;
}




/*
   FUNCTION: 		ICA_ITT

   Synopsis: 
       Insert Tag Tool.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON
       910418   cso      Because the tags in the grammar are
                          uppercase, the tag list here needs to be
			  uppercase as well.        
       910523   cso      Output a complete ITT spec.
       910628   cso      Added printgroupparents to make sure the ITT
                          spec was complete.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
ICA_ITT(fptr)

FILE
   *fptr; /* File being output */

{
   int
      i;      /* Loop counter for accessing symbol table */
   NodeList
      *nlptr; /* Loop pointer to visit each occurrence of a symbol 
		 in a production */
   Node
      *nptr;  /* Pointer to trace back and find the immediate 
		 nonGROUP parent of a node */

   /*
    * For each symbol in the symbol table write out its dependency list
    */
   for (i = 0; i < Ndx; i++)
   {
      /*
       * If the symbol is a nonterminal nonGROUP
       * output its dependency list
       */
      if ((Info[i].terminal == TriF) && (Info[i].symtype == Part))
      {
	 /*
	  * If it is the root symbol just output it
	  * Else output the symbol and "IN" separator
	  * For each definition that uses this symbol,
	  * output its immediate nonGROUP parent
	  * Output a linefeed to separate the symbols
	  */
	 fprintf(fptr, "<%s>\t</%s>\t", MakeUpper(Info[i].symbol),
		 MakeUpper(Info[i].symbol));
	 PrintContent(fptr, i);
	 fprintf(fptr, ":IN:");
	 if (i == 0)
	 {
	    fprintf(fptr, "\t:TOPLEVEL:");
	 }
	 for (nlptr = Info[i].definition; nlptr != NULL; nlptr = nlptr->next)
	 {
	    /*
	     * Access the node in a production, if it isn't NULL
	     * Find its immediate nonGROUP parent and output it
	     */
	    nptr = nlptr->nodeptr;
	    if (nptr != NULL)
	    {
	       if (nptr->parent != NULL)
	       {
		  nptr = nptr->parent;	       
		  /*
		   * If the parent node is not NULL
		   * Output that immediate parent
		   */
		  if (GetSymbolType(nptr->symbol) == Group)
		  {
		     PrintGroupParents(fptr, nptr);
		  }
		  else
		  {
		     fprintf(fptr, "\t<%s>", MakeUpper(nptr->symbol));
		  }
	       }
	    }
	 }
	 fprintf(fptr, "\n");
      }
   }
}




/*
   FUNCTION: 		YaccPre

   Synopsis: 
       Generate the Header info for Yacc files.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Globals used:
       Info    Need to look at the symbol table to find the terminal symbols,
               and the start symbol
       Ndx     The number of symbols in the symbol table

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
YaccPre(fptr)

FILE
   *fptr;  /* The file to output into */

{
   int
      i;   /* Loop counter to check each symbol in the symbol table. */

   /*
    * Generate the tokens/terminal symbols
    */
   fprintf(fptr,"%%token\n");
   for (i = 0; i < Ndx; i++)
   {
      if (Info[i].terminal == TriT)
      {
	 fprintf(fptr,"\t%s\n", Info[i].symbol);
      }
   }

   /*
    * Generate the start symbol.
    */
   fprintf(fptr, "\n%%start\t%s\n\n", Info[0].symbol);

   /*
    * Generate the production delimiter.
    */
   fprintf(fptr, "%%%%\n\n");
   return;
}



/*
   FUNCTION: 		YaccBNF

   Synopsis: 
       This generates the extra productions to represent
       the different occurrences (*, +, ?).  It also outputs the
       token list and the start symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added extra linefeeds in the extra productions.

   Globals used:
       Info    Used to access all symbols requiring additional prods,
       Ndx     To limit the symbols being checked

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
YaccBNF(fptr)

FILE
   *fptr;        /* The file with all of the productions generated
		    by YaccGen */

{
   int
      i;         /* Loop counter through the symbol table */

   /*
    * Generate the extra productions for (*, +, ?)
    */
   for (i = 0; i < Ndx; i++)
   {
      /*
       * Info[].optional, .star, and .plus are the number of uses of those
       * occurrences for a particular symbol
       */
      if (Info[i].optional > 0)
      {
	 fprintf(fptr, "%sOPTL\t: %s\n\t{\n\n\t}\n",
		 Info[i].symbol, Info[i].symbol);
	 fprintf(fptr, "\t| /* empty */\n\t{\n\n\t}\n\t;\n\n");
      }
      if (Info[i].star     > 0)
      {
	 fprintf(fptr, "%sSTAR\t: %sSTAR %s\n\t{\n\n\t}\n", 
		 Info[i].symbol, Info[i].symbol, Info[i].symbol);
	 fprintf(fptr, "\t| /* empty */\n\t{\n\n\t}\n\t;\n\n");
      }
      if (Info[i].plus     > 0)
      {
	 fprintf(fptr, "%sPLUS\t: %sPLUS %s\n\t{\n\n\t}\n",
		 Info[i].symbol, Info[i].symbol, Info[i].symbol);
	 fprintf(fptr, "\t| %s\n\t{\n\n\t}\n\t;\n\n", Info[i].symbol);
      }
   }
   return;
}




/*
   FUNCTION: 		YaccGen

   Synopsis: 
       Generate Yacc notation.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added a test so that NULL symbols are not printed.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
YaccGen(nptr, fptr)

Node
   *nptr;  /* The node under consideration */
FILE
   *fptr;  /* The file being output */

{
   Alt
      *aptr;  /* Loop pointer for each definition of nptr */
   Node
      *qptr;  /* Loop pointer for each node in a definition */
   
   /*
    * If a node has a definition, generate the yacc notation for it
    */
   if (nptr->alts != NULL)
   {
      /*
       * Output the lefthand side symbol
       */
      fprintf(fptr, "%s\t: ", nptr->symbol);

      /*
       * For each definition generate the notation
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 /*
	  * If this is not the first definition, output an OR symbol
	  */
	 if (IsFirstAlt(aptr) == False)
	 {
	    fprintf(fptr, "\n\t| ");
	 }

	 /*
	  * For each symbol in a definition output it and its occurrence
	  * representation, in this case OPTL, STAR, PLUS
	  */
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       fprintf(fptr, "%s", qptr->symbol);
	       PrintOccur(fptr, *qptr->occurrence,
			  " ", " ", "OPTL ", "STAR ", "PLUS ");
	    }
	 }
	 
	 /*
	  * Output the end of a production
	  */
	 fprintf(fptr, "\n\t{\n");
#ifdef AUX

	 /*
	  * Output any auxiliary information for this notation
	  */
	 if (aptr->auxinf != NULL)
	 {
	    /*right now only yacc types exist so...*/
	    if (*aptr->auxinf->info != '\0')
	    {
	       fprintf(fptr, "%s", aptr->auxinf->info);
	    }
	 }
#endif

	 /*
	  * Output the end of the auxiliary information
	  */
	 fprintf(fptr, "\n\t}");
      }

      /*
       * Output the end of all definitions for this node
       */
      fprintf(fptr, "\n\t;\n\n");

      /*
       * For each node in each definition output the notation
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    YaccGen(qptr, fptr);
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		IFYaccPre

   Synopsis: 
       Generate the Header info for Intermediate Form Yacc files.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Globals used:
       Info    Need to look at the symbol table to find the terminal symbols,
               and the start symbol
       Ndx     The number of symbols in the symbol table

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
IFYaccPre(fptr)

FILE
   *fptr;  /* The file to output into */

{
   String
      upstr;
   int
      i;   /* Loop counter to check each symbol in the symbol table. */

   /*
    * Generate the tokens/terminal symbols
    */
	 
   fprintf(fptr, "%%{\n#include <stdio.h>\n#include <ctype.h>\n");
   fprintf(fptr, "%%}\n\n");
   fprintf(fptr,"%%token\n");
   fprintf(fptr,"\tICA_Char_DATA\n\n");
   fprintf(fptr,"%%token\n");
   for (i = 0; i < Ndx; i++)
   {
      if ((Info[i].symtype == Part) && (Info[i].terminal == TriF))
      {
	 upstr = MakeUpper(Info[i].symbol);
	 fprintf(fptr, "\t%sa\n\t%sz\n", upstr, upstr);
	 XtFree(upstr);
      }
   }

/*
   fprintf(fptr,"\n%%type\n");
   for (i = 0; i < Ndx; i++)
   {
      if ((Info[i].symtype == Part) && (Info[i].terminal == TriF))
      {
	 fprintf(fptr, "\t%s", Info[i].symbol);
	 if (Info[i].optional > 0)
	 {
	    fprintf(fptr, "\t%sOPTL", Info[i].symbol);
	 }
	 if (Info[i].star     > 0)
	 {
	    fprintf(fptr, "\t%sSTAR", Info[i].symbol);
	 }
	 if (Info[i].plus     > 0)
	 {
	    fprintf(fptr, "\t%sPLUS", Info[i].symbol);
	 }
	 fprintf(fptr, "\n", Info[i].symbol);
      }
   }
*/
   /*
    * Generate the start symbol.
    */
   fprintf(fptr, "\n%%start\t%s\n\n", Info[0].symbol);

   /*
    * Generate the production delimiter.
    */
   fprintf(fptr, "%%%%\n\n");
   for (i = 0; i < Ndx; i++)
   {
      if ((Info[i].symtype == Part) && (Info[i].terminal == TriT))
      {
	 fprintf(fptr, "%s\t: ICA_Char_DATASTAR\n\t{\n\n\t}\n\t;\n\n",
		 Info[i].symbol);
      }
   }
   return;
}



/*
   FUNCTION: 		IFYaccBNF

   Synopsis: 
       This generates the extra productions to represent
       the different occurrences (*, +, ?).  It also outputs the
       token list and the start symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added extra linefeeds to the extra productions.

   Globals used:
       Info    Used to access all symbols requiring additional prods,
       Ndx     To limit the symbols being checked

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
IFYaccBNF(fptr)

FILE
   *fptr;        /* The file with all of the productions generated
		    by IFYaccGen */

{
   int
      i;         /* Loop counter through the symbol table */

   /*
    * Generate the extra productions for (*, +, ?)
    */
   for (i = 0; i < Ndx; i++)
   {
      /*
       * Info[].optional, .star, and .plus are the number of uses of those
       * occurrences for a particular symbol
       */
      if (Info[i].optional > 0)
      {
	 fprintf(fptr, "%sOPTL\t: %s\n\t{\n\n\t}\n",
		 Info[i].symbol, Info[i].symbol);
	 fprintf(fptr, "\t| /* empty */\n\t{\n\n\t}\n\t;\n\n");
      }
      if (Info[i].star     > 0)
      {
	 fprintf(fptr, "%sSTAR\t: %sSTAR %s\n\t{\n\n\t}\n", 
		 Info[i].symbol, Info[i].symbol, Info[i].symbol);
	 fprintf(fptr, "\t| /* empty */\n\t{\n\n\t}\n\t;\n\n");
      }
      if (Info[i].plus     > 0)
      {
	 fprintf(fptr, "%sPLUS\t: %sPLUS %s\n\t{\n\n\t}\n",
		 Info[i].symbol, Info[i].symbol, Info[i].symbol);
	 fprintf(fptr, "\t| %s\n\t{\n\n\t}\n\t;\n\n", Info[i].symbol);
      }
   }
   fprintf(fptr, "ICA_Char_DATASTAR\t: ICA_Char_DATASTAR ICA_Char_DATA\n");
   fprintf(fptr, "\t{\n\n\t}\n\t| /* empty */\n\t{\n\n\t}\n\t;\n\n");
   return;
}




/*
   FUNCTION: 		IFYaccGen

   Synopsis: 
       Generate Intermediate Form Yacc notation.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/05/28   cso    Added a test so that NULL symbols are not printed.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
IFYaccGen(nptr, fptr)

Node
   *nptr;  /* The node under consideration */
FILE
   *fptr;  /* The file being output */

{
   Alt
      *aptr;  /* Loop pointer for each definition of nptr */
   String
      upstr;
   Node
      *qptr;  /* Loop pointer for each node in a definition */
   
   /*
    * If a node has a definition, generate the yacc notation for it
    */
   if (nptr->alts != NULL)
   {
      /*
       * Output the lefthand side symbol
       */
      fprintf(fptr, "%s\t: ", nptr->symbol);

      /*
       * For each definition generate the notation
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 /*
	  * If this is not the first definition, output an OR symbol
	  */
	 if (IsFirstAlt(aptr) == False)
	 {
	    fprintf(fptr, "\n\t| ");
	 }

	 upstr = MakeUpper(nptr->symbol);
	 
	 if ((GetSymbolType(nptr->symbol) == Part) && 
	     (IsTerminal(nptr->symbol) == TriF))
	 {
	    fprintf(fptr, "%sa ", upstr);
	 }
	 
	 /*
	  * For each symbol in a definition output it and its occurrence
	  * representation, in this case OPTL, STAR, PLUS
	  */
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    if (*qptr->symbol != '\0')
	    {
	       fprintf(fptr, "%s", qptr->symbol);
	       PrintOccur(fptr, *qptr->occurrence,
			  " ", " ", "OPTL ", "STAR ", "PLUS ");
	    }
	 }
	 
	 if ((GetSymbolType(nptr->symbol) == Part) && 
	     (IsTerminal(nptr->symbol) == TriF))
	 {
	    fprintf(fptr, "%sz ", upstr);
	 }
	 XtFree(upstr);
	 
	 /*
	  * Output the end of a production
	  */
	 fprintf(fptr, "\n\t{\n");
#ifdef AUX

	 /*
	  * Output any auxiliary information for this notation
	  */
	 if (aptr->auxinf != NULL)
	 {
	    /*right now only yacc types exist so...*/
	    if (*aptr->auxinf->info != '\0')
	    {
	       fprintf(fptr, "%s", aptr->auxinf->info);
	    }
	 }
#endif

	 /*
	  * Output the end of the auxiliary information
	  */
	 fprintf(fptr, "\n\t}");
      }

      /*
       * Output the end of all definitions for this node
       */
      fprintf(fptr, "\n\t;\n\n");

      /*
       * For each node in each definition output the notation
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    IFYaccGen(qptr, fptr);
	 }
      }
   }
   return;
}
