/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: get.c,v $
   Purpose:     
       To provide routines that Get information from the data structures
       or widgets.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.6 $
   Date:        $Date: 92/03/18 14:15:45 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   EXPORT  GetOddChars

   EXPORT  GetMaxNameLength

   EXPORT  GetViewFilename

   EXPORT  GetLabel

   EXPORT  GetStringWidthInPixels

   EXPORT  GetWidgetHeightOffset

   EXPORT  GetIndex

   EXPORT  GetDefinition
                           GetIndex
   EXPORT  GetDimensions

   EXPORT  GetSymbolType
                           GetIndex
   EXPORT  GetNumProds

*/

/* HISTORY
   $Log:	get.c,v $
 * Revision 1.6  92/03/18  14:15:45  ware
 * Working on consistent interface.
 * 
 * Revision 1.5  1992/01/14  17:56:36  cso
 *  - Reorganizing header files.
 *
 * Revision 1.4  1992/01/03  21:30:37  cso
 * GetSymbolType:
 * 	Expanded the return to check for the Ndx limit and to make
 * sure to return a symbol type of Unknown.
 *
 * Revision 1.3  1991/11/20  00:03:25  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.2  1991/10/17  18:32:29  cso
 * Because SGML uses the entity symbols % and & included the string
 * length, we need to check if the symbol is a Group (SGML = Entity),
 * then increment its length by 1.
 *
 * Revision 1.1  1991/10/04  19:05:54  jbm
 * Initial revision
 *
 * Revision 2.1  91/05/28  10:02:21  cso
 * Had a problem occasionally with specs being generated with the NMCHARS b
 * being some garbage string and I think the problem was that the oddstr was
 * not initialized with a \0 in the first position to get things going.
 * 
 * Revision 2.0  91/03/30  13:48:37  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.4  91/02/26  15:24:37  cso
 * Changed the default values for a NULL tree widget.
 * 
 * Revision 1.3  91/02/22  09:39:14  cso
 * Completed documentation.
 * Added GetOddChars and GetMaxNameLength for sgml.c
 * 
 * Revision 1.2  90/11/29  14:24:57  cso
 * Added GetDimensions from display.c
 * 
 * Revision 1.1  90/11/26  13:13:56  cso
 * Initial revision
 * 
*/

/* #INCLUDES */

#include "std.h"
#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "main.h"



/*
   FUNCTION: 		GetOddChars

   Synopsis: 
       Get and return any non-alphanumeric characters occurring in
       symbol names.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON
       91/05/24 cso      Initialized the string being created to "".

   Globals used:
       Info    To access the symbol table
       Ndx     To limit access

   Globals Modified:
       none

   Return value:	String

   Range:		"" or more
*/

String
GetOddChars()

{
   int
      i,       /* Loop counter */
      j,       /* Loop counter */
      len,     /* The length of the symbol */
      oddndx;  /* Index into oddstr */
   String
      oddstr,  /* The string containing the odd chars */
      str;     /* A pointer to the symbol to make coding more readable */

   /*
    * Allocate space for the characters, at most only 256 - ASCII set
    * Initialize index into oddstr.  oddndx points to the next available 
    * spot for an odd character in oddstr.
    */
   oddstr = XtMalloc(256);
   oddndx = 0;
   oddstr[oddndx] = '\0';

   /*
    * For each symbol in the symbol table, check it for odd characters
    */
   for (i = 0; i < Ndx; i++)
   {
      /*
       * Assign str to point to the symbol to make the code easier to read;
       * one could easily have replaced str[j] with Info[i].symbol[j], but 
       * I thought str[j] would be easier to understand.
       * Get the length of the string, so there is only one call to strlen.
       */
      str = Info[i].symbol;
      len = strlen(str);

      /*
       * For each character in the symbol, check for oddity
       */
      for (j = 0; j < len; j++)
      {
	 /*
	  * If the character is punctuation (non-alphanumeric), but not
	  * a space, tab, linefeed, etc. then it is odd, so place that 
	  * character in oddstr.
	  */
	 if ((ispunct(str[j]) != 0) && (isspace(str[j]) == 0))
	 {
	    /*
	     * If the character exists in oddstr, strchr returns non-NULL,
	     * so if it is NULL then add that character to oddstr
	     */
	    if (strchr(oddstr, str[j]) == NULL)
	    {
	       /*
		* Assign the odd character to the next spot in the character
		* array, increment the index, and put '\0' in that spot 
		* so that oddstr is always NULL-terminated
		*/
	       oddstr[oddndx] = str[j];
	       oddndx++;
	       oddstr[oddndx] = '\0';
	    }
	 }
      }
   }
   return(oddstr);
}




/*
   FUNCTION: 		GetMaxNameLength

   Synopsis: 
       Find the longest symbol string and return that value.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Globals used:
       Info     To access all of the symbols
       Ndx      To limit the search

   Globals Modified:
       none

   Return value:	int

   Range:		>= 0
*/

int
GetMaxNameLength()

{
   int
      i,   /* Loop counter */
      len, /* Length value of each symbol,
	      Used to reduce the number of strlen calls */
      max; /* Maximum length so far in table, and returned */

   /*
    * Initialize max
    */
   max = 0;

   /*
    * Search symbol table
    * If a symbol's length is greater than max,
    * Update max
    * Return max
    */
   for (i = 0; i < Ndx; i++)
   {
      len = strlen(Info[i].symbol);
      if (Info[i].symtype == Group)
      {
	 len++;
      }
      
      if (len > max)
      {
	 max = len;
      }
   }
   return(max);
}




/*
   FUNCTION: 		GetViewFilename

   Synopsis: 
       Returns the filename contained in a widget.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	String

   Range:		"" or a filename
*/

String
GetViewFilename(w)

Widget
   w;  /* The widget that we are getting the filename from */

{
   Arg
      args[MAXARGS];/* List of arguments */
   int
      cnt;       /* Number of arguments */
   String
      filename;  /* Filename contained in the widget */

   /*
    * If the widget is not NULL,
    * Use GetValues to get the filename
    * Else return ""
    */
   if (w != NULL)
   {
      cnt = 0;
      XtSetArg(args[cnt], XtNstring, &filename);  cnt++;
      XtGetValues(w, args, cnt);
   }
   else
   {
      filename = "";
   }
   return(filename);
}




/*
   FUNCTION: 		GetLabel

   Synopsis: 
       Gets a label out of a widget.
       
   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	String

   Range:	        "" or the label value
*/

String
GetLabel(w)

Widget
   w;  /* Widget to access */

{
   Arg
      args[MAXARGS];  /* Argument for the label */
   int
      cnt;
   String
      str;      /* Label to return */
   
   /*
    * If the widget is not NULL
    * Get the label
    * Else return ""
    */
   if (w != NULL)
   {
      cnt = 0;
      XtSetArg(args[cnt], XtNstring, &str); cnt++;
      XtGetValues(w, args, cnt);
   }
   else
   {
      str = "";
   }
   return(str);
}




/*
   FUNCTION: 		GetStringWidthInPixels

   Synopsis: 
       Using XTextExtents, find the width of a string in pixels, making
       the application font independent.
       
   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Dimension

   Range:		0 <=
*/

Dimension
GetStringWidthInPixels(w, str)

Widget
   w;   /* The widget under consideration */
String
   str; /* The string being measured */

{
   Arg
      args[MAXARGS];   /* List of arguments */
   Dimension
      width;           /* Width value to be returned */
   int
      cnt,             /* Number of arguments */
      direction_hint,  /* Direction of the font */
      ascent,          /* Amount above the baseline used by the font */
      descent;         /* Amount below the baseline used by the font */
   Position
      lmargin,         /* The left margin allowed in the widget */
      rmargin;         /* The right margin allowed in the widget */
   XCharStruct
      overall;         /* Structure describing the font */
   XFontStruct
      *font;           /* The font structure used by the widget */

   /*
    * Get the font and margins of the widget
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNfont,        &font);    cnt++;
   XtSetArg(args[cnt], XtNleftMargin,  &lmargin); cnt++;
   XtSetArg(args[cnt], XtNrightMargin, &rmargin); cnt++;
   XtGetValues(w, args, cnt);

   /*
    * Get some of the dimensions used in the font of the widget
    */
   XTextExtents(font, str, strlen(str), &direction_hint, &ascent, &descent,
		&overall);

   /*
    * The width returned equals the margins of the widget (to make sure
    * we don't cut it off too short) plus the left and right bearings that
    * define the rectangle around the string in that font
    */
   width = lmargin + rmargin + overall.lbearing + overall.rbearing;
#ifdef DEBUG
   fprintf(stderr, "Width in pixels of [%s] is [%d]\n", str, width);
#endif
   return(width);
}




/*
   FUNCTION: 		GetWidgetHeightOffset

   Synopsis: 
       Determine the height of widgets using font and form widget 
       information.  Used by the compass to place the west widget.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	int

   Range:		0 <=
*/

int
GetWidgetHeightOffset(w)

Widget
   w;   /* A widget in a form widget */

{
   Arg
      args[MAXARGS]; /* List of arguments */
   Dimension
      bwidth,    /* Border width of the widget */
      height;    /* Height of the widget */
   int
      cnt,       /* Number of arguments */
      defdist,   /* Default distance between widgets in the form widget */
      hoffset,   /* Height offset to be returned */
      vdist;     /* Vertical distance between widgets */
   
   /*
    * Get the border width, height and vertical distance of the widget
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNborderWidth,     &bwidth);  cnt++;
   XtSetArg(args[cnt], XtNheight,          &height);  cnt++;
   XtSetArg(args[cnt], XtNvertDistance,    &vdist);   cnt++;
   XtGetValues(w, args, cnt);

   /*
    * Get the default distance used by the form widget which may or
    * may not be equal to the vertical distance
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNdefaultDistance, &defdist); cnt++;
   XtGetValues(XtParent(w), args, cnt);

   /*
    * The height offset from the top of the form widget to the widget equals
    * twice the borderwidth of widgets + height + vertical distance + 
    * default distance
    *   ---------------------
    *   |  ^      form widget
    *   |                [north widget]
    *   |
    *   |  ^[west widget]
    *
    * The distance between the ^'s is the height offset.  Essentially, it is 
    * as if a widget existed between the widget and the top of the form widget
    */
   hoffset = (2 * bwidth) + (2 * (height + vdist)) + defdist;
#ifdef DEBUG
   fprintf(stderr, "Height offset in pixels is [%d]\n", hoffset);
#endif
   return(hoffset);
}




/*
   FUNCTION: 		GetIndex

   Synopsis: 
       Gets the index in the symbol table of a symbol.
       
   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info    The symbol table to search
       Ndx     The limit of the symbol table and the next available spot

   Globals Modified:
       none

   Return value:	int

   Range:		<= Ndx
*/

int
GetIndex(str)

String
   str;  /* The symbol being looked up in the symbol table */

{
   int
      i;  /* Loop counter */

   /*
    * This function searches the Info[] GrammarData structure
    * for a name match.    If it finds a match then it returns the index.
    * If not, then it returns the index of the next empty position.
    */
   for (i = 0; i < Ndx; i++)
   {
      if (strcmp(str, Info[i].symbol) == 0)
      {
	 return(i);
      }
   }
   return(Ndx);
}




/*
   FUNCTION: 		GetDefinition

   Synopsis: 
       It returns a pointer to the position in the tree where a symbol's
       definition exists.  A symbol's definition only exists once in the
       tree and that pointer is returned.
       
   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		NULL or valid
*/

Node
*GetDefinition(str)

String
   str;  /* Symbol whose definition is being searched */

{
   int
      i; /* Loop counter */

   /*
    * Get the index of the symbol's location in the symbol table
    */
   i = GetIndex(str);

   /*
    * If a definition exists, return it
    * else NULL is returned
    */
   if (Info[i].definition != NULL)
   {
      return(Info[i].definition->nodeptr);
   }
   return(NULL);
}

   


/*
   FUNCTION: 		GetDimensions

   Synopsis: 
       Get the dimensions of a tree widget for positioning.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void, but modifying all dimension parameters

   Range:		none
*/

void
GetDimensions(treew, childflag, x, y, width, height, hdist, vdist)

Widget
   treew;     /* The widget we are finding the dimension of */
Boolean
   childflag; /* Used for returning defaults in the case of an initial
		 node being placed */
Position
   *x,        /* X coordinate */
   *y;        /* Y coordinate */
Dimension
   *width,    /* Width of the widget */
   *height;   /* Height of the widget */
int
   *hdist,    /* Horizontal distance of the widget from its neighbor */
   *vdist;    /* Vertical distance of the widget from its neighbor */

{
   Arg
      args[MAXARGS]; /* List of widgets */
   int
      cnt;      /* Number of widgets */

   /*
    * If the widget exists, get its dimensions
    * Else if defaults are needed, set them as such
    */
   if (treew != NULL)
   {
      cnt = 0;
      XtSetArg(args[cnt], XtNx,             x);      cnt++;
      XtSetArg(args[cnt], XtNy,             y);      cnt++;
      XtSetArg(args[cnt], XtNwidth,         width);  cnt++;
      XtSetArg(args[cnt], XtNheight,        height); cnt++;
      XtSetArg(args[cnt], XtNhorizDistance, hdist);  cnt++;
      XtSetArg(args[cnt], XtNvertDistance,  vdist);  cnt++;
      XtGetValues(treew, args, cnt);
   }
   else if (childflag == True)
   {
      *x      = DEFAULTSPACE;
      *y      = DEFAULTSPACE;
      *width  = 0;
      *height = 0;
      *hdist  = 0;
      *vdist  = 0;
   }
   return;
}




/*
   FUNCTION: 		GetSymbolType

   Synopsis: 
       Return the type of a symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	SymType

   Range:		Unknown, Part, Group, File
*/

SymType
GetSymbolType(str)

String
   str;  /* The symbol being checked */

{
   int
      i;
   
   /*
    * Get the symbol's index and return the symbol type
    */
   i = GetIndex(str);
   if (i == Ndx)
   {
      return(Unknown);
   }
   else
   {
      return(Info[i].symtype);
   }
}



/*
   FUNCTION: 		GetNumProds

   Synopsis: 
       Count the number of symbols defined.
       
   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Ndx     Limit the count for number of productions

   Globals Modified:
       none

   Return value:	int

   Range:	        0 <=
*/

int
GetNumProds()

{
   int
      i,   /* Loop counter */
      cnt; /* Counter for number of symbols being defined */
   
   /*
    * Initialize counter
    */
   cnt = 0;

   /*
    * Check symbol table for nonterminal symbols that have been defined,
    * for each that meet that criteria, increment counter
    */
   for (i = 0; i < Ndx; i++)
   {
      if ((Info[i].terminal != TriT) && (Info[i].defined == True))
      {
	 cnt++;
      }
   }
   return(cnt);
}
