/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: grammenu.c,v $
   Purpose:     
       Create the menu, and provide the functions for manipulating
       the grammar, e.g., deleting, defining, inserting, etc.

   Programmer:  Con O'Connell
   Reviewer:    
   Revision:    $Revision: 1.10 $
   Date:        $Date: 92/03/18 14:15:53 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  DefineSym
                                           GetDefinition
					   IsTerminal
					   MakeSubtree
					   PutDefined
					   UpdateInterface
   HIDDEN  CreateAlt
                                           FindAlt
					   NewAlt
					   UpdateInterface
   HIDDEN  InsBefSym
                                           FindAlt
					   IsFirstNode
					   IsRoot
					   NewNode
					   UpdateInterface
   HIDDEN  InsAftSym
                                           IsLastNode
					   IsRoot
					   NewNode
					   UpdateInterface
   HIDDEN  CheckTerm
                                           CheckTerminates
					   InitTerminate
   HIDDEN  RemoveNulls
                                           RemoveNullSymbols
					   UpdateInterface
   HIDDEN  DeleteSym
                                           Alert
					   HideDisplayTree
					   Initialize
					   PrettyPrintTree
					   RemoveNode
					   RemoveSymbol
					   UpdateInterface
   HIDDEN  EraseAlt
                                           Alert
					   FindAlt
					   HideDisplayTree
					   IsRoot
					   PrettyPrintTree
					   RemoveNode
					   RemoveSymbol
					   UpdateInterface
   EXPORT  CreateGrammarMenu
*/

/* HISTORY
   $Log:	grammenu.c,v $
 * Revision 1.10  92/03/18  14:15:53  ware
 * Working on consistent interface.
 * 
 * Revision 1.9  1992/01/28  15:36:22  cso
 * Delaying the form layout until after prettyprinting
 *
 * Revision 1.8  1992/01/14  20:11:15  cso
 *  - Reorganizing header files.
 *  - Fixed a problem with RemoveNullSymbols.  It was not updating the
 * pointer to the current focus properly and the display was getting
 * messed up.
 *  - Made sure that after deleting a symbol that the interface gets
 * updated correctly.
 *
 * Revision 1.7  1992/01/03  21:26:20  cso
 * Using the message box widgets.
 *
 * Revision 1.6  1991/11/22  23:35:11  cso
 * Finally, have the display working as expected, perhaps a bit quicker,
 * but there is still a small amount of flicker on the redraws.  The
 * fixes included some small problems when removing empty place holders
 * and complete alternatives.
 *
 * Revision 1.5  1991/11/20  00:03:33  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.4  1991/11/07  12:05:35  cso
 * Fixed Bad Window error!!!!  Combination of two functions causing the
 * problem.  In HighlightDefinedStatus, I was always managing the tree
 * widget after modifying its border width - now I check to see if it was
 * already managed before managing it after the change.  In DefineSym,
 * the call to MakeSubtree for the symbol being defined was creating a
 * new alt with exposed set to true and the Tree ptr being bumped down to
 * that alt, but the interface hadn't been updated to get to the correct
 * plane yet - now if the planes of the tree aren't the same, I hide the
 * old plane first update the oldptr and then bump Tree ptr along.
 *
 * Revision 1.3  1991/11/06  16:11:55  cso
 * Initialized uninitialized variables found by QA_C...
 *
 * Revision 1.2  1991/10/17  18:34:07  cso
 * Added the ifdef around Push.
 *
 * Revision 1.1.1.3  1991/10/04  20:16:00  jbm
 * First HaL mods to ICA
 *
 * Revision 2.1  91/05/28  10:04:07  cso
 * Removed Chosen_Symbol.
 * Changed the menu selections to read with Alternative instead of Definition.
 * 
 * Revision 2.0  91/03/30  13:48:51  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.6  91/02/22  09:42:10  cso
 * Completed documentation.
 * Fixed one problem with segmentation faults during deletes.  Other ones
 * are occurring however.
 * 
 * Revision 1.5  90/11/26  13:14:24  cso
 * Cosmetic and lint modifications.
 * 
 * Revision 1.4  90/11/20  15:27:54  cso
 * Added a RemoveNulls function to get rid of the extra Null symbols.
 * Added A Check Termination function for the user also.`
 * Modified DeleteSym and Erase Alts so they work!
 * 
 * Revision 1.3  90/11/05  18:16:57  cso
 * Added function template to all functions, but still need to fill it in.
 * Else, cosmetics.  I still need to rethink Delete and Erase.  Also, need
 * to add blanks in menu to break up functionality.
 * 
 * Revision 1.2  90/10/29  10:21:37  cso
 * Added the appropriate decls to make it work.  Also, modified the ordering
 * of the callbacks to associate them with their respective menu choices.
 * This file is a result of combining two menus into one.
 * 
 * Revision 1.1  90/10/25  13:34:46  cso
 * Initial revision
 * 
*/

/* #INCLUDES */

#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "display.h"
#include "get.h"
#include "main.h"
#include "node.h"
#include "put.h"
#include "remove.h"
#include "status.h"
#include "symtab.h"
#include "utils.h"



/*
   FUNCTION: 		DefineSym

   Synopsis: 
       This function moves the user to the first node of the 
       first definition of the node currently in focus.  If a definition 
       does not exist, one is created.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       WHO       DATE    REASON
       cso	91-05-23 Removed Chosen_Symbol
       cso      91-11-07 Fixed Bad Window error:
                          During MakeSubtree, the new alt is returned
			  with its exposed flag already true, so I
			  need to update the tree (hiding/exposing)
			  before making the subtree and bumping the
			  Tree pointer down its definition.		  

   Globals used:
       Chosennode     Check to see if we know which node to define
       Tree            Used to move with the user

   Globals Modified:
       Tree            Will point to a definition of the node

   Return value:	void

   Range:		none
*/

static void
DefineSym(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Node
      *oldptr;  /* Place holder for updating the tree */

   /*
    * If the user hasn't chosen a node to define, inform the user to do so
    * Else if the user is trying to define a terminal symbol, don't 
    * let that happen either
    * Else define the node
    */
   if (Chosennode == NULL)
   {
      XicaStatus(Statw, "Please select a non-empty symbol first.\n");
      XtSetSensitive(XtNameToWidget(Toplevel, "*compassframe*south"), False);
   }
   else if (IsTerminal(Chosennode->symbol) == TriT)
   {
      XicaStatus(Statw, "Sorry can not define a terminal symbol\n");
   }
   else
   {
      /*
       * Remember what the current focus is,
       * Mark the symbol as defined
       * Get the location where the definition (will) exists
       * If the symbol being defined is not the same as the old position
       * * Hide the subtree of the symbol where this function was called
       * * Update the tree to the symbol being defined position
       * * Update the old pointer to be the symbol being defined
       * Make a subtree (the function will check if it is necessary it
       *  better be!)
       * Move the Tree pointer to the first definition
       * Update the interface with the new focus
       */
#ifdef PATH
      Push(Tree);
#endif
      oldptr = Tree;
      PutDefined(Chosennode, True);
      Tree = GetDefinition(Chosennode->symbol);
      if (oldptr != Tree)
      {
	 HideSubtree(oldptr);
	 UpdateTree(oldptr, Tree);
	 oldptr = Tree;
      }
      MakeSubtree(Tree);
      Tree = Tree->alts->prod;
      UpdateInterface(oldptr, Tree);
   }
   return;
}




/*
   FUNCTION: 		CreateAlt

   Synopsis: 
       Create an alternative definition for a symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Tree    Used to know what the current focus is

   Globals Modified:
       Tree    Moved to a new focus

   Return value:	void

   Range:		none
*/

static void
CreateAlt(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Alt
      *aptr;   /* Loop pointer to find the location where to add the new
		  alternative definition */
   Node
      *oldptr, /* Remembers the current focus */
      *qptr;   /* Loop pointer to hide each of the nodes in the currently
		  exposed definition */

   /*
    * If the current focus is the Root symbol a new definition cannot be
    * created
    * Else create a new alternative definition
    */
   if (Tree->parent == NULL)
   {
      XicaStatus(Statw, "There can only be one root symbol\n");
   }
   else
   {
      /*
       * Remember the current focus
       * Position the alternative pointer
       */
      oldptr = Tree;
      aptr = FindAlt(Tree);
      if (aptr != NULL)
      {
	 /*
	  * If an alternative pointer exists,
	  * Set the exposed flag for the present definition False
	  * Hide each of the nodes in that definition
	  * Wipe the tree window clean of lines
	  * Posistion it at the end of definitions
	  * Create a new definition
	  * Set the new definition's exposed flag to True
	  * Move the Tree pointer to the first node in the new definition
	  * Update the interface
	  */
	 aptr->exposed = False;
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    HideDisplayTree(qptr);
	 }
	 WipeWindow();
	 for ( ; aptr->next != NULL; aptr = aptr->next)
	 {
	 }
	 aptr->next = NewAlt(aptr->whom, aptr);
	 aptr->next->exposed = True;
	 Tree = aptr->next->prod;
	 UpdateInterface(oldptr, Tree);
      }
   }
   return;
}




/*
   FUNCTION: 		InsBefSym

   Synopsis: 
       Insert a new symbol in front of the one currently in focus.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Chosennode   Makes sure we know where we are doing this operation
       Tree          To know the new focus

   Globals Modified:
       Tree          Moved to a new symbol

   Return value:	void

   Range:		none
*/

static void
InsBefSym(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Alt
      *aptr;    /* Pointer to find the definition where the new 
		   symbol will be placed */
   Node
      *oldptr,  /* Remember the current focus */
      *nptr,    /* Points to the symbol we are inserting before */
      *qptr;    /* The new symbol being inserted */

   /*
    * If a node hasn't been chosen to insert before, inform the user
    * Else if it is a root symbol, that isn't allowed either
    * Else do the insertion
    */
   if (Chosennode == NULL)
   {
      XicaStatus(Statw, "Please select a non-empty symbol first.\n");
   }
   else if (IsRoot(Chosennode) == True)
   {
      XicaStatus(Statw, "Cannot have more than one Root symbol.\n");
   }
   else
   {
      /*
       * Use nptr to point at the node being inserted before
       * Remember this node for updating later
       * Create a new node and assign it to qptr
       * Do the insertion
       */
      nptr = Chosennode;
      oldptr = nptr;
      qptr = NewNode(nptr->parent, nptr->prev);

      /*
       * If nptr is not at the front of the production,
       * make the node currently before nptr point to qptr
       * Else find the alternative definition pointer and
       * set it to point at qptr (the new head of the list)
       */
      if (IsFirstNode(nptr) == False)
      {
	 nptr->prev->next = qptr;
      }
      else
      {
	 aptr = FindAlt(nptr);
	 aptr->prod = qptr;
      }

      /*
       * Set nptr to point at qptr previous
       * Set qptr to point at nptr next
       * Set Tree to point to the new node
       * Update the interface
       */
      nptr->prev = qptr;
      qptr->next = nptr;
      Tree = qptr;
      UpdateInterface(oldptr, Tree);
   }
   return;
}




/*
   FUNCTION: 		InsAftSym

   Synopsis: 
       Insert a new symbol behind the one currently in focus.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Chosennode   Makes sure we know where we are doing this operation
       Tree          To know the new focus

   Globals Modified:
       Tree          Moved to a new symbol

   Return value:	void

   Range:		none
*/

static void
InsAftSym(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Node
      *oldptr,  /* Remember the current focus */
      *nptr,    /* Points to the symbol we are inserting after */
      *qptr;    /* The new symbol being inserted */

   /*
    * If a node hasn't been chosen to insert after, inform the user
    * Else if it is a root symbol, that isn't allowed either
    * Else do the insertion
    */
   if (Chosennode == NULL)
   {
      XicaStatus(Statw, "Please select a non-empty symbol first.\n");
   }
   else if (IsRoot(Chosennode) == True)
   {
      XicaStatus(Statw, "Cannot have more than one Root symbol.\n");
   }
   else
   {
      /*
       * Use nptr to point at the node being inserted after
       * Remember this node for updating later
       * Create a new node and assign it to qptr
       * Do the insertion
       */
      nptr = Chosennode;
      oldptr = nptr;
      qptr = NewNode(nptr->parent, nptr);

      /*
       * If nptr is not at the back of the production,
       * make the node currently after nptr point to qptr
       */
      if (IsLastNode(nptr) == False)
      {
	 nptr->next->prev = qptr;
      }

      /*
       * Set qptr point to the node after nptr
       * Set nptr point to qptr after nptr
       * Set Tree to point to the new node
       * Update the interface
       */
      qptr->next = nptr->next;
      nptr->next = qptr;
      Tree = qptr;
      UpdateInterface(oldptr, Tree);
   }
   return;
}




/*
   FUNCTION: 		CheckTerm

   Synopsis: 
       Check whether the grammar terminates or not.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Root     To give an initial node to start checking

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
CheckTerm(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   /*
    * Initialize the symbol table for doing this
    */
   InitTerminate();

   /*
    * Check the termination starting at Root.
    * If it returns False tell the user it doesn't terminate
    * Else tell the user that the grammar does terminate
    */
   if (CheckTerminates(Root) == False)
   {
      XicaStatus(Statw, "WARNING: Grammar does not terminate\n");
   }
   else
   {
      XicaStatus(Statw, "Grammar terminates\n");
   }
   return;
}




/*
   FUNCTION: 		RemoveNulls

   Synopsis: 
       Remove any NULL symbols that the user may have left in the grammar.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Root     To give a starting point from which to remove NULL symbols

   Globals Modified:
       Tree     It points to the Root to let the user know that the function
                has completed

   Return value:	void

   Range:		none
*/

static void
RemoveNulls(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   /*
    * Remove the empty place holders starting at Root
    * Set Tree to Root
    * Update the interface
    */
   if (*Tree->symbol == '\0')
   {
      if ((Tree->prev != NULL) && (*Tree->prev->symbol != '\0'))
      {
	 Tree = Tree->prev;
      }
      else if ((Tree->next != NULL) && (*Tree->next->symbol != '\0'))
      {
	 Tree = Tree->next;
      }
      else if ((Tree->parent != NULL) && (*Tree->parent->symbol != '\0'))
      {
	 Tree = Tree->parent;
      }
      else
      {
	 Tree = Root;
      }
   }
   
   RemoveNullSymbols(Root);
   UpdateInterface(Tree, Tree);
   return;
}




/*
   FUNCTION: 		DeleteSym

   Synopsis: 
       Delete a symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       WHO       DATE    REASON
       cso	91-05-23 Removed Chosen_Symbol

   Globals used:
       Chosennode      The node being deleted
       Root             A reference for prettyprinting
       Tree             The current focus

   Globals Modified:
       Tree             Points to the new focus

   Return value:	void

   Range:		none
*/

static void
DeleteSym(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Node
      *oldptr;  /* Remember the current focus (probably being deleted) */
   String
      str;      /* The prompt to ask the user to confirm */

   /*
    * The user needs to select a symbol to delete, so check for this
    * first.
    */
   if (Chosennode == NULL)
   {
      XicaStatus(Statw, "Please select a non-empty symbol first.\n");
   }
   else
   {
      /*
       * Allocate enough space for the prompt string,
       * Create the prompt string to confirm with the user that
       * this symbol be deleted
       */
      str = (char *) XtMalloc((unsigned) strlen(Chosennode->symbol) + 64);
      sprintf(str, "Delete one instance of [%s]?", Chosennode->symbol);

      /*
       * Alert the user, and wait for a response.
       * If that response is to continue and delete the symbol, do so
       */
      if (YesNoMessage(Frame, str, False) == True)
      {
	 /*
	  * Remember where the current focus is, set Tree to point
	  * to the node being deleted (they should be the same anyway)
	  */
	 oldptr = Tree;
	 Tree   = Chosennode;

	 /*
	  * If the symbol being deleted is the Root symbol, initialize
	  * the interface
	  * Else delete an ordinary symbol
	  */
	 if (Tree == Root)
	 {
	    Initialize(Ndx);
	 }
	 else
	 {
	    /*
	     * Hide the tree displayed from this node down its subtree
	     * Remove the symbol from the symbol table
	     * Remove the symbol from the node data structure
	     * Set the dirty flag
	     */
	    HideDisplayTree(Tree);
	    RemoveSymbol(Tree, True);
	    Tree       = RemoveNode(Tree, True);
	    Dirtyflag = True;
	 }

	 /*
	  * The displayed tree will need to be prettyprinted
	  * and the rest of the interface updated
	  */
	 if (oldptr == Chosennode)
	 {
	    UpdateInterface(Root, Tree);
	 }
	 else
	 {
	    UpdateInterface(oldptr, Tree);
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		EraseAlt

   Synopsis: 
       Erase an entire definition of a symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       02-14-91   cso    Fixed the seg fault occurring when deletions
                         take place.

   Globals used:
       Tree    Used to determine which definition to erase

   Globals Modified:
       Tree    Points to a new location in the tree

   Return value:	void

   Range:		none
*/

static void
EraseAlt(w, client_data, call_data)

Widget
   w;           /* Unused */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Alt
      *aptr;   /* Points to the definition being erased */
   Node
      *oldptr, /* Remembers the current focus of the interface */
      *nptr,   /* Loop pointer to delete each symbol in the definition */
      *qptr;   /* Points to a new valid node that is not deleted */

   /*
    * The user cannot erase the Root, but also inform the user that
    * the user CAN delete the symbol
    * Else if the user has confirmed the action, do the erasure
    */
   if (IsRoot(Tree) == True)
   {
      XicaStatus(Statw, "You can not erase the root symbol.\n");
      XicaStatus(Statw, "You can DELETE the root symbol!\n");
   }
   else if (YesNoMessage(Frame, "Erase this definition?", False) == True)
   {
      /*
       * Remember the current focus
       * Find the definition containing Tree
       */
      oldptr = Tree;
      aptr   = FindAlt(Tree);
      
      /*
       * If a definition exists, in other words, no error,
       * erase it
       */
      if (aptr != NULL)
      {
	 /*
	  * For each symbol in the definition, 
	  * Hide its subtree and remove it.
	  */
	 qptr = NULL;
	 for (nptr = aptr->prod; nptr != NULL; nptr = nptr->next)
	 {
	    /*
	     * 02-14-91 cso
	     * HideDisplayTree(Tree) -> HideDisplayTree(nptr)
	     */
	    HideDisplayTree(nptr);
	    RemoveSymbol(nptr, True);
	    qptr = RemoveNode(nptr, True);
	 }

	 /*
	  * qptr now equals what is valid
	  * after removing all of the symbols (should be the first
	  * node at an equal or higher level of the erased alt that is
	  * now exposed.) and
	  * aptr has already been removed in RemoveNode, so don't expect it 
	  * to be valid.
	  * If qptr is not NULL have Tree point to qptr ELSE to Root
	  */
	 if (qptr != NULL)
	 {
	    Tree = qptr;
	 }
	 else
	 {
	    Tree = Root;
	 }

	 /*
	  * Set the dirty flag
	  */
	 Dirtyflag = True;
      }

      /*
       * Prettyprint the tree
       * Update the interface
       */
      XawFormDoLayout(XtNameToWidget(Toplevel, "*treeframe"),
		      False);
      (void) PrettyPrintTree(Root, (Widget) NULL, True);
      XawFormDoLayout(XtNameToWidget(Toplevel, "*treeframe"),
		      True);
      UpdateInterface(oldptr, Tree);
   }
   return;
}




/*
   FUNCTION: 		CreateGrammarMenu

   Synopsis: 
       Create the menu list of grammar options.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91-05-23   cso    Changed Definition to Alternative in the
                          Create and Erase selections.

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
CreateGrammarMenu(parent)

Widget
   parent;  /* The parent widget of the menu list */

{
   Arg
      args[MAXARGS];/* List of arguments */
   int
      cnt;       /* Number of arguments */
   Widget
      checkterm, /* Check termination option */
      createsel, /* Create definition option */
      definesel, /* Define symbol option */
      delsymsel, /* Delete symbol option */
      erasesel,  /* Erase definition option */
      insaftsel, /* Insert after option */
      insbefsel, /* Insert before option */
      remnulls;  /* Remove Null symbols option */

   /*
    *   -----------------------------
    *   |     parent                |
    *   | [Define Symbol]           |
    *   | [Create Alternative]      |
    *   | [Insert Before Symbol]    |
    *   | [Insert After Symbol]     |
    *   | ------------------------- |
    *   | [Check Termination]       |
    *   | [Remove Null Symbols]     |
    *   | ------------------------- |
    *   | [Delete Symbol]           |
    *   | [Erase  Alternative]      |
    *   -----------------------------
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNlabel,    "Define Symbol"); cnt++;
   definesel = XtCreateManagedWidget("definesel", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(definesel, XtNcallback, DefineSym, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel,    "Create Alternative"); cnt++;
   createsel = XtCreateManagedWidget("createsel", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(createsel, XtNcallback, CreateAlt, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel,    "Insert Before Symbol"); cnt++;
   insbefsel = XtCreateManagedWidget("insbefsel", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(insbefsel, XtNcallback, InsBefSym, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel,    "Insert After Symbol"); cnt++;
   insaftsel = XtCreateManagedWidget("insaftsel", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(insaftsel, XtNcallback, InsAftSym, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlineWidth, 3); cnt++;
   (void) XtCreateManagedWidget("linesep", smeLineObjectClass,
				parent, args, cnt);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel,    "Check  Termination"); cnt++;
   checkterm = XtCreateManagedWidget("checkterm", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(checkterm, XtNcallback, CheckTerm, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlabel,    "Remove Place Holders"); cnt++;
   remnulls = XtCreateManagedWidget("remnulls", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(remnulls, XtNcallback, RemoveNulls, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNlineWidth, 3); cnt++;
   (void) XtCreateManagedWidget("linesep", smeLineObjectClass,
				parent, args, cnt);

   cnt = 0;
   XtSetArg(args[cnt], XtNsensitive, True);            cnt++;
   XtSetArg(args[cnt], XtNlabel,     "Delete Symbol"); cnt++;
   delsymsel = XtCreateManagedWidget("delsymsel", smeBSBObjectClass,
				     parent, args, cnt);
   XtAddCallback(delsymsel, XtNcallback, DeleteSym, NULL);

   cnt = 0;
   XtSetArg(args[cnt], XtNsensitive, True);                 cnt++;
   XtSetArg(args[cnt], XtNlabel,     "Erase  Alternative"); cnt++;
   erasesel = XtCreateManagedWidget("erasesel", smeBSBObjectClass,
				    parent, args, cnt);
   XtAddCallback(erasesel, XtNcallback, EraseAlt, NULL);

   return;
}
