/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: interface.c,v $
   Purpose:
       To call the appropriate create routines and handle the "look"
       of the interface.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.11 $
   Date:        $Date: 92/05/29 14:46:09 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  QuickQuit

   EXPORT  CreateInterface
					   CreateAltDisplay
					   CreateAttribShell
					   CreateAuxShell
					   CreateDefSymbolList
					   CreateFileIOShell
					   CreateMenubar
					   CreateTreeShell
					   CreateUndSymbolList
*/

/* HISTORY
   $Log:	interface.c,v $
 * Revision 1.11  92/05/29  14:46:09  boyd
 * Changes for improved resizing, consisting mostly of replacing Form widgets
 * with Group, Table, and Viewform widgets.
 * 
 * 
 * Revision 1.10  92/03/18  14:16:07  ware
 * Working on consistent interface.
 * 
 * Revision 1.9  1992/02/11  16:24:42  cso
 * Changed the focus feedback to a switch of bg/fg colors and caret display
 *
 * Revision 1.8  1992/02/06  04:42:24  cso
 * Changed the parent of the file dialog for proper positioning
 *
 * Revision 1.7  1992/01/28  15:38:50  cso
 * Added a define action for double clicking in Undefined list
 *
 * Revision 1.6  1992/01/14  20:14:33  cso
 *  - Reorganizing header files.
 *  - Added the keypad accelerator actions for the compass and depth
 * buttons.
 *
 * Revision 1.5  1992/01/03  21:19:48  cso
 * Removed background fallbacks b/c they are dependent on color for now.
 * Moved accelerators for symbol attributes.
 * Standardized the SYNC compile variable.
 * Removed all calls for the alert shell.
 * Cross installed accelerators.
 *
 * Revision 1.4  1991/11/20  00:03:52  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.3  1991/11/19  16:37:40  cso
 * Added keyboard accelerators and panner settings to fallback resources.
 *
 * Revision 1.2  1991/11/18  23:47:18  cso
 * Added the Release 5 panner and porthole widgets to replace the
 * scrollbars on the navigation tree window.
 * They are compile time usable by using -DR5 in the CFLAGS.
 *
 * Revision 1.1.1.3  1991/10/04  20:16:07  jbm
 * First HaL mods to ICA
 *
 * Revision 2.1  91/05/28  10:12:28  cso
 * Changed the order of the creation invocations.
 * 
 * Revision 2.0  91/03/30  13:49:35  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.18  91/03/30  13:29:14  cso
 * Taking advantage of XtInstallAllAccelerators.
 * Added the install routines for the symbol and occurrence types.
 * 
 * Revision 1.17  91/03/28  14:44:42  cso
 * Added the AddActions and pixmap creation to CreateInterface.
 * 
 * Revision 1.16  91/02/22  09:44:45  cso
 * Completed documentation.
 * Changed the name of quit2 to QuickQuit.
 * Installed keyboard accelerators for navigating.
 * 
 * Revision 1.15  90/11/26  13:14:53  cso
 * Cosmetic and lint modifications.
 * 
 * Revision 1.14  90/11/05  18:19:25  cso
 * Added function template to all functions, but still need to fill it in.
 * Only allow QuickQuit if debugging. Else, cosmetics.
 * 
 * Revision 1.13  90/10/29  10:23:30  cso
 * Removed CreateViewShell, and placed the call in semmenu.c
 * 
 * Revision 1.12  90/10/24  15:34:28  cso
 * Added ifdef SYNCED for debugging purposes.
 * 
 * Revision 1.11  90/10/04  15:56:24  cso
 * Cosmetic changes.
 * 
 * Revision 1.10  90/10/02  13:34:55  cso
 * Cosmetic changes.
 * 
 * Revision 1.9  90/09/07  09:39:58  cso
 * Cosmetics.
 * 
 * Revision 1.8  90/09/03  18:24:42  cso
 * Function name changes.
 * 
 * Revision 1.7  90/09/03  14:33:31  cso
 * Changed function names to be more consistent.
 * 
 * Revision 1.6  90/08/29  17:48:10  cso
 * changed create_attdisplay to create_attribshell because of the
 * migration to popups.
 * Also,
 *  rearranged the look of the interface.
 * 
 * Revision 1.5  90/08/26  17:41:22  cso
 * Comment all calls RE: auxiliary information.
 * 
 * Revision 1.4  90/08/26  14:13:19  cso
 * added module template
 * 
*/

/* #INCLUDES */

#include "std.h"
#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "alt.h"
#include "att.h"
#include "compass.h"
#include "deflist.h"
#include "fileio.h"
#include "main.h"
#include "menubar.h"
#include "status.h"
#include "threed.h"
#include "tree.h"
#include "undlist.h"

/*  GLOBALS */

Pixel
   Background,
   Foreground;

static char
   *Fallbacks[] = {
   /************************/
   /***      Panner      ***/
   /************************/
   "*Panner*backgroundStipple:	gray",
   "*treepan*defaultScale:	25",
   "*treepan*internalSpace:	6",
   "*treepan*shadowThickness:	4",
   /************************/
   /***   Accelerators   ***/
   /************************/
   /*       Compass        */
   "*north*accelerators:	<Key>Up:	set() notify() reset()",
   "*south*accelerators:	<Key>Down:	set() notify() reset()",
   "*east*accelerators:		<Key>Right:	set() notify() reset()",
   "*west*accelerators:		<Key>Left:	set() notify() reset()",
   "*prevalt*accelerators:      <Key>R9:	set() notify() reset()",
   "*nextalt*accelerators:	<Key>R15:	set() notify() reset()",
   "*Navigation Tree.title:	Develop Grammar: Navigation Tree",
   "*Navigation Tree.iconName:	Develop Grammar: Navigation Tree",
   "*Semantic View.title:	Develop Grammar: Semantic View",
   "*Semantic View.iconName:	Develop Grammar: Semantic View",
   "*DeveGram.title:		Develop Grammar",
   "*DeveGram.iconName:		Develop Grammar",
   /************************/
   /***     Sizing       ***/
   /************************/
   "*nodeview.prefWidth:        550",
   "*nodeview.minWidth:         100",
   "*listdef.stretch:           True",
   "*listund.stretch:           True",
   "*listviews.stretch:         True",
   "*defview.prefWidth:         300",
   "*defview.minWidth:          100",
   "*defview.prefHeight:        300",
   "*defview.minHeight:         50",
   "*undview.prefWidth:         300",
   "*undview.minWidth:          100",
   "*undview.prefHeight:        300",
   "*undview.minHeight:         50",
   "*Statw.prefHeight:          96",
   "*Statw.minHeight:           20",
   "*Statw.stretch:             False",
   "*statframe.stretch:         False",
   "*treeview.prefWidth:        650",
   "*treeview.minWidth:         100",
   "*treeview.prefHeight:       250",
   "*treeview.minHeight:        50",
   "*treeport.prefWidth:        600",
   "*treeport.minWidth:         100",
   "*treeport.prefHeight:       300",
   "*treeport.minHeight:        50",
   "*semdefs.prefWidth:         450",
   "*semdefs.minWidth:          100",
   "*semdefs.prefHeight:        300",
   "*semdefs.minHeight:         50",
   (char *) NULL
   };




/*
   FUNCTION: 		QuickQuit

   Synopsis: 
       Create a quick quit during development.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   System files modified:
	none

   Return value:	void

   Range:		none
*/

static void
QuickQuit(w, client_data, event)

Widget
   w;           /* Unused */
caddr_t
   client_data; /* Unused */
XEvent
   *event;      /* The event causing this routine to be called */

{
   int
      i;        /* Count of characters being returned */
   char
      text[10]; /* String of characters buffered */
   KeySym
      keys;     /* Key structure */
   
   /*
    * Find the string of characters being typed by the user
    * If the user hit a 'Q' or 'q', exit out of the environment
    */
   i = XLookupString(event, text, 10, &keys, 0);
   if ((i == 1) && ((text[0] == 'Q') || (text[0] == 'q')))
   {
      exit(0);
   }
   return;
}




/*
   FUNCTION: 		CreateInterface

   Synopsis: 
       Call all of the routines to create the complete interface.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/22   cso    Added the creation of Threed Pixmap
       91/03/29   cso    Taking advantage of InstallAllAccelerators
                         Added accelerators to the occurrences and
			  symbol types.
   Globals used:
       Treeshell  To create the global pixmap

   Globals Modified:
       Appcon     It's created
       Frame       It's created
       Toplevel    It's created
       Threed      It's created

   Return value:	void

   Range:		none
*/

void CreateInterface( argc, argv )
int
   argc;   /* The number of variable args being passed in */
char
   **argv; /* The list of string options being passed in */
{
   Display
      *pixdisp; /* Display pointer of a window */
   int
      depth;    /* Depth of the workstation screen for the pixmap */
   Pixel
      background, /* Background pixel value */
      foreground; /* Foreground pixel value */
   Widget
      listframe,
      listlbls,listviews, listdef, listund, 
      treeform;
    
   Window
      pixwin;     /* Drawable for the pixmap */
   static XtActionsRec
      NodeActions[] = {
      {"AltAction",   AltAction},
      {"CompassAction",   CompassAction},
      {"DefineSymAction",   DefineSymAction},
      {"NodeDone",   NodeDone},
      {"NodeSelect", NodeSelect},
      {NULL,         NULL},
   };

   /*
    * Create the toplevel shell and initialize the application
    * context, then add some resources.
    */
   Toplevel = XtVaAppInitialize( 
       &Appcon, TITLE, NULL, 0, &argc, argv, Fallbacks,
       XtNinput,            True,
#ifdef HARDCODEDTITLE
       XtNtitle,            TITLE,
       XtNiconName,         TITLE,
#endif
       XtNallowShellResize, True,
       NULL );

   XtAppAddActions( Appcon, NodeActions, XtNumber(NodeActions) );

#ifdef SYNC
   /*
    * Go into synchronize mode for debugging
    */
   XSynchronize(XtDisplay(Toplevel), True);
#endif

   /*
    * Create the Frame that contains the pieces of the main widget
    */
   Frame = XtVaCreateManagedWidget( "Frame",
       groupWidgetClass, Toplevel,
       XtNspacing, 8,
       NULL );

   XtVaGetValues( Frame,
       XtNborderColor, &Foreground,
       XtNbackground,  &Background,
       NULL );

   /*
    * Create the menubar
    * Create the compass
    * Create the definitions display
    * Create the available symbols and undefined symbols lists
    * Realize the topshell
    * Create a shell for displaying the node tree
    */
   CreateMenubar(Frame);
   CreateCompass(Frame);
   CreateAltDisplay(Frame);

   /*
    * The following sets up the two lists and their labels as a Table
    * with two rols and two columns.  The table is now the parent, and
    * the appropriate col and both row gadgets are passed to Create...List.
    */
   listframe = XtVaCreateManagedWidget( "listframe",
       tableWidgetClass, Frame,
       XtNalignment,   gaFill,
       XtNmargin,      8,
       XtNhorzMargin,  0,
       XtNvertMargin,  0,
       XtNhorzSpacing, 8,
       XtNvertSpacing, 0,
       XtNstretch,     True,
       NULL );
   listdef = XtVaCreateWidget( "listdef",
       tableColObjectClass, listframe,
/*
       XtNstretch, True,
*/
       NULL );
   listund = XtVaCreateWidget( "listund",
       tableColObjectClass, listframe,
/*
       XtNstretch, True,
*/
       NULL );
   listlbls = XtVaCreateWidget( "listlbls",
       tableRowObjectClass, listframe,
       NULL );
   listviews = XtVaCreateWidget( "listviews",
       tableRowObjectClass, listframe,
/*
       XtNstretch, True,
*/
       NULL );

   CreateDefSymbolList( listframe, listdef, listlbls, listviews );
   CreateUndSymbolList( listframe, listund, listlbls, listviews );

   CreateStatus(Frame);

   XtRealizeWidget(Toplevel);

   CreateTreeShell();
   /*
    * Create shells for alerting the user, getting symbol 
    * attributes, and doing file I/O
    */
   CreateAttribShell(Toplevel);

   CreateFileIOShell(Toplevel);

#ifdef AUX
   /*
    * Create a shell to get auxiliary information
    */
   CreateAuxShell();
#endif

   /*
    * Install keyboard accelerators
    */
   treeform = XtNameToWidget(Treeshell, "*treeform");
   
   XtInstallAllAccelerators(Frame,    Frame);
   XtInstallAllAccelerators(Frame,    treeform);
   XtInstallAllAccelerators(treeform, Frame);
   XtInstallAllAccelerators(treeform, treeform);

#ifdef DEBUG
   /*
    * Add the event handler for the quick quit
    */
   XtAddEventHandler(Frame,    KeyPressMask, False, QuickQuit, NULL);
   XtAddEventHandler(treeform, KeyPressMask, False, QuickQuit, NULL);
#endif

#ifdef ThreeD
   /*
    * Create the background pixmap to show which nodes have more than
    * one definition.
    */
   XtVaGetValues( Treeshell,
       XtNbackground,  &background,
       XtNborderColor, &foreground,
       XtNdepth,       &depth,
       NULL );

#ifdef DEBUG
   fprintf(stderr, "Depth of Treeshell = [%d]\n", depth);
#endif

   pixdisp = XtDisplay(Treeshell);
   pixwin  = XtWindow(Treeshell);
   Threed  = XCreatePixmapFromBitmapData(pixdisp, pixwin,
					 threed_bits,
					 threed_width, threed_height,
					 foreground, background, depth);
   
   if (Threed == None)
   {
      Threed = XtUnspecifiedPixmap;
   }
#endif

   return;
}
