/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: loadfile.c,v $
   Purpose:     
       Contains all of the functions relevant to loading a file in the
       format used by DeveGram.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.6 $
   Date:        $Date: 92/03/18 14:16:15 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  StripLoadString

   HIDDEN  ProcessOccur
                                           PutDefined
                                           PutOccur
                                           PutTerminal
   HIDDEN  ProcessNewSymbol
                          ProcessOccur
                                           AddSymbol
					   IsDefined
					   PutDefined
   HIDDEN  ProcessAuxInfo
                                           NewAux
   EXPORT  EditFile
			   ProcessAuxInfo
			   ProcessNewSymbol
                           StripLoadString
					   AddSymbol
					   FindAlt
					   GetDefinition
					   NewAlt
					   NewNode
					   PutDefined
					   PutTerminal

*/

/* HISTORY
   $Log:	loadfile.c,v $
 * Revision 1.6  92/03/18  14:16:15  ware
 * Working on consistent interface.
 * 
 * Revision 1.5  1992/01/14  20:15:43  cso
 *  - Reorganizing header files.
 *  - Disable the symbol lists from updating while a file is loaded -
 * significant speed up.
 *
 * Revision 1.4  1991/11/20  00:04:01  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.3  1991/11/06  16:12:07  cso
 * Initialized uninitialized variables found by QA_C...
 *
 * Revision 1.2  1991/10/17  18:35:19  cso
 * Added a level of debugging so that the loading diagnostics don't
 * interfere.
 *
 * Revision 1.1.1.3  1991/10/04  20:16:11  jbm
 * First HaL mods to ICA
 *
 * Revision 2.0  91/03/30  13:50:02  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.1  91/02/22  09:46:51  cso
 * Initial revision
 * 
*/

/* #INCLUDES */

#include "std.h"
#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "get.h"
#include "main.h"
#include "put.h"
#include "status.h"
#include "symtab.h"
#include "utils.h"



/*
   FUNCTION: 		StripLoadString

   Synopsis: 
       Strips a token during load to significant pieces: 
       symbol type, occurrence, and symbol string.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
StripLoadString(str, type, occur)

String
   *str,  /* The token to be stripped */
   type,  /* The symbol type */
   occur; /* The occurrence */

{

   /*
    * Based on the first character of the token, strip off the pieces.
    * IMPORTANT: A symbol cannot contain any whitespace, else it won't be 
    *            collected as a token.
    */
   switch (**str)
   {
   default:
      /*
       * A token occurring in a production.
       */
      *type  = **str;
      (*str)++;
      *occur = **str;
      (*str)++;
      break;

   case '!':
      /*
       * This marks a symbol about to be defined.
       * In other words a lefthand-side symbol.
       * No type or occurrence involved.
       */
      /* Same as '#' */

   case '#':
      /*
       * This marks the beginning of auxiliary information which is a buffer
       * of characters of a certain length.
       */
      (*str)++;
      *type  = ' ';
      *occur = ' ';
      break;

   case '&':
      /*
       * This marks a definition for a symbol.
       * If no production actually exists, i.e., no symbols in the production,
       * it is a NULL production, so ignore it.
       * Else get the rest of the symbol information for the first symbol 
       * in the production.
       */
      if (strcmp(*str, "&") == 0)
      {
	 (*str)++;
	 *type  = ' ';
	 *occur = ' ';
#ifdef DEBUG
	 fprintf(stderr, "Found NULL alternative production\n");
#endif
      }
      else
      {
	 (*str)++;
	 *type  = **str;
	 (*str)++;
	 *occur = **str;
	 (*str)++;
      }
      break;
   }

   /*
    * The occurrence was unknown when the file was saved out.
    * We should demand the user to tell us the information after
    * loading in the entire file.
    */
   if (*occur == '$')
   {
      *occur = ' ';
      XicaStatus(Statw, "WARNING: Symbol [%s] needs an occurrence\n", *str);
   }
   return;
}




/*
   FUNCTION: 		ProcessOccur

   Synopsis: 
       Process the occurrences loaded.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

static void
ProcessOccur(ptr)

Node
   *ptr;    /* The node under consideration */

{
   String
      str; /* The symbol string */
   
   /*
    * Assign the symbol string to str
    */
   str = XtNewString(ptr->symbol);

   /*
    * Depending on the occurrence character
    */
   switch(ptr->occurrence[0])
   {
   default:
      /*
       * If none of the occurrences matched there must be an error
       */
      XicaStatus(Statw, "ERROR: Unknown occurrence [%c], for token [%s]\n",
	      ptr->occurrence[0], str);
      break;

   case '^':
      /*
       * The occurrence symbol for a terminal symbol
       */
      PutTerminal(ptr, TriT);
      PutDefined(ptr, True);
      break;

   case '_':
      /*
       * An occurrence of exactly once for a nonterminal
       */
      PutTerminal(ptr, TriF);
      break;

   case '?':
      /*
       * An occurrence of zero or one (optional) for a nonterminal
       */
      PutTerminal(ptr, TriF);
      break;

   case '*':
      /*
       * An occurrence of zero or more (star) for a nonterminal
       */
      PutTerminal(ptr, TriF);
      break;

   case '+':
      /*
       * An occurrence of one or more (plus) for a nonterminal
       */
      PutTerminal(ptr, TriF);
      break;
   }
   PutOccur(str, ptr->occurrence[0]);
}




/*
   FUNCTION: 		ProcessNewSymbol

   Synopsis: 
       Based on the type of symbol, process it.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/


static void
ProcessNewSymbol(type_char, occur_char, symbol, nptr)

char
   type_char,  /* The character indicating the symbol type */
   occur_char; /* The character indicating the occurrence */
String
   symbol;     /* The symbol being processed */
Node
   *nptr;      /* The node being changed */

{
   /*
    * Depending on the symbol type, add the symbol to the node tree
    */
   switch (type_char)
   {
   default:
      XicaStatus(Statw, "Syntax error on type char [%c] for symbol [%s]\n",
	      type_char, symbol);
      return;
      break;
      
   case 'P':
      AddSymbol(nptr, symbol, Part, True);
      break;
      
   case 'G':
      AddSymbol(nptr, symbol, Group, True);
      break;
      
   case 'F':
      break;
   }

   /*
    * Assign the occurrence character to the node
    * Process that occurrence value
    * Assign the defined flag to the data structures
    */
   nptr->occurrence[0] = occur_char;
   ProcessOccur(nptr);
   PutDefined(nptr, IsDefined(symbol));
   return;
}




/*
   FUNCTION: 		ProcessAuxInfo

   Synopsis: 
       Process the auxiliary information.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       Auxtree     Points to the current location in the auxtree 
                   data structure

   Return value:	void

   Range:		none
*/

static void
ProcessAuxInfo(fptr, aptr)

FILE
   *fptr;  /* The file being loaded */
Alt
   *aptr;  /* The definition currently under consideration */

{
   char
      auxc;     /* The character being read in */
   int
      i,        /* Loop counter */
      info_len; /* The length in the auxiliary information */

#ifdef AUX
   /*
    * Create a new spot for the auxiliary information in the definition
    */
   if (aptr->auxinf == NULL)
   {
      aptr->auxinf = NewAux(aptr);
      Auxtree = aptr->auxinf;
   }
   else
   {
      if (*aptr->auxinf->info != '\0')
      {
	 Auxtree = NewAux(aptr);
	 Auxtree->next = aptr->auxinf;
	 Auxtree->next->prev = Auxtree;
	 aptr->auxinf = Auxtree;
      }
   }

   /*
    * ???????????  Need the aux info type passed in for this ?????
    */
   Auxtree->type = XtNewString(symbol);
   /*
     addaux_tree(Root, Auxtree->type);
     */
#endif

   /*
    * Read in the length of the aux info
    * If it isn't EOF then read in that amount of info
    */
   if (fscanf(fptr, "%d ", &info_len) != EOF)
   {
#ifdef AUX
      Auxtree->info = (char *) XtMalloc((unsigned) info_len +1);
#endif
      for (i = 0; i < info_len; i++)
      {
	 auxc = fgetc(fptr);
#ifdef AUX
	 Auxtree->info[i] = auxc;
#endif
      }
#ifdef AUX
      Auxtree->info[i] = '\0';
#endif
   }
   return;
}




/*
   FUNCTION: 		EditFile

   Synopsis: 
       Load the file.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
EditFile(fptr)

FILE
   *fptr;  /* The file being loaded */

{
   Alt
      *aptr;           /* Definition pointer */
   String
      symbol;          /* The symbol string */
   char
      str[SYMBOLSIZE], /* Input buffer read in and stripped */
      type_char,       /* The symbol type character */
      occur_char;      /* The occurrence character */
   Node
      *nptr;            /* A node pointer to help the Tree pointer 
			   The Tree pointer tracks each new node
			   nptr tracks a node being defined */
   
   /*
    * Disable the form widgets with the lists until the end
    * While the file is being scanned and not EOF, load the symbols
    * and definitions
    */
   XtUnmanageChild(XtNameToWidget(Frame, "*deflist"));
   XtUnmanageChild(XtNameToWidget(Frame, "*undlist"));

   while (fscanf(fptr, "%s", str) != EOF)
   {
#if (DEBUG == 2)
      fprintf(stderr, "Edit string = [%s]\n", str);
#endif

      /*
       * Assign the loaded buffer to symbol
       * Strip the buffer down into the symbol, type, and occurrence
       */
      symbol = XtNewString(str);
      StripLoadString(&symbol, &type_char, &occur_char);
#if (DEBUG == 2)
      fprintf(stderr, "symbol = [%s] ty = [%c] oc = [%c]\n",
	      symbol, type_char, occur_char);
#endif

      /*
       * Based on the first character of the input buffer,
       * create the tree data structure
       */
      switch (*str)
      {
      default:
	 /*
	  * It should be a node in a definition,
	  * checking the type_char verifies that
	  */
	 if ((type_char == 'P') ||
	     (type_char == 'G') ||
	     (type_char == 'F'))
	 {
	    /*
	     * If a new node is needed, create it
	     * Move Tree along
	     * Process the symbol with occurrence and type.
	     */
	    if (Tree->next == NULL)
	    {
	       Tree->next = NewNode(Tree->parent, Tree);
	    }
	    Tree = Tree->next;
	    ProcessNewSymbol(type_char, occur_char, symbol, Tree);
	 }
	 else
	 {
	    /*
	     * A syntax error was found, might be a misplaced
	     * space in the file.
	     */
	    XicaStatus(Statw, "Syntax error on token [%s]\n", str);
	    XicaStatus(Statw, "Might be a misplaced space\n");
	    return;
	 }
	 break;

      case '!':
	 /*
	  * This signifies a symbol about to be defined
	  */
	 if (Tree == Root)
	 {
	    AddSymbol(Tree, symbol, Part, True);
	 }

	 /*
	  * Move Tree to where the definition for a symbol will be created
	  */
	 Tree = GetDefinition(symbol);

	 /*
	  * Set nptr to point to the parent of the next set of
	  * alternative definitions
	  */
	 nptr = Tree;
	 PutTerminal(Tree, TriF);
	 break;
	 
      case '&':
	 /*
	  * This signifies the beginning of a new definition for a symbol
	  * nptr points to the symbol being defined
	  * If a definition doesn't exist, create one
	  */
	 if (nptr == NULL)
	 {
	    fprintf(stderr,
		    "ERROR: syntax error in dgm file, unexpected '&'\n");
	    exit(1);
	 }
	 
	 if (nptr->alts == NULL)
	 {
	    /*
	     * Create a new definition
	     * Assign aptr to point to it
	     * Flag nptr as being defined
	     */
	    nptr->alts = NewAlt(nptr, (Alt *) NULL);
	    aptr = nptr->alts;
	    PutDefined(nptr, True);
	 }
	 else
	 {
	    /*
	     * Find a location for the new definition
	     * Create the new definition
	     * Assign aptr to that definition
	     */
	    for (aptr = nptr->alts; aptr->next != NULL; aptr = aptr->next)
	    {
	    }
	    aptr->next = NewAlt(aptr->whom, aptr);
	    aptr = aptr->next;
	 }

	 /*
	  * Move Tree to the beginning of the newly created definition
	  * Process the new symbol
	  */
	 Tree = aptr->prod;
	 ProcessNewSymbol(type_char, occur_char, symbol, Tree);
	 break;
	 
      case '#':
	 /*
	  * This signifies auxiliary information
	  * Put aptr at the right definition
	  * Process the auxiliary information
	  */
	 aptr = FindAlt(Tree);
	 ProcessAuxInfo(fptr, aptr);
	 break;
      } /* end SWITCH */
   } /* WHILE fscanf */

   XtManageChild(XtNameToWidget(Frame, "*deflist"));
   XtManageChild(XtNameToWidget(Frame, "*undlist"));
   return;
}
