/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: remove.c,v $
   Purpose:     
       To provide Remove routines that manipulate the 
       data structures and widgets.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.7 $
   Date:        $Date: 92/03/18 14:16:57 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  RemoveAlt
                                           IsFirstAlt
					   PutDefined
   EXPORT  RemoveNode
                           RemoveAlt
                                           FindAlt
					   GetExposedAlt
					   IsFirstNode
					   IsRoot
					   KillNodeWidgets
   EXPORT  RemoveNullSymbols
                           RemoveAlt
			   RemoveNullSymbols
			                   FindAlt
					   IsRoot
					   KillNodeWidgets
   HIDDEN  RemoveAll

   EXPORT  RemoveOccur

   EXPORT  RemoveSymbol
                           RemoveAll
			   RemoveOccur
			   RemoveSubtree
			                   GetIndex
					   RemoveDefItem
					   RemoveUndItem
   EXPORT  RemoveSubtree
                           RemoveNode
			   RemoveSymbol
*/

/* HISTORY
   $Log:	remove.c,v $
 * Revision 1.7  92/03/18  14:16:57  ware
 * Working on consistent interface.
 * 
 * Revision 1.6  1992/01/28  15:41:43  cso
 * Allocated empty strings so that freeing won't cause problems
 *
 * Revision 1.5  1992/01/14  20:22:02  cso
 *  - Reorganizing header files.
 *  - Fixed pointer progress in RemoveNullSymbols and RemoveSubtree so
 * that the freed pointer is not being accessed after it's been freed.
 *
 * Revision 1.4  1991/11/22  23:35:19  cso
 * Finally, have the display working as expected, perhaps a bit quicker,
 * but there is still a small amount of flicker on the redraws.  The
 * fixes included some small problems when removing empty place holders
 * and complete alternatives.
 *
 * Revision 1.3  1991/11/20  00:04:56  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.2  1991/11/06  16:12:16  cso
 * Initialized uninitialized variables found by QA_C...
 *
 * Revision 1.1  1991/10/04  19:07:09  jbm
 * Initial revision
 *
 * Revision 2.0  91/03/30  13:51:49  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.3  91/03/28  14:46:58  cso
 * Added the appropriate checks and assignments for updating the exposed
 * flags for any alternatives that may be removed.
 * 
 * Revision 1.2  91/02/22  09:55:32  cso
 * Completed documentation.
 * 
 * Revision 1.1  90/11/26  13:16:03  cso
 * Initial revision
 * 
*/

/* #INCLUDES */

#include "std.h"
#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "deflist.h"
#include "display.h"
#include "main.h"
#include "node.h"
#include "put.h"
#include "status.h"
#include "tree.h"
#include "undlist.h"
#include "utils.h"

/* Forward definition for internal routines */

void
   RemoveSubtree();




/*
   FUNCTION: 		RemoveAlt

   Synopsis: 

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/27   cso    Added the checks and assignments for exposed alts

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
RemoveAlt(aptr)

Alt
   *aptr;  /* Pointer to the definition being removed */

{
   Alt
      *bptr; /* Pointer to another definition to maintain the structure */
   Node
      *nptr; /* Pointer to the node whose definition is being removed */

   /*
    * If the pointer is not NULL, perform the removal
    */
   if (aptr != NULL)
   {
      /*
       * Get the owner of this definition
       * If this is the only definition, reset the node as undefined
       * Else If the first definition is being removed, fix pointers
       * Else fix pointers
       * Free the pointer
       */
      nptr = aptr->whom;
      if ((IsFirstAlt(aptr) == True) && (IsLastAlt(aptr) == True))
      {
	 nptr->alts = NULL;
	 PutDefined(nptr, False);
      }
      else if (IsFirstAlt(aptr) == True)
      {
	 nptr->alts          = aptr->next;
	 aptr->next->prev    = NULL;
	 aptr->next->exposed = True;
      }
      else
      {
	 /*
	  * Access the previous definition
	  * Point it to the definition after the one removed
	  * If that is not NULL, make it point to the previous definition
	  */
	 bptr       = aptr->prev;
	 bptr->next = aptr->next;
	 if (bptr->next != NULL)
	 {
	    bptr->next->prev = bptr;
	 }
	 bptr->exposed = True;
      }
      XtFree(aptr);
   }
}




/*
   FUNCTION: 		RemoveNode

   Synopsis: 
       Remove a node from the node data structure.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/27   cso    If an alt is removed then set a new one to
                          being exposed

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		none
*/

Node
*RemoveNode(nptr, vertical)

Node
   *nptr;     /* Node being removed */
Boolean
   vertical;  /* Display orientation */

{
   Alt
      *aptr;  /* The definition from which the node is being removed */
   Node
      *qptr;  /* A sibling or parent node of the one removed */
   
   /*
    * If the node is already NULL, or ifit's the Root,
    * set the return node to the Root
    * Else remove the node
    */
   qptr = NULL;
   if ((nptr == NULL) || (IsRoot(nptr) == True))
   {
      qptr = Root;
   }
   else
   {
      /*
       * Destroy the widgets for this node
       * If the definition for this node can be found
       * Remove the node from the definition
       */
      KillNodeWidgets(nptr, vertical);
      if ((aptr = FindAlt(nptr)) != NULL)
      {
	 /*
	  * If it is the only node in the definition
	  * Set the return node to the removed node's parent
	  * Remove the definition pointer
	  * Else there is a sibling
	  */
	 if ((IsFirstNode(nptr) == True) && (IsLastNode(nptr) == True))
	 {
	    qptr = nptr->parent;
	    RemoveAlt(aptr);
	    aptr = GetExposedAlt(qptr);
	    ShowDisplayTree(qptr);
	    if (aptr != NULL)
	    {
	       qptr = aptr->prod;
	    }
	 } /***** IF the node is the only symbol in the alternative */
	 else if (IsFirstNode(nptr) == True)
	 {
	    /*
	     * The removed node is the first one in the definition
	     * Set the definition to point to the next sibling
	     */
	    qptr       = nptr->next;
	    qptr->prev = NULL;
	    aptr->prod = qptr;
	 } /***** ELSE IF the node is first in the production */
	 else
	 {
	    /*
	     * The removed node is either in the middle or the end
	     * Set the siblings up to point around the removed node
	     */
	    qptr       = nptr->prev;
	    qptr->next = nptr->next;
	    if (qptr->next != NULL)
	    {
	       qptr->next->prev = qptr;
	    }
	 } /***** ELSE the node has a sibling to the left */
      } /***** IF an alternative exists for this node */
   } /***** ELSE it is not the root or null node */
   return(qptr);
}




/*
   FUNCTION: 		RemoveNullSymbols

   Synopsis: 
       Because of the default setting to provide an empty (NULL) node
       for the user to fill, many of them may exist, so elimnate them.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		none
*/

void
RemoveNullSymbols(nptr)

Node
   *nptr;  /* The node under consideration */

{
   Alt
      *aptr,
      *bptr;   /* Pointer to definitions (in a loop too) */
   Node
      *qptr,
      *rptr;   /* Loop pointer and a pointer to nonNULL siblings */

   /*
    * If the symbol under consideration is NULL
    * Remove it
    * Else keep searching for NULL nodes
    */
   if (*nptr->symbol == '\0')
   {
      /*
       * If it is not the Root symbol
       * Remove it.
       */
      if (IsRoot(nptr) == False)
      {
	 /*
	  * Remove any widgets for it
	  * Find its definition pointer
	  */
	 KillNodeWidgets(nptr, True);
	 aptr = FindAlt(nptr);
	 if (aptr != NULL)
	 {
	    /*
	     * If it is the first Node, remove it and see if it is the only
	     * node in the definition (if so remove the definition)
	     */
	    if (aptr->prod == nptr)
	    {
	       aptr->prod = nptr->next;
	       if (nptr->next != NULL)
	       {
		  nptr->next->prev = NULL;
	       }
	       else
	       {
		  RemoveAlt(aptr);
	       }
	    } /***** IF nptr is first in the alternative */
	    else
	    {
	       /*
		* Otherwise, it is in the middle or at the end
		* It is guaranteed that a node exists before it so fix 
		* the pointers.
		*/
	       qptr = nptr->prev;
	       qptr->next = nptr->next;
	       if (qptr->next != NULL)
	       {
		  qptr->next->prev = qptr;
	       }
	    }
	 }
	 XtFree(nptr);
	 nptr = NULL;
      }
   }
   else
   {
      /*
       * Visit any children of the current node
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = bptr)
      {
	 bptr = aptr->next;
	 for (qptr = aptr->prod; qptr != NULL; qptr = rptr)
	 {
	    rptr = qptr->next;
	    RemoveNullSymbols(qptr);
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		RemoveAll

   Synopsis: 
       Remove a symbol completely from the symbol table.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info    Removing a symbol from the table
       Ndx     To limit the adjustment

   Globals Modified:
       Info   A symbol is removed
       Ndx    It is decremented

   Return value:	static void

   Range:		none
*/

static void
RemoveAll(i)

int
   i;   /* Index of the symbol to remove */

{
   int
      j;  /* Loop counter */

   /*
    * If it is not the next available symbol (i.e., not one to remove)
    * Remove the symbol
    */
   if (i != Ndx)
   {
      /*
       * Because the symbol table is an array, we have to do this the
       * old-fashioned way, from this point down the array, move
       * each symbol up one position.
       * Decrement Ndx
       * Initialize Ndx
       * Make sure that all parts are covered (check types.h)
       */
      for (j = i; j < (Ndx - 1); j++)
      {
	 Info[j].symbol      = XtNewString(Info[j+1].symbol);
	 Info[j].symtype     = Info[j+1].symtype;
	 Info[j].filename    = XtNewString(Info[j+1].filename);
	 Info[j].defined     = Info[j+1].defined;
	 Info[j].definition  = Info[j+1].definition;
	 Info[j].count       = Info[j+1].count;
	 Info[j].star        = Info[j+1].star;
	 Info[j].plus        = Info[j+1].plus;
	 Info[j].optional    = Info[j+1].optional;
	 Info[j].terminal    = Info[j+1].terminal;
	 Info[j].terminate   = Info[j+1].terminate;
	 Info[j].visits      = Info[j+1].visits;
      }
      Ndx--;
      Info[Ndx].symbol      = XtNewString("");
      Info[Ndx].symtype     = Unknown;
      Info[Ndx].filename    = XtNewString("");
      Info[Ndx].defined     = False;
      Info[Ndx].definition  = NULL;
      Info[Ndx].count       = 0;
      Info[Ndx].star        = 0;
      Info[Ndx].plus        = 0;
      Info[Ndx].optional    = 0;
      Info[Ndx].terminal    = TriU;
      Info[Ndx].terminate   = TriU;
      Info[Ndx].visits      = 0;
   }
   return;
}




/*
   FUNCTION: 		RemoveOccur

   Synopsis: 
       Remove an occurrence of a symbol from the symbol table.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info    The symbol table is accessed

   Globals Modified:
       Info    The symbol table is modified

   Return value:	void

   Range:		none
*/

void
RemoveOccur(i, occurchar)

int
   i;         /* Index into the table */
char
   occurchar; /* The type of occurrence to remove */

{
   /*
    * Removal is accomplished by decrementing a counter
    */
   switch (occurchar)
   {
   default:
      break;

   case '^':
      break;

   case '_':
      break;

   case '?':
      Info[i].optional--;
      break;

   case '*':
      Info[i].star--;
      break;

   case '+':
      Info[i].plus--;
      break;
   }
   return;
}




/*
   FUNCTION: 		RemoveSymbol

   Synopsis: 
       Removes one instance of a symbol from the symbol table.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info     To access the symbol table

   Globals Modified:
       Info     An instance is removed

   Return value:	void

   Range:		none
*/

void
RemoveSymbol(nptr, vertical)

Node
   *nptr;     /* The instance being removed */
Boolean
   vertical;  /* The display orientation */

{
   int
      cnt,        /* Number of arguments */
      i;          /* Index into the symbol table */
   Alt
      *aptr;      /* Pointer to definitions */
   Arg
      args[MAXARGS];/* List of arguments */
   Node
      *qptr;      /* Loop pointer to access nodes in a definition */
   NodeList
      *nlptr,     /* Pointer to the definition where this instance is 
		     being removed */
      *olptr;     /* Pointer to a possible replacement for the 
		     definition holder for this symbol */

   /*
    * Get the index of the symbol in the symbol table
    * If only one instance remains, remove it completely
    * Else If an instance exists remove it
    * Else an error
    */
   i = GetIndex(nptr->symbol);
   if (i == Ndx)
   {
      /*
       * It must be an empty place holder, so ignore this instance
       */
      return;
   }
   else if (Info[i].count == 1)
   {
      RemoveAll(i);
      RemoveSubtree(nptr);
      RemoveDefItem(nptr->symbol);
      RemoveUndItem(nptr->symbol);
   }
   else if (Info[i].definition != NULL)
   {
      /*
       * If a node list exists for this symbol (should be one for every
       * count)
       * Decrement count
       * Remove the occurrence (decrement the respective count)
       * Find the definition that this instance is being removed from
       * If one exists, remove the instance
       * Else an error
       */
      Info[i].count--;
      RemoveOccur(i, *nptr->occurrence);
      for (nlptr = Info[i].definition;
	   ((nlptr != NULL) && (nlptr->nodeptr != nptr));
	   nlptr = nlptr->next)
      {
      }
      if (nlptr != NULL)
      {
	 /*
	  * If only one definition remains, this is an error
	  * Else If it is the first definition (this is special!)
	  */
	 if ((nlptr->prev == NULL) && (nlptr->next == NULL))
	 {
	    XicaStatus(Statw, "RemoveSymbol: ERROR count != # of defs\n");
	 }
	 else if (nlptr->prev == NULL)
	 {
	    /*
	     * This is a special case because a symbol's definition exists
	     * where it is first used in the tree.
	     * Adjust the nodelist pointers
	     * Change the subsequent instance of this symbol to be the
	     * position of its definition
	     */
	    Info[i].definition   = nlptr->next;
	    olptr                = nlptr->next;
	    olptr->prev          = NULL;
	    olptr->nodeptr->alts = nlptr->nodeptr->alts;
	    for (aptr = olptr->nodeptr->alts; aptr != NULL; aptr = aptr->next)
	    {
	       /*
		* All of the symbol's children should reference the 
		* correct parent
		*/
	       aptr->whom    = olptr->nodeptr;
	       for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	       {
		  qptr->parent = olptr->nodeptr;
		  if (qptr->treew != NULL)
		  {
		     /*
		      * Similar to fixing the parent pointers, are the
		      * parent resources for the tree widgets
		      * Depending on the orientation the new parent
		      * must be made to be above or to the left of 
		      * the children
		      */
		     XtUnmanageChild(qptr->treew);
		     cnt = 0;
		     if (vertical == True)
		     {
			XtSetArg(args[cnt], XtNfromVert,  qptr->parent->treew);
		     }
		     else
		     {
			XtSetArg(args[cnt], XtNfromHoriz, qptr->parent->treew);
		     }
		     cnt++;
		     XtSetValues(qptr->treew, args, cnt);
		     XtManageChild(qptr->treew);
		  } /* IF treew exists */
	       } /* FOR each node in a definition */
	    } /* FOR each definition */
	 } /* ELSE IF nlptr is the first definition */
	 else
	 {
	    /*
	     * The simple case, fix the nodelist pointers to point around
	     * the symbol's existence
	     */
	    olptr       = nlptr->prev;
	    olptr->next = nlptr->next;
	    if (olptr->next != NULL)
	    {
	       olptr->next->prev = olptr;
	    }
	 } /* ELSE if nlptr is in the middle or at the end */

	 /*
	  * Free up the space for that instance
	  */
	 XtFree(nlptr);
      } /* IF a node list exists for this symbol */
      else
      {
	 XicaStatus(Statw, "RemoveSymbol: ERROR in symbol table\n");
      }
   } /* ELSE IF a definition exists */
   else
   {
      XicaStatus(Statw, "RemoveSymbol: ERROR in symbol table\n");
   }
   return;
}




/*
   FUNCTION: 		RemoveSubtree

   Synopsis: 
       Remove a subtree.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
RemoveSubtree(nptr)

Node
   *nptr;  /* Node whose subtree is being removed */

{
   Alt
      *aptr,
      *bptr; /* Loop pointer to access each definition */
   Node
      *qptr,
      *rptr; /* Loop pointer to access each node in a definition */

   if (nptr != NULL)
   {
      /*
       * Remove an instance of each node in the subtree under nptr
       */
      for (aptr = nptr->alts; aptr != NULL; aptr = bptr)
      {
	 bptr = aptr->next;
	 for (qptr = aptr->prod; qptr != NULL; qptr = rptr)
	 {
	    rptr = qptr->next;
	    RemoveSymbol(qptr, True);
	    (void) RemoveNode(qptr, True);
	 }
      }
   }
   return;
}


