/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: symtab.c,v $
   Purpose:
       Functions to handle the symbol table.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.5 $
   Date:        $Date: 92/03/18 14:17:31 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   EXPORT  ChangeSymbol
                                           GetIndex
					   PutLabel
					   PutString
   EXPORT  IsDefined
                                           GetIndex
   EXPORT  IsGroup
                                           GetSymbolType
   EXPORT  IsTerminal
                                           GetIndex
   EXPORT  Visited
                                           GetIndex
   EXPORT  ResetVisits

   EXPORT  InitTerminate

   EXPORT  CheckTerminates
                           CheckTerminates
			   Visited
			                   GetDefinition
					   GetIndex
					   
*/

/* HISTORY
   $Log:	symtab.c,v $
 * Revision 1.5  92/03/18  14:17:31  ware
 * Working on consistent interface.
 * 
 * Revision 1.4  1992/02/20  04:44:01  cso
 * Fixed a couple of minor problems with changing (editing) a symbol's
 * name, and ensured that the proper allocation and freeing of strings
 * occurred.
 *
 * Revision 1.3  1992/01/14  17:57:24  cso
 *  - Reorganizing header files.
 *
 * Revision 1.2  1991/11/20  00:05:20  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.1.1.3  1991/10/04  20:16:47  jbm
 * First HaL mods to ICA
 *
 * Revision 2.2  91/05/28  10:17:15  cso
 * Added an explicit return for lint.
 * 
 * Revision 2.1  91/04/01  09:51:54  cso
 * Added an additional check for NULL symbols when checking termination.
 * 
 * Revision 2.0  91/03/30  13:52:49  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.13  91/03/28  14:48:19  cso
 * Added a check for NULL widgets.
 * 
 * Revision 1.12  91/02/22  10:00:13  cso
 * Completed documentation.
 * 
 * Revision 1.11  90/11/26  13:20:39  cso
 * Placed Get, Put, and Remove routines in their own modules.
 * Thinking about a boolean module or is module for all of the Is routines.
 * That may eliminate node.c.
 * 
 * Revision 1.10  90/11/20  15:56:41  cso
 * Modified the ChangeSymbol routine.
 * Fixes from lint.
 * Fixed the Remove functions for Delete and Erase.
 * 
 * Revision 1.9  90/11/05  18:27:29  cso
 * Added function template to all functions, but still need to fill it in.
 * Modified the Definition functions because of a change to the data structure.
 * Now there is a list of node ptrs so that I can track all occurrences of
 * a symbol and hopefully do Deletes better.
 * 
 * Revision 1.8  90/10/24  15:38:04  cso
 * Added a function to count the number of unique productions.
 * 
 * Revision 1.7  90/10/02  13:37:03  cso
 * Cosmetic changes.
 * 
 * Revision 1.6  90/09/07  09:40:36  cso
 * Cosmetics.
 * 
 * Revision 1.5  90/09/03  14:40:29  cso
 * Changed function name to be consistent.
 * 
 * Revision 1.4  90/08/26  17:43:24  cso
 * Comment all calls RE: auxiliary information.
 * 
 * Revision 1.3  90/08/26  14:13:37  cso
 * added module template
 * 
*/

/* #INCLUDES */

#include "widgets.h"
#include "types.h"

#include "get.h"
#include "main.h"
#include "put.h"
#include "status.h"



/*
   FUNCTION: 		ChangeSymbol

   Synopsis: 
       This function changes a symbol everywhere in the data structure.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/13   cso    Check if widgets exist before assigning the symbol
       91/07/16   cso    Check to see if the new symbol, that an old
                          one is being changed to, already exists.  If
			  so, do not allow the change.

   Globals used:
       Info       Access to the symbol table
       Ndx        Comparison purposes

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

Boolean
ChangeSymbol(nptr, oldstr)

Node
   *nptr;      /* The node whose symbol is being changed */
String
   oldstr;     /* The new symbol string */

{
   int
      i;       /* Index into the symbol table */
   NodeList
      *nlptr;  /* Loop pointer to change all of the 
		  instances of that symbol */
   String
      newstr;
   
   /*
    * Check to see if the new symbol already exists...
    * Get the index of the old symbol in the table
    * Replace the old with the new by first
    * Freeing the old string and the assigning the new string
    */
   newstr = XtNewString(nptr->symbol);
   i = GetIndex(newstr);
   if (i < Ndx)
   {
      /*
       * Warn the user that it already exists and reset the node
       * widget.
       */
      XicaStatus(Statw,
	      "WARNING: [%s] already exists, delete [%s] first\n",
	      newstr, oldstr);
      PutString(nptr->nodew, oldstr);
      return(False);
   }

   i = GetIndex(oldstr);
   XtFree(Info[i].symbol);
   Info[i].symbol = XtNewString(newstr);
   for (nlptr = Info[i].definition; nlptr != NULL; nlptr = nlptr->next)
   {
      /*
       * For each instance of the symbol in the data structures
       * Replace the strings displayed in widgets
       * Free up the old string in the data structure instances
       * If the appropriate widgets exist
       * Assign the new string in the data structure instances
       */
      if (nlptr->nodeptr->nodew != NULL)
      {
	 PutString(nlptr->nodeptr->nodew, newstr);
      }
      if (nlptr->nodeptr->treew != NULL)
      {
	 PutLabel(nlptr->nodeptr->treew,  newstr);
      }
      
      XtFree(nlptr->nodeptr->symbol);
      nlptr->nodeptr->symbol = XtNewString(newstr);
   }
   XtFree(newstr);
   return(True);
}




/*
   FUNCTION: 		IsDefined

   Synopsis: 
       Returns the Boolean indicating whether a symbol is defined or not.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info       Access to the symbol table

   Globals Modified:
       none

   Return value:	Boolean

   Range:		True or False
*/

Boolean
IsDefined(str)

String
   str;   /* The symbol under consideration */

{
   int
      i;  /* Index into the symbol table */

   /*
    * Get the index of the symbol
    * Return the value in the defined part of the symbol table
    */
   i = GetIndex(str);
   return(Info[i].defined);
}




/*
   FUNCTION: 		IsGroup

   Synopsis: 
       Return the boolean indicating if the symbol is a group symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Boolean

   Range:		True or False
*/

Boolean
IsGroup(str)

String
   str; /* The symbol under consideration */

{
   /*
    * If the symbol type is Group, return True, else False
    */
   if (GetSymbolType(str) == Group)
   {
      return(True);
   }
   else
   {
      return(False);
   }
}




/*
   FUNCTION: 		IsTerminal

   Synopsis: 
       Return a Boolean indicating the terminal status of a symbol.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info       Access to the symbol table

   Globals Modified:
       none

   Return value:	TriState

   Range:		TriT, TriF, TriU (True, False, Unknown)
*/

TriState
IsTerminal(str)

String
   str;  /* Symbol under consideration */

{
   int
      i; /* Index into the symbol table */

   /*
    * Get the index
    * Return the terminal status
    */
   i = GetIndex(str);
   return(Info[i].terminal);
}




/*
   FUNCTION: 		Visited

   Synopsis: 
       Returns the number of times this symbol has been visited 
       using this function.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info       Access to the symbol table

   Globals Modified:
       none

   Return value:	int

   Range:		> 0
*/

int
Visited(str)

String
   str;  /* Symbol under consideration */

{
   int
      i; /* Index into the symbol table */

   /*
    * Get the index
    * Increment the visit counter
    * Return that value
    */
   i = GetIndex(str);
   Info[i].visits++;
   return(Info[i].visits);
}




/*
   FUNCTION: 		ResetVisits

   Synopsis: 
       Usually called before using the Visits() function to reset the visits
       value of the symbol table.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info       Access to the symbol table
       Ndx        To limit access to the symbol table

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
ResetVisits()

{
   int
      i; /* Loop counter to go through the symbol table */

   /*
    * For each symbol
    * Reset the visits counter to 0
    */
   for (i = 0; i < Ndx; i++)
   {
      Info[i].visits = 0;
   }
   return;
}




/*
   FUNCTION: 		InitTerminate

   Synopsis: 
       Initialize the terminate status of the symbol table.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Info       Access to the symbol table
       Ndx        To limit access to the symbol table

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
InitTerminate()

{
   int
      i;  /* Loop counter to initialize each symbol */

   /*
    * For each symbol
    * If the symbol is a terminal, then it terminates
    * Else If the symbol is not defined, then it is assumed to terminate
    * Else it is unknown if the symbol terminates
    */
   for (i = 0; i < Ndx; i++)
   {
      Info[i].visits = 0;
      if (Info[i].terminal == TriT)
      {
	 Info[i].terminate = TriT;
      }
      else if (Info[i].defined != True)
      {
	 /*
	  * Warn the user that Info[i].symbol has not been defined
	  * Set the terminate to True (assumption that if it were defined
	  * that it would terminate)
	  */
	 XicaStatus(Statw, "WARNING: Symbol [%s] has NOT been defined!\n",
		 Info[i].symbol);
	 Info[i].terminate = TriT;
      }
      else
      {
	 Info[i].terminate = TriU;
      }
   }
   return;
}




/*
   FUNCTION: 		CheckTerminates

   Synopsis: 
       Check whether or not a symbol terminates, i.e., the grammar can 
       generate a language more than the empty language.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/04/01   cso    Had to add an additional check for NULL
                          symbols to make sure that they do not affect
			  the check for termination.
       91/05/23   cso    Added an explicit return of False.

   Globals used:
       Info       Access to the symbol table

   Globals Modified:
       none

   Return value:	Boolean

   Range:		True or False
*/

Boolean
CheckTerminates(nptr)

Node
   *nptr;       /* Node under consideration */

{
   Alt
      *aptr;    /* Loop pointer to visit each definition of nptr */
   Node
      *qptr;    /* Loop pointer for each symbol in a definition */
   String
      nptrstr,  /* Symbol in nptr */
      qptrstr;  /* Symbol in qptr */
   int
      ndx1,     /* Index of nptr in symbol table */
      ndx2;     /* Index of qptr in symbol table */

   /*
    * Assign the node symbol to nptrstr
    * If the symbol is not NULL, check it
    * Else return True
    */
   nptrstr = XtNewString(nptr->symbol);
   if (*nptrstr != '\0')
   {
      /*
       * Get the index in the symbol table of the symbol under consideration
       * Increment its visits counter, so we don't get into an infinite loop
       * If it terminates because it is a terminal or undefined, return True
       * Else check it out
       */
      ndx1 = GetIndex(nptrstr);
      Info[ndx1].visits++;
      if (Info[ndx1].terminate == TriT)
      {
	 return(True);
      }
      else
      {
	 /*
	  * Get the location of this symbol's definition
	  * Make sure it isn't NULL
	  * For each definition, for each symbol, check if it terminates
	  * If an entire definition terminates then the symbol in question
	  * terminates, else check the next definition
	  */
	 nptr = GetDefinition(nptrstr);
	 if (nptr != NULL)
	 {
	    for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
	    {
	       for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	       {
		  /*
		   * Get the symbol of this node
		   * Get the symbol table index of this symbol
		   * If the occurrence cannot be zero (? or *),
		   * delve deeper into the termination question
		   */
		  if (*qptr->symbol != '\0')
		  {
		     qptrstr = qptr->symbol;
		     ndx2    = GetIndex(qptrstr);
		     if ((*qptr->occurrence != '?') &&
			 (*qptr->occurrence != '*'))
		     {
			/*
			 * If this symbol has not been visited yet,
			 * check if it terminates
			 * Else if this symbol is not known to terminate
			 * Stop checking this definition (break)
			 */
			if (Visited(qptrstr) == 1)
			{
			   /*
			    * If it returns False,
			    * Stop checking this definition (break)
			    */
			   if (CheckTerminates(qptr) == False)
			   {
			      break;
			   }
			}
			else if (Info[ndx2].terminate != TriT)
			{
			   break;
			}
		     } /* IF occurrence requires at least one instance */
		  } /* IF the symbol is not NULL */
	       } /* FOR each symbol in the definition */
	       
	       /*
		* If we reached the end of a definition, i.e., the 
		* FOR loop ended because qptr == NULL,
		* the symbol under consideration has a definition that
		* terminates, therefore, that symbol terminates
		* Set the terminate status as True, and return True
		*/
	       if (qptr == NULL)
	       {
		  Info[ndx1].terminate = TriT;
		  return(True);
	       }
	    } /* FOR each definition of nptr */
	    
	    /*
	     * The symbol under consideration did not terminate
	     * inform the user of which symbol
	     * Set the terminate status to False, and return False
	     */
	    XicaStatus(Statw, "WARNING: Symbol [%s] does NOT terminate!\n",
		    nptrstr);
	    Info[ndx1].terminate = TriF;
	    return(False);
	 } /* IF nptr is not NULL */
      } /* IF (it terminates) ELSE */
   } /* IF the symbol is nonNULL */
   else
   {
      return(True);
   }
   return(False);
}
