/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: tree.c,v $
   Purpose:
       To provide functions for an absolute, tree navigation interface.

   Programmer:  Conleth O'Connell, Laura Lodge, and Susan Keiff
   Reviewer:    
   Revision:    $Revision: 1.10 $
   Date:        $Date: 92/05/29 14:46:35 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   HIDDEN  FindSelectedNode
                           FindSelectedNode
   HIDDEN  JumpTree
                           FindSelectedNode
			                   UpdateInterface
   EXPORT  GetExposedAlt

   EXPORT  GetFirstExposedChild
                           GetExposedAlt
   EXPORT  CreateTreeW
                           GetFirstExposedChild
			                   IsFirstNode
					   IsLastNode
					   IsRoot
   HIDDEN  RedrawLegs
                           GetExposedAlt
                           RedrawLegs
			                   DrawDogLegs
					   IsRoot
   HIDDEN  RefreshTree
                           RedrawLegs
   EXPORT  CreateTreeShell
					   CreateCompass
*/

/* HISTORY
   $Log:	tree.c,v $
 * Revision 1.10  92/05/29  14:46:35  boyd
 * Changes for improved resizing, consisting mostly of replacing Form widgets
 * with Group, Table, and Viewform widgets.
 * 
 * 
 * Revision 1.9  92/03/18  14:17:41  ware
 * Working on consistent interface.
 * 
 * Revision 1.8  1992/01/14  17:57:10  cso
 *  - Reorganizing header files.
 *
 * Revision 1.7  1991/11/22  23:35:30  cso
 * Finally, have the display working as expected, perhaps a bit quicker,
 * but there is still a small amount of flicker on the redraws.  The
 * fixes included some small problems when removing empty place holders
 * and complete alternatives.
 *
 * Revision 1.6  1991/11/22  15:35:28  cso
 * 	Added a graphics context to the node type for easier
 * manipualtion of the line drawing in the navigation tree.  Tried to
 * eliminate the number of refreshes occurring for a particular expose
 * event, tough to debug in synchronous mode.
 *
 * Revision 1.5  1991/11/20  00:05:31  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.4  1991/11/19  16:37:33  cso
 * Added keyboard accelerators and panner settings to fallback resources.
 *
 * Revision 1.3  1991/11/18  23:47:23  cso
 * Added the Release 5 panner and porthole widgets to replace the
 * scrollbars on the navigation tree window.
 * They are compile time usable by using -DR5 in the CFLAGS.
 *
 * Revision 1.2  1991/11/07  21:03:51  cso
 * *** empty log message ***
 *
 * Revision 1.1.1.3  1991/10/04  20:16:52  jbm
 * First HaL mods to ICA
 *
 * Revision 2.1  91/05/28  10:17:47  cso
 * Changed the shell being used to be a popup.
 * 
 * Revision 2.0  91/03/30  13:53:07  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.19  91/03/28  14:48:48  cso
 * Changed GetExposedAlt to return NULL if no exposed alts are found.
 * Getting foreground color from borderColor.
 * 
 * Revision 1.18  91/02/26  15:26:11  cso
 * Because the fromHoriz/Vert resources aren't being used for positioning 
 * a widget relative to a sibling, don't hassle with it any longer.
 * 
 * Revision 1.17  91/02/22  10:01:11  cso
 * Completed documentation.
 * In the process of changing CreateTreeW.
 * Removed the ClipSubtree and RemoveFromTree functions from
 * the instance editor code.
 * 
 * Revision 1.16  90/11/29  14:26:17  cso
 * Widened the default tree display.
 * 
 * Revision 1.15  90/11/20  15:58:22  cso
 * Put functions into display.c from here to break this down a bit.
 * 
 * Revision 1.14  90/11/05  18:28:37  cso
 * Added function template to all functions, but still need to fill it in.
 * Still need to break PrettyPrint apart.  It's FIVE pages now!  Else, 
 * cosmetics.
 * 
 * Revision 1.13  90/10/29  10:29:04  cso
 * Changed Refresh to RefreshTree.  Also, I'm still fiddling with 
 * PrettyPrintTree and the placement of the parent widget.  There seems
 * to be a problem especially with alternative defs.  (3-D stuff!)
 * 
 * Revision 1.12  90/10/24  15:39:18  cso
 * Cosmetic changes.
 * 
 * Revision 1.11  90/10/04  15:58:28  cso
 * Finally fixed the problems with the form widget and the PrettyPrinting
 * in 2D.  Now, I have to work on the 3D.  Essentially, the fix revolves
 * around using Unmanage/Manage with resetting resources in order to
 * get the Form widget to know that changes took place and to grow
 * appropriately.
 * I also moved PrettyPrinting to take place only from PutSymbolInTree.
 * Eventually it will also take place after RemoveSymbolFromTree.  Else,
 * it should never have to take place.
 * 
 * Revision 1.10  90/10/02  13:38:11  cso
 * Trying to determine why the form widget keeps getting clip
 * getting clipped by the viewport widget.
 * 
 * Revision 1.9  90/09/25  15:40:43  cso
 * Now devegram draws lines in treeframe in reverse video!  I had to use
 * fill_style and GCFillStyle.
 * 
 * Revision 1.8  90/09/07  09:40:57  cso
 * Fixed the two-dimensional layout problems.  Except I think the form widget
 * is still giving us problems.
 * 
 * Revision 1.7  90/09/03  18:27:38  cso
 * Changed some data structures.  Altered the PrettyPrintTree and DrawDogLEgs
 * routines.  Removed unused routines including CreateTreewAnd Struct.
 * 
 * Revision 1.6  90/09/03  14:41:08  cso
 * Changed function names to be consistent.  Changed to XtMoveWidget maintain
 * the correct position with node_x/node_y.
 * 
 * Revision 1.5  90/08/30  11:31:39  cso
 * This is a new version from the Instance Editor.
 * 
*/

/* #INCLUDES */

#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "display.h"
#include "main.h"
#include "node.h"
#include "utils.h"

/* Globals */

GC
   Gc,        /* Graphics context for the drawing routines */
   Pathgc;    /* Graphics context for drawing thick lines */
Widget
   Treeshell; /* The shell widget for getting access to other widgets
		 contained in its widget subtree */





/*
   FUNCTION: 		*FindSelectedNode

   Synopsis: 
       Finds a node in the display tree that the user has selected.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static Node *

   Range:		NULL or valid pointer
*/

static Node
*FindSelectedNode(nptr, w)

Node
   *nptr;     /* Node under consideration */
Widget
   w;         /* The widget selected by the user */

{
   
   Node 
      *qptr,  /* Loop pointer to visit each node in a definition */
      *rptr;  /* Node pointer being returned from recursive calls */
   Alt
      *aptr;  /* Loop pointer to visit each definition of a node */
   
   /*
    * If the tree widget of the node equals the widget selected,
    * return that node
    * Else visit the subtree under nptr to find that node
    * If no node can be found return NULL
    */
   if (nptr->treew == w)
   {
      return(nptr);
   }
   else
   {
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next) 
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    /*
	     * If the returned node is not NULL, return it
	     */
	    rptr = FindSelectedNode(qptr, w);
	    if (rptr != NULL)
	    {
	       return(rptr);
	    }
	 }
      }
   }
   return(NULL);
}




/*
   FUNCTION: 		JumpTree

   Synopsis: 
       Callback that is used when the user has selected a tree widget.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
JumpTree(w, client_data, call_data)

Widget
   w;           /* Widget calling this function */
caddr_t
   client_data, /* Unused */
   call_data;   /* Unused */

{
   Node
      *nptr;    /* Node to jump to in the data structures */
   
   /*
    * Find the node whose tree widget has been selected by the user
    * If the node is not NULL
    * Update the interface from where the user was to where the user selected
    */
   nptr = FindSelectedNode(Root, w);
   if (nptr != NULL) 
   {
      UpdateInterface(Tree, nptr);
      Tree = nptr;
   }
   return;
}




/*
   FUNCTION: 		GetExposedAlt

   Synopsis: 
       Find the definition for a node that is exposed to the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/26   cso    If no exposed definition is found, return NULL

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Alt *

   Range:		NULL or valid pointer
*/

Alt
*GetExposedAlt(nptr)

Node
   *nptr;    /* Node under consideration */

{
   Alt
      *aptr; /* Loop pointer to find the exposed definition */

   /*
    * If the node is not NULL
    * search for each of its definitions to see which one is exposed
    * That definition and its parent are not exposed - return NULL
    */
   if (nptr != NULL)
   {
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 if (aptr->exposed == True)
	 {
	    return(aptr);
	 }
      }
   }
   return(NULL);
}




/*
   FUNCTION: 		GetFirstExposedChild

   Synopsis: 
       Return the first node in an exposed definition of a node.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		NULL or valid pointer
*/

Node
*GetFirstExposedChild(nptr)

Node
   *nptr;      /* Node under consideration */

{
   Alt
      *aptr;   /* Definition pointer to an exposed definition */

   /*
    * If the node has definitions
    * Get an exposed definition
    * If one exists (nonNULL)
    * return the first node in that definition
    */
   if (nptr->alts != NULL)
   {
      aptr = GetExposedAlt(nptr);
      if (aptr != NULL)
      {
	 return(aptr->prod);
      }
#ifdef DEBUG
      fprintf(stderr, "Could not find an exposed alt for [%s]\n",
	      nptr->symbol);
#endif
   }
   return(NULL);
}




/*
   FUNCTION: 		CreateTreeW

   Synopsis: 
       Create a tree widget for a node.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Widget

   Range:		Widget
*/

Widget
CreateTreeW(nptr, vertical)

Node
   *nptr;         /* Node under consideration */
Boolean
   vertical;      /* Display orientation */

{
   Arg
      args[MAXARGS];/* List of arguments */
   int
      cnt;        /* Number of arguments */
   Widget
      parent,     /* Parent widget of the one being created */
      top,        /* My parent's tree widget */
      w;          /* Widget being created and returned */

   /*
    * Get the parent widget
    * Get node parent's tree widget
    * Create the widget
    * Put the node's tree widget relative to node parent's tree widget
    * Add the callback
    * Return the widget
    */
   parent = XtNameToWidget(Treeshell, "*treeframe");
   if (IsRoot(nptr) == True)
   {
      top = NULL;
   }
   else
   {
      top = nptr->parent->treew;
   }
   cnt = 0;
   XtSetArg(args[cnt], XtNresizable, True);         cnt++;
   XtSetArg(args[cnt], XtNlabel,     nptr->symbol); cnt++;
   XtSetArg(args[cnt], XtNtop,       XtChainTop);   cnt++;
   XtSetArg(args[cnt], XtNbottom,    XtChainTop);   cnt++;
   if (vertical == True)
   {
      XtSetArg(args[cnt], XtNfromVert,  top); cnt++;
   }
   else
   {
      XtSetArg(args[cnt], XtNfromHoriz, top); cnt++;
   }
   w  = XtCreateWidget("treew", commandWidgetClass,
		       parent, args, cnt);
   XtAddCallback(w, XtNcallback, JumpTree, NULL);
   return(w);
}




/*
   FUNCTION: 		RedrawLegs

   Synopsis: 
       Redraw the lines connecting the display tree nodes.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
RedrawLegs(nptr, vertical)

Node
   *nptr;     /* Node under consideration */
Boolean
   vertical;  /* Display orientation */

{
   Alt
      *aptr;  /* Loop pointer for each definition of nptr */
   Node
      *qptr;  /* Loop pointer for each node of a definition */

   /*
    * If the node is not NULL and has a definition
    * Get the exposed definition of the node (to limit the looping)
    * If an exposed definition exist, for each of the nodes, recurse
    */
   if ((nptr != NULL) && (nptr->alts != NULL))
   {
      aptr = GetExposedAlt(nptr);
      if (aptr != NULL)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr= qptr->next)
	 {
	    /*
	     * Keep going until the leaves are found,
	     * If a node has a tree widget, and it isn't the Root
	     * Check if the tree widget is managed and draw the doglegs
	     */
	    RedrawLegs(qptr, vertical);
	    if ((qptr->treew != NULL) && (IsRoot(qptr) == False))
	    {
	       if (XtIsManaged(qptr->treew) != True)
	       {
		  XtManageChild(qptr->treew);
	       }
	       DrawDogLegs(qptr, Gc, vertical);		  
	    }
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		RefreshTree

   Synopsis: 
       Refresh the tree display.  This is needed because of the lines being
       drawn are not handled by any widgets.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Root       To initialize the RedrawLegs routine

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
RefreshTree(w, data, event)

Widget
   w;      /* Unused */
caddr_t
   data;   /* Unused */
XEvent
   *event; /* Check the expose event counter */

{
   /*
    * If the last expose event is left to be handled,
    * make sure the doglegs are redrawn
    */
   if (event->xexpose.count == 0)
   {
      RedrawLegs(Root, True);
      UpdateTree(Tree, Tree);
   }
   return;
}  




/*
   FUNCTION: 		MovePorthole

   Synopsis: 
       Keep the porthole up to date from the panner widget.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Operating System issues:
       CALL		 PURPOSE

   System files modified:
       none

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

#ifdef R5

static void
MovePorthole(panw, portw, report)

Widget
   panw,
   portw;
XawPannerReport
   *report;

{
   Arg
      args[MAXARGS];
   int
      cnt;
   Widget
      treew;
   
   treew = XtNameToWidget(portw, "*treeframe");
   if (treew == NULL)
   {
      return;
   }
   
   cnt = 0;
   XtSetArg(args[cnt], XtNx, -report->slider_x); cnt++;
   XtSetArg(args[cnt], XtNy, -report->slider_y); cnt++;
   XtSetValues(treew, args, cnt);
   
   return;
}

#endif



/*
   FUNCTION: 		MovePanner

   Synopsis: 
       Update the panner based on a change in the porthole.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE     WHO      REASON

   Operating System issues:
       CALL		 PURPOSE

   System files modified:
       none

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

#ifdef R5

static void
MovePanner(portw, panw, report)

Widget
   portw,
   panw;
XawPannerReport
   *report;

{
   Arg
      args[MAXARGS];
   int
      cnt;
   
   cnt = 0;
   XtSetArg(args[cnt], XtNsliderX, report->slider_x); cnt++;
   XtSetArg(args[cnt], XtNsliderY, report->slider_y); cnt++;
   if (report->changed != (XawPRSliderX | XawPRSliderY))
   {
      XtSetArg(args[cnt], XtNcanvasHeight, report->canvas_height); cnt++;
      XtSetArg(args[cnt], XtNcanvasWidth,  report->canvas_width);  cnt++;
      XtSetArg(args[cnt], XtNsliderHeight, report->slider_height); cnt++;
      XtSetArg(args[cnt], XtNsliderWidth,  report->slider_width);  cnt++;
   }
   XtSetValues(panw, args, cnt);
   return;
}
#endif



/*
   FUNCTION: 		CreateTreeShell

   Synopsis: 
       Create the shell that contains the tree display.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/03/26   cso    Use the borderColor resource for the foreground
       91/05/23   cso    Changed the Shell being used to be a Popup so
                          that font and other information is passed down.
         
   Globals used:
       Toplevel      To access the display pointer

   Globals Modified:
       Gc            The Graphics Context is created.
       Pathgc        The Path Graphics Context is created.
       Treeshell     This widget is created

   Return value:	void

   Range:		none
*/

void
CreateTreeShell()

{
   static char
      dashlst[] = { 2, 3 };
   XGCValues
      values;   /* GC values structure */
   Widget
#ifdef R5
      treepan,
      treeport,
#else
      treeview,  /* Viewport of the treeframe */
#endif
      treeform,
      treeframe, /* Form widget containing tree widgets */
      vertw;
   
   /*
    * Create an independent shell
    * add the viewport and its child the form widget
    * Realize the shell
    * Create the graphics context for drawing lines
    */
   Treeshell = XtVaCreatePopupShell( "Navigation Tree",
       applicationShellWidgetClass, Toplevel,
       XtNallowShellResize, True,
       NULL );

   treeform = XtVaCreateManagedWidget( "treeform",
       groupWidgetClass, Treeshell,
       XtNspacing,  6,
       NULL );

   vertw = (Widget) NULL;

#ifdef R5
   treepan = XtVaCreateManagedWidget( "treepan",
       pannerWidgetClass, treeform,
       XtNalignment, gafill,
       XtNmargin,    6,
       NULL );

   treeport = XtVaCreateManagedWidget( "treeport",
       portholeWidgetClass, treeform,
       XtNpreference, False,
/*
       XtNprefWidth,  600,
       XtNminWidth,   100,
       XtNprefHeight, 300,
       XtNminHeight,  50,
*/
       XtNalignment,  gaFill,
       XtNstretch,    True,
       XtNmargin,     6,
       NULL );
   
   XtAddCallback(treeport, XtNreportCallback, MovePanner,   treepan);
   XtAddCallback(treepan,  XtNreportCallback, MovePorthole, treeport);
   
#else
   treeview = XtVaCreateManagedWidget( "treeview",
       viewportWidgetClass, treeform,
       XtNallowHoriz, True,
       XtNallowVert,  True,
       XtNforceBars,  True,
       XtNpreference, False,
/*
       XtNprefWidth,  650,
       XtNminWidth,   100,
       XtNprefHeight, 250,
       XtNminHeight,  50,
*/
       XtNalignment,  gaFill,
       XtNstretch,    True,
       XtNmargin,     6,
       NULL );
#endif

#ifdef R5
   treeframe = XtVaCreateManagedWidget( "treeframe",
       viewformWidgetClass, treeport
       XtNdefaultDistance, DEFAULTSPACE,
       NULL );

#else
   treeframe = XtVaCreateManagedWidget( "treeframe",
       viewformWidgetClass, treeview,
       XtNdefaultDistance, DEFAULTSPACE,
       NULL );
#endif
   XtAddEventHandler(treeframe, ExposureMask, False, RefreshTree, NULL);
   
   XtPopup(Treeshell, XtGrabNone);
   
   XtVaGetValues( treeframe,
       XtNborderColor, &values.foreground,
       XtNbackground,  &values.background,
       NULL );
   values.line_width = 2;
   values.fill_style = FillSolid;
   Gc = XtGetGC(treeframe, 
                (GCLineWidth  | GCFillStyle | GCForeground | GCBackground),
		&values);

   values.line_width = 3;
   values.foreground = values.background;
   values.line_style = LineOnOffDash;
   Pathgc = XtGetGC(treeframe, 
		    (GCLineWidth  | GCFillStyle | GCLineStyle | 
		     GCForeground | GCBackground),
		    &values);
   XSetDashes(XtDisplay(treeframe), Pathgc, -2, dashlst, 2);
   
   return;
}
