/* CONTENTS
   Owner:       Copyright 1990 The Ohio State University,
                Integrated Chameleon Architecture
   Module:      $RCSfile: utils.c,v $
   Purpose:
       To provide functions for handling the node data structure.

   Programmer:  Conleth O'Connell
   Reviewer:    
   Revision:    $Revision: 1.12 $
   Date:        $Date: 92/05/29 14:46:52 $

   Functions Index:
   AVAIL   NAME DEPENDS ON INTERNALLY      EXTERNALLY
   -----   ----            ----------      ----------
   EXPORT  Pop

   EXPORT  Push

   EXPORT  MakeUpper

   EXPORT  NewNode

   EXPORT  NewAlt
                                           NewNode
   EXPORT  NewAux

   EXPORT  FindAlt

   EXPORT  FindNodeW

   EXPORT  FindOccurW

   EXPORT  FindTypeW

   EXPORT  HideSubtree

   EXPORT  MakeSubtree
                           NewAlt
			   NewNode
			                   SensitizeAltButtons
   HIDDEN  UpdateW
                                           CreateNodeW
					   CreateTypeW
					   CreateOccurW
					   PutLabel
   HIDDEN  DisplaySubtree
                           FindAlt
			   UpdateW
			                   IsTerminal
					   PutLabel
					   SensitizeAltButtons
   EXPORT  AddSymbol
                                           ChangeSymbol
					   InsertDefItem
					   InsertUndItem
					   IsDefined
					   PutDefinition
					   PutString
					   PutSymbol
					   PutSymbolInTree
					   PutSymbolType
					   RemoveDefItem
					   RemoveUndItem
   HIDDEN  ReconcileWidgets

   EXPORT  KillNodeWidgets
                           ReconcileWidgets
   EXPORT  KillWidgets
                           KillNodeWidgets
			   KillWidgets
   HIDDEN  AdjustHorizontal
                                           GetDimensions
   HIDDEN  AdjustVertical
                                           GetDimensions
   HIDDEN  AdjustScrollBars
                           AdjustHorizontal
			   AdjustVertical
   EXPORT  UpdateInterface
                           AdjustScrollBars
			   DisplaySubtree
			   HideSubtree
			                   UpdateAux
					   UpdateCompass
					   UpdateTree
*/

/* HISTORY
   $Log:	utils.c,v $
 * Revision 1.12  92/05/29  14:46:52  boyd
 * Changes for improved resizing, consisting mostly of replacing Form widgets
 * with Group, Table, and Viewform widgets.
 * 
 * 
 * Revision 1.11  92/03/18  14:18:00  ware
 * Working on consistent interface.
 * 
 * Revision 2.4  92/03/06  17:46:14  boyd
 * Changed production updating so that widgets don't "bunch up" in the corner;
 * Required change to AdjustHorizontal as well, which was using Form's
 * horizDistance for widgets which are now children of a Group...
 * 
 * Revision 2.3  92/03/05  17:40:42  boyd
 * Changes associated with boxing text fields in production:
 *  . UpdateInterface handles nodeframe differently, since it is now a
 *    Group widget.
 * 
 * Revision 1.10  1992/02/20  04:43:53  cso
 * Fixed a couple of minor problems with changing (editing) a symbol's
 * name, and ensured that the proper allocation and freeing of strings
 * occurred.
 *
 * Revision 1.9  1992/02/11  16:24:51  cso
 * Changed the focus feedback to a switch of bg/fg colors and caret display
 *
 * Revision 1.8  1992/01/28  15:44:21  cso
 * Removed the use of left and right alts - they weren't being used anyway.
 *
 * Revision 1.7  1992/01/14  20:27:09  cso
 *  - Reorganizing header files.
 *  - Fixed a typo that was unmanaging a different pointer's widget
 * (nptr->qptr).
 *  - Globalized keyboard focus.
 *
 * Revision 1.6  1992/01/03  19:41:23  cso
 * HideSubtree:
 * 	Reset borderwidth to 1 for all nodes being unmanaged (because
 * the thicker borderwidths are showing keyboard focus).
 *
 * AddSymbol:
 * 	Moved the calls to insert a symbol into the interface lists.
 *
 * UpdateInterface:
 * 	Using borderwidth to show keyboard focus and delaying layout
 * changes until after everything is settled to help limit the amount of
 * "blinking" in the node form widget.
 *
 * Revision 1.5  1991/11/22  15:35:40  cso
 * 	Added a graphics context to the node type for easier
 * manipualtion of the line drawing in the navigation tree.  Tried to
 * eliminate the number of refreshes occurring for a particular expose
 * event, tough to debug in synchronous mode.
 *
 * Revision 1.4  1991/11/20  00:05:47  cso
 * Reorganized the interface by removing borders from around widgets and
 * relocating widgets.
 * Added a status widget to display info that was being output to stderr.
 * The xterm used to instantiate DeveGram is no longer required b/c of
 * this widget.
 * Created a couple of constants for maximum arguments used and width of
 * the interface.
 *
 * Revision 1.3  1991/11/18  23:47:35  cso
 * Added the Release 5 panner and porthole widgets to replace the
 * scrollbars on the navigation tree window.
 * They are compile time usable by using -DR5 in the CFLAGS.
 *
 * Revision 1.2  1991/11/14  14:31:29  cso
 * types.h utils.c utils.h CVS: CVS: Enter Log.  Lines beginning with
 * 'CVS: ' are removed automatically CVS:
 * ----------------------------------------------------------------------
 * Made the necessary additions and changes to provide a mechanism to
 * change symbol types between nonterminal types Nonterminal & Grouping.
 * In SGML terms, therefore, intermediate form terms, this is the
 * difference between having element tags and not.
 *
 * Revision 1.1.1.3  1991/10/04  20:17:01  jbm
 * First HaL mods to ICA
 *
 * Revision 2.1  91/05/28  10:18:45  cso
 * Added a couple of explicit returns.
 * Removed Chosen_Symbol.
 * 
 * Revision 2.0  91/03/30  13:53:53  cso
 * This module is now in Alpha Release form.
 * 
 * Revision 1.20  91/03/28  14:50:22  cso
 * Added the focus setting for getting symbols from a user.
 * 
 * Revision 1.19  91/02/22  10:05:06  cso
 * Completed documentation.
 * Changed the use of Pathndx to be like Ndx.
 * 
 * Revision 1.18  90/11/29  14:28:59  cso
 * Cosmetics.
 * Added functions for updating the scrollbars as the user works on the 
 * grammar.
 * 
 * Revision 1.17  90/11/26  13:22:44  cso
 * Placed Get, Put, and Remove routines in their own modules.
 * Thinking about a boolean module or is module for all of the Is routines.
 * That may eliminate node.c.
 * 
 * Revision 1.16  90/11/20  15:59:59  cso
 * Cosmetic changes.
 * Changes from lint.
 * Added the functions for getting Pixel values for offsets and sizes.
 * Added FindOccurW.
 * Modified AddSymbol to use vertical in calls, and to change a symbol 
 * correctly.
 * Added ReconcileWidgets (for sibling relationships.).
 * Added KillNodeWidgets to destroy one set of widgets at a time.
 * Added the appropriate remove routines for Delete and Erase.
 * Added RemoveNUllSymbols routine.
 * Do I need a get.c put.c remove.c change.c modules?
 * 
 * Revision 1.15  90/11/05  18:32:11  cso
 * Added function template to all functions, but still need to fill it in.
 * Fixed Push!  Removed setting the right part of Node for now.  Changed
 * FindNodeW so that it doesn't call FindAlt.  Added an UpdateW to pare
 * down DisplaySubtree.  Still need to fix Remove functions.
 * Else, cosmetics.
 * 
 * Revision 1.14  90/10/29  10:30:48  cso
 * Removed some SetSensitive calls.
 * 
 * Revision 1.13  90/10/25  13:36:14  cso
 * Trying to trace down NULL alts after a deletion occurs.
 * 
 * Revision 1.12  90/10/24  15:40:02  cso
 * Used XtNew instead of my own malloc.  Also, other changes based on the 
 * new types in types.h.
 * 
 * Revision 1.11  90/10/04  16:07:14  cso
 * I added PutString for changing the value in a nodew, and I use it instead 
 * of Destroying and recreating widgets.
 * 
 * Revision 1.10  90/10/02  13:40:25  cso
 * Added managing widgets instead of destroying and recreating widgets.
 * Added Unset selection routine also.
 * 
 * Revision 1.9  90/09/07  09:42:13  cso
 * Use of NODE routines and cosmetic changes.
 * 
 * Revision 1.8  90/09/03  18:30:41  cso
 * Added left and right to NewNode.  Still need updat routines for
 * these additional ptrs.
 * 
 * Revision 1.7  90/09/03  14:43:49  cso
 * Moved CreateOccurW to alt.c.  Changed function names to be 
 * consistent.
 * 
 * 
 * Revision 1.6  90/08/29  17:52:40  cso
 * Added GetOccurValue and GetSymbolType routines to get attributes via
 * popup shells.
 * 
 * Revision 1.5  90/08/26  17:44:15  cso
 * Comment all calls RE: auxiliary information.
 * 
 * Revision 1.4  90/08/26  14:13:45  cso
 * added module template
 * 
*/

/* #INCLUDES */

#include "std.h"
#include "widgets.h"
#include "types.h"
#include "defs.h"

#include "alt.h"
#include "compass.h"
#include "deflist.h"
#include "display.h"
#include "get.h"
#include "interface.h"
#include "main.h"
#include "node.h"
#include "put.h"
#include "status.h"
#include "symtab.h"
#include "tree.h"
#include "undlist.h"

extern Aux
   *Auxtree;
/*
#include "aux.h"
*/



/*
   FUNCTION: 		Pop

   Synopsis: 
       Pop a node off of a path mechanism.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Pathlist  To access the list of path nodes
       Pathndx   To access a specific node in the list

   Globals Modified:
       Pathlist  The top item of the stack is removed
       Pathndx   Decremented

   Return value:	Node *

   Range:		none
*/

Node
*Pop()

{
   /*
    * If there are items in the list pop one off
    */
   if (Pathndx > 0)
   {
      /*
       * Decrement the Top location
       * Return the top item off the list
       */
      Pathndx--;
      return(Pathlist[Pathndx]);
   }
   return(NULL);
}




/*
   FUNCTION: 		Push

   Synopsis: 
       Push a node onto the stack (path mechanism).

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       Pathlist  To access the list stack
       Pathndx   The position of the top item in the stack

   Globals Modified:
       Pathlist  A new node is added
       Pathndx   Increment index

   Return value:	void

   Range:		none
*/

void
Push(nptr)

Node
   *nptr;  /* The node being pushed onto the stack */

{
   /*
    * Put the node onto the stack
    * Increment the index (always points to the next available spot)
    */
   Pathlist[Pathndx] = nptr;
   Pathndx++;
   return;
}




/*
   FUNCTION: 		MakeUpper

   Synopsis: 
       Make a string into the uppercase equivalent.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	String

   Range:		An upper case version of the input string
*/

String
MakeUpper(str)

String
   str;  /* The string being transformed */

{
   String
      tmp; /* A temporary copy */
   int
      i;   /* Loop counter to examine at each character */
   
   /*
    * Make a copy of the string
    * For each character, consider if it should be changed
    */
   tmp = XtNewString(str);
   for (i = 0; i < strlen(tmp); i++)
   {
      /*
       * If the character is a lower case alphabetic character
       * Make it upper case
       */
      if (islower(tmp[i]))
      {
	 tmp[i] = toupper(tmp[i]);
      }
   }
   return(tmp);
}




/*
   FUNCTION: 		NewNode

   Synopsis: 
       Create a new node and return a pointer to it.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       Dirtyflag   Has been set because a new node is created

   Return value:	Node *

   Range:		none
*/

Node
*NewNode(parent, sibling)

Node
   *parent,  /* The parent of this node in the tree */
   *sibling; /* The sibling previous to the new node */

{
   Node
      *tmpnode; /* A temporary pointer to the new node */
   
   /*
    * Set the Dirtyflag because a new node is being created
    * Use X to allocate space and create a data structure of a new Node
    * If the pointer is not NULL, inform the user.
    * Initialize the data structure
    * Return the pointer to the new node
    */
   Dirtyflag = True;
   tmpnode    = XtNew(Node);
   if (tmpnode == NULL)
   {
      fprintf(stderr, "Ran out of memory - could not create new Node\n");
      exit(1);
   }
   tmpnode->symbol     = "";
   tmpnode->occurrence = XtNewString(" ");
   tmpnode->parent     = parent;
   tmpnode->next       = NULL;
   tmpnode->prev       = sibling;
   tmpnode->alts       = NULL;
   tmpnode->nodew      = NULL;
   tmpnode->occurw     = NULL;
   tmpnode->typew      = NULL;
   tmpnode->treew      = NULL;
   tmpnode->treegc     = Gc;
   return(tmpnode);
}




/*
   FUNCTION: 		NewAlt

   Synopsis: 
       Create a new Alternative definition structure and return a pointer
       to it.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       Dirtyflag       Set because a new structure is created

   Return value:	Alt *

   Range:		none
*/

Alt
*NewAlt(whom, sibling)

Node
   *whom;    /* The node that this definition will belong */
Alt
   *sibling; /* A sibling alternative previous to this new one */

{
   Alt
      *tmpalt; /* A temporary pointer that is returned */
   
   /*
    * Set the dirty_flag because a new structure is being created
    * Allocate space for the creation of the structure
    * If the pointer is NULL, inform the user
    * Initialize the data structure (includes creating a new node)
    */
   Dirtyflag = True;
   tmpalt = XtNew(Alt);
   if (tmpalt == NULL)
   {
      fprintf(stderr, "Ran out of memory - could not create new Alt\n");
      exit(1);
   }
   tmpalt->whom    = whom;
   tmpalt->prod    = NewNode(whom, (Node *) NULL);
   tmpalt->next    = NULL;
   tmpalt->prev    = sibling;
   tmpalt->exposed = True;
   tmpalt->auxinf  = NULL;
   return(tmpalt);
}




/*
   FUNCTION: 		NewAux

   Synopsis: 
       Allocate space and create a new structure for auxiliary information.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Aux *

   Range:		none
*/

Aux
*NewAux(aptr)

Alt
   *aptr;      /* The definition that needs this new auxiliary information */

{
   Aux
      *tmpaux; /* A pointer to the new aux structure being returned */
   
   /*
    * Get a pointer to a new structure
    * If the pointer is NULL, inform the user
    * Initialize the structure
    * Return the pointer
    */
   tmpaux = XtNew(Aux);
   if (tmpaux == NULL)
   {
      fprintf(stderr, "Ran out of memory - could not create new Aux\n");
      exit(1);
   }
   tmpaux->which = aptr;
   tmpaux->type  = "";
   tmpaux->info  = "";
   tmpaux->next  = NULL;
   tmpaux->prev  = Auxtree;
   return(tmpaux);
}




/*
   FUNCTION: 		FindAlt

   Synopsis: 
       Find the definition containing a node.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Alt *

   Range:		none
*/

Alt
*FindAlt(nptr)

Node
   *nptr;     /* The node whose definition is being searched */

{
   Alt
      *aptr;  /* Loop pointer to find the definition containing the node */
   
   /*
    * If the node does not exists, or if it is the Root, return NULL
    * Else find the definition
    */
   if ((nptr == NULL) || (IsRoot(nptr) == True))
   {
      return(NULL);
   }
   else
   {
      /*
       * Move the node pointer to the head of the list in a production
       */
      for ( ; nptr->prev != NULL; nptr = nptr->prev)
      {
      }
#ifdef DEBUG
      fprintf(stderr, "nptr = [%s]\n", nptr->symbol);
#endif

      /*
       * For each definition of the node's parent, look for the definition
       * If the definition pointer finds the head of the production list,
       * the definition is found
       * 
       */
      for (aptr = nptr->parent->alts;
	   ((aptr != NULL) && (aptr->prod != nptr));
	   aptr = aptr->next)
      {
#ifdef DEBUG
	 fprintf(stderr, "aptr->prod = [%s]\n", aptr->prod->symbol);
#endif
      }
#ifdef DEBUG
      if (aptr == NULL)
      {
	 fprintf(stderr, "aptr being returned is NULL\n");
      }
#endif   
    
      /*
       * Return the definition pointer
       */
      return(aptr);
   }
}




/*
   FUNCTION: 		FindNodeW

   Synopsis: 
       Find the node whose widget has been selected by the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		none
*/

Node
*FindNodeW(w, nptr)

Widget
   w;      /* Widget selected by the user */
Node
   *nptr;  /* A reference node in the same definition as the widget */

{
   Node
      *qptr; /* Loop pointer to find the node whose widget was selected */
   
   /*
    * From the reference node to previous nodes, check for the widget
    * If found, return the node pointer
    * From the reference node to next nodes, check for the widget
    * If found, return the node pointer
    * Otherwise there is an error.
    */
   for (qptr = nptr; qptr != NULL; qptr = qptr->prev)
   {
      if (qptr->nodew == w)
      {
	 return(qptr);
      }
   }
   for (qptr = nptr; qptr != NULL; qptr = qptr->next)
   {
      if (qptr->nodew == w)
      {
	 return(qptr);
      }
   }
   fprintf(stderr, "FindNodeW: Data structure corrupted.\n");
   exit(5);
   return(NULL);
}




/*
   FUNCTION: 		FindOccurW

   Synopsis: 
       Find the node whose occurrence widget has been selected by the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		none
*/

Node
*FindOccurW(w, nptr)

Widget
   w;     /* Widget selected by the user */
Node
   *nptr; /* Reference node to search for the occurrence widget */

{
   Node
      *qptr; /* Loop pointer to look for the widget */
   
   /*
    * For each node to previous ones, if the widgets match
    * Return the node
    * For each node to next ones, if the widgets match
    * Return the node
    * Inform the user if the data structure might be corrupted
    */
   for (qptr = nptr; qptr != NULL; qptr = qptr->prev)
   {
      if (qptr->occurw == w)
      {
	 return(qptr);
      }
   }
   for (qptr = nptr; qptr != NULL; qptr = qptr->next)
   {
      if (qptr->occurw == w)
      {
	 return(qptr);
      }
   }
   fprintf(stderr, "FindOccurW: Data structure corrupted.\n");
   exit(5);
   return(NULL);
}




/*
   FUNCTION: 		FindTypeW

   Synopsis: 
       Find the node whose type widget has been selected by the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	Node *

   Range:		none
*/

Node
*FindTypeW(w, nptr)

Widget
   w;     /* Widget selected by the user */
Node
   *nptr; /* Reference node to search for the occurrence widget */

{
   Node
      *qptr; /* Loop pointer to look for the widget */
   
   /*
    * For each node to previous ones, if the widgets match
    * Return the node
    * For each node to next ones, if the widgets match
    * Return the node
    * Inform the user if the data structure might be corrupted
    */
   for (qptr = nptr; qptr != NULL; qptr = qptr->prev)
   {
      if (qptr->typew == w)
      {
	 return(qptr);
      }
   }
   for (qptr = nptr; qptr != NULL; qptr = qptr->next)
   {
      if (qptr->typew == w)
      {
	 return(qptr);
      }
   }
   fprintf(stderr, "FindTypeW: Data structure corrupted.\n");
   exit(5);
   return(NULL);
}




/*
   FUNCTION: 		HideSubtree

   Synopsis: 
       Hide the widgets making up the production that is currently displayed,
       i.e., unmanage the widgets.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
HideSubtree(nptr)

Node
   *nptr;    /* Reference node for hiding the whole production */

{
   Arg
      args[MAXARGS];
   int
      cnt;
   Node
      *qptr; /* Loop pointer to visit each sibling of nptr */
   
   /*
    * If the node is the Root, hide the Root's node widget
    * Else hide the production.
    */
   if (IsRoot(nptr) == True)
   {
      if (nptr->nodew != NULL)
      {
	 XawTextUnsetSelection(nptr->nodew);
	 XtUnmanageChild(nptr->nodew);
	 cnt = 0;
	 XtSetArg(args[cnt], XtNdisplayCaret, False);      cnt++;
	 XtSetArg(args[cnt], XtNborderWidth,  1);          cnt++;
	 XtSetValues(nptr->nodew, args, cnt);
      }
   }
   else
   {
      /*
       * For each sibling before the reference node
       * hide node and occurrence widgets
       * For each sibling after the reference node
       * hide node and occurrence widgets
       */
      for (qptr = nptr; qptr != NULL; qptr = qptr->prev)
      {
	 if (qptr->occurw != NULL)
	 {
	    XtUnmanageChild(qptr->occurw);
	 }
	 if (qptr->typew != NULL)
	 {
	    XtUnmanageChild(qptr->typew);
	 }
	 if (qptr->nodew != NULL)
	 {
	    XawTextUnsetSelection(qptr->nodew);
	    XtUnmanageChild(qptr->nodew);
	    cnt = 0;
	    XtSetArg(args[cnt], XtNdisplayCaret, False);      cnt++;
	    XtSetArg(args[cnt], XtNborderWidth,  1);          cnt++;
	    XtSetValues(qptr->nodew, args, cnt);
	 }
      }
      for (qptr = nptr; qptr != NULL; qptr = qptr->next)
      {
	 if (qptr->occurw != NULL)
	 {
	    XtUnmanageChild(qptr->occurw);
	 }
	 if (qptr->typew != NULL)
	 {
	    XtUnmanageChild(qptr->typew);
	 }
	 if (qptr->nodew != NULL)
	 {
	    XawTextUnsetSelection(qptr->nodew);
	    XtUnmanageChild(qptr->nodew);
	    cnt = 0;
	    XtSetArg(args[cnt], XtNdisplayCaret, False);      cnt++;
	    XtSetArg(args[cnt], XtNborderWidth,  1);          cnt++;
	    XtSetValues(qptr->nodew, args, cnt);
	 }
      }
   }
   return;
}




/*
   FUNCTION: 		MakeSubtree

   Synopsis: 
       If a node has no subtree, create one.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
MakeSubtree(nptr)

Node
   *nptr;  /* Node requiring a subtree created */

{
   /*
    * If the definition pointer is NULL, create a new one
    * If no nodes exist in the definition, create one
    * Sensitize the previous and next definition buttons
    */
   if (nptr->alts == NULL)
   {
      nptr->alts = NewAlt(nptr, (Alt *) NULL);
   }
   if (nptr->alts->prod == NULL)
   {
      nptr->alts->prod = NewNode(nptr, (Node *) NULL);
   }
   SensitizeAltButtons(nptr->alts);
   return;
}




/*
   FUNCTION: 		UpdateW

   Synopsis: 
       Update the node and occurrence widgets in a production so that after
       an insertion or deletion, the resources are correct.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
UpdateW(nptr, wtype, horiz)

Node
   *nptr;  /* Node being updated */
char
    wtype; /* The type of widget being updated */
Widget
    horiz; /* The widget that should be placed in the fromHoriz resource */

{
   Arg
      args[MAXARGS];/* List of arguments */
   int
      cnt;      /* Number of arguments */
   SymType
      stype;
   
   /*
    * Depending on the type of widget
    * Ensure that it placed correctly
    */
   switch (wtype)
   {
   default:
      XicaStatus(Statw, "UpdateW passed wrong widget type [%c]\n", wtype);
      break;

   case 'n':
      /*
       * For a node widget,
       * If it is NULL, create a new one with the proper
       * resources
       * Else set the resources
       */
      if (nptr->nodew == NULL)
      {
	 nptr->nodew = CreateNodeW(horiz, nptr->symbol);
      }
      else
      {
	 XtUnmanageChild(nptr->nodew);
	 cnt = 0;
	 XtSetArg(args[cnt], XtNfromHoriz, horiz); cnt++;
	 XtSetValues(nptr->nodew, args, cnt);
	 XtManageChild(nptr->nodew);
      }
      break;

   case 'o':
      /*
       * For an occurrence widget, 
       * If it is NULL, create a new one with the proper
       * resources and place the occurrence value in the widget
       * If an occurrence exist, set the occurw sensitive
       * Else set the resources
       */
      if (nptr->occurw == NULL)
      {
	 nptr->occurw = CreateOccurW(horiz);
	 PutLabel(nptr->occurw, nptr->occurrence);
	 if ((*nptr->occurrence != '\0') && (*nptr->occurrence != ' '))
	 {
	    XtSetSensitive(nptr->occurw, True);
	 }
      }
      else
      {
	 XtUnmanageChild(nptr->occurw);
	 cnt = 0;
	 XtSetArg(args[cnt], XtNfromHoriz, horiz); cnt++;
	 XtSetValues(nptr->occurw, args, cnt);
	 XtManageChild(nptr->occurw);
      }
      break;

   case 't':
      /*
       * For a type widget, 
       * If it is NULL, create a new one with the proper
       * resources and place the symbol type in the widget
       * If a type exists, set the typew sensitive
       * Else set the resources
       */
      stype = GetSymbolType(nptr->symbol);
      if (nptr->typew == NULL)
      {
	 nptr->typew = CreateTypeW(horiz);
	 switch (stype)
	 {
	 default:
	    break;
	    
	 case Unknown:
	    PutLabel(nptr->typew, " ");
	    break;
	    
	 case Part:
	    PutLabel(nptr->typew, "N");
	    break;
	    
	 case Group:
	    PutLabel(nptr->typew, "G");
	    break;
	    
	 case File:
	    PutLabel(nptr->typew, "F");
	    break;
	 }
	 if (stype != Unknown)
	 {
	    XtSetSensitive(nptr->typew, True);
	 }
      }
      else
      {
	 XtUnmanageChild(nptr->typew);
	 cnt = 0;
	 XtSetArg(args[cnt], XtNfromHoriz, horiz); cnt++;
	 XtSetValues(nptr->typew, args, cnt);
	 XtManageChild(nptr->typew);
      }
      break;
   }
}




/*
   FUNCTION: 		DisplaySubtree

   Synopsis: 
       Display a definition for a node.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
DisplaySubtree(nptr)

Node
   *nptr;     /* A reference node for displaying the occurrence it is a
		 part of */

{
   Alt
      *aptr;  /* Pointer to the definition containing the node */
   Node
      *qptr;  /* Loop pointer to visit each node in a definition */
   
   /*
    * Get the definition pointer containing the reference node
    * Sensitize the definition buttons
    * If the definition pointer is NULL, assume that means the node is the Root
    * Assign nptr to qptr
    * Else assign the first node in a production to qptr
    */
   aptr = FindAlt(nptr);
   SensitizeAltButtons(aptr);
   if (aptr == NULL)
   {
      qptr = nptr;
   }
   else
   {
      qptr = aptr->prod;
   }
   
   /*
    * Put the parent's symbol in the focus widget, "" in the case of Root
    */
   if (qptr != NULL)
   {
      if (IsRoot(qptr) == False)
      {
	 PutLabel(XtNameToWidget(Frame, "altframe*focus"),
		  qptr->parent->symbol);
      }
      else
      {
	 PutLabel(XtNameToWidget(Frame, "altframe*focus"), "");
      }
   }

   /*
    * For each node in the definition being displayed
    * Update the node, type, and occurrence widgets
    */
   for ( ; qptr != NULL; qptr = qptr->next)
   {
      /*
       * If the node widget is the first, update it to be from NULL
       * Else If a previous sibling has an occurrence widget, use it
       * Else use the previous sibling's node widget
       */
      if (IsFirstNode(qptr) == True)
      {
	 UpdateW(qptr, 'n', (Widget) NULL);
      }
      else if (qptr->prev->occurw != NULL)
      {
	 UpdateW(qptr, 'n', qptr->prev->occurw);
      }
      else
      {
	 UpdateW(qptr, 'n', qptr->prev->nodew);
      }
      
      /*
       * If the node is not Root and not a terminal,
       * Update the occurrence widget
       */
      if ((IsRoot(qptr) == False) && (IsTerminal(qptr->symbol) != TriT))
      {
	 UpdateW(qptr, 't', qptr->nodew);
	 UpdateW(qptr, 'o', qptr->typew);
      }
   }
   return;
}




/*
   FUNCTION: 		AddSymbol

   Synopsis: 
       Add a symbol to the tree display, node data structure
       (actually just a matter of changing a label), symbol table and
       appropriate symbol lists.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
AddSymbol(nptr, str, type, loadflag)

Node
   *nptr;     /* Node under consideration */
String
   str;       /* The symbol string being added */
SymType
   type;      /* The symbol type */
Boolean
   loadflag;  /* Whether or not a file is being loaded */

{
   String
      oldstr;
   
   /*
    * If no symbol exist already in that node location
    * Add the symbol as being new
    * Else treat it as an edited symbol
    */
   if (*nptr->symbol == '\0')
   {
      /*
       * Update the node structure to use the symbol string
       * If a node widget exist, put the string in the widget
       * Put the symbol in the symbol table
       * Put the symbol type in the symbol table
       * Put the node in the nodelist of definitions where this symbol occurs
       * Insert the symbol in the available symbols list
       */
      nptr->symbol = XtNewString(str);
      if (nptr->nodew != NULL)
      {
	 PutString(nptr->nodew, nptr->symbol);
      }
      PutSymbol(nptr->symbol);
      PutDefinition(nptr);
      PutSymbolType(nptr->symbol, type);
      PutSymbolInTree(nptr, loadflag);
   }
   else if (strcmp(nptr->symbol, str) != 0)
   {
      /*
       * Remove the symbol from the symbol lists
       * Change the symbol in the node and table structures
       * Insert the symbol in the appropriate lists
       */
      oldstr       = XtNewString(nptr->symbol);
      XtFree(nptr->symbol);
      nptr->symbol = XtNewString(str);
      
      RemoveDefItem(oldstr);
      RemoveUndItem(oldstr);
      if (ChangeSymbol(nptr, oldstr) == False)
      {
	 XtFree(nptr->symbol);
	 nptr->symbol = XtNewString(oldstr);
      }
      XtFree(oldstr);
   }
   InsertDefItem(nptr->symbol);
   if (IsDefined(nptr->symbol) == False)
   {
      InsertUndItem(nptr->symbol);
   }
   return;
}




/*
   FUNCTION: 		ReconcileWidgets

   Synopsis: 
       Reconcile widgets in the tree display because one the tree
       widgets is being destroyed.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
ReconcileWidgets(nptr, vertical)

Node
   *nptr;       /* Node under consideration */
Boolean
   vertical;    /* Orientation of the widgets */

{
   Arg
      args[MAXARGS];/* List of arguments */
   int
      cnt;      /* Number of arguments */
   Node
      *qptr;    /* Sibling of nptr */
   Widget
      fromw;    /* Widget the is used for the fromHoriz resource */

   /*
    * If the node under consideration nor its next sibling are NULL
    * reconcile the resources
    */
   if ((nptr != NULL) && (nptr->next != NULL))
   {
      /*
       * Set qptr to be the next sibling
       * If both node widgets are not NULL
       * reset the sibling's resources
       * If both tree widgets are not NULL
       * reset the sibling's resources
       */
      qptr = nptr->next;
      if ((nptr->nodew != NULL) && (qptr->nodew != NULL))
      {
	 /*
	  * Get the fromHoriz resource for the node widget
	  * Assign the same value to the sibling widget
	  */
	 cnt = 0;
	 XtSetArg(args[cnt], XtNfromHoriz, &fromw); cnt++;
	 XtGetValues(nptr->nodew, args, cnt);

	 XtUnmanageChild(qptr->nodew);
	 cnt = 0;
	 XtSetArg(args[cnt], XtNfromHoriz, fromw);  cnt++;
	 XtSetValues(qptr->nodew, args, cnt);
	 XtManageChild(qptr->nodew);
      }
      if ((nptr->treew != NULL) && (qptr->treew != NULL))
      {
	 /*
	  * Depending on the orientation
	  * Get the appropriate resource for the tree widget
	  * Assign the same value to the sibling widget
	  */
	 cnt = 0;
	 if (vertical == True)
	 {
	    XtSetArg(args[cnt], XtNfromHoriz, &fromw); cnt++;
	 }
	 else
	 {
	    XtSetArg(args[cnt], XtNfromVert,  &fromw); cnt++;
	 }
	 XtGetValues(nptr->treew, args, cnt);

	 XtUnmanageChild(qptr->treew);
	 cnt = 0;
	 if (vertical == True)
	 {
	    XtSetArg(args[cnt], XtNfromHoriz, fromw); cnt++;
	 }
	 else
	 {
	    XtSetArg(args[cnt], XtNfromVert,  fromw); cnt++;
	 }
	 XtSetValues(qptr->treew, args, cnt);
	 XtManageChild(qptr->treew);
      }
   }
}




/*
   FUNCTION: 		KillNodeWidgets

   Synopsis: 
       Destroy the widgets for the node.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
KillNodeWidgets(nptr, vertical)

Node
   *nptr;    /* Node under consideration */
Boolean
   vertical; /* Display orientation */

{
   /*
    * If the node was not NULL
    * Attempt to reconcile the widgets for a node and its sibling
    * Destroy its widgets
    */
   if (nptr != NULL)
   {	
      ReconcileWidgets(nptr, vertical);
      if (nptr->occurw != NULL)
      {
	 XtDestroyWidget(nptr->occurw);
	 nptr->occurw = NULL;
      }
      if (nptr->typew != NULL)
      {
	 XtDestroyWidget(nptr->typew);
	 nptr->typew = NULL;
      }
      if (nptr->nodew != NULL)
      {
	 XtDestroyWidget(nptr->nodew);
	 nptr->nodew = NULL;
      }
      if (nptr->treew != NULL)
      {
	 XtDestroyWidget(nptr->treew);
	 nptr->treew = NULL;
      }
   }
   return;
}




/*
   FUNCTION: 		KillWidgets

   Synopsis: 
       In a postfix manner, destroy a subtrees' widgets.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	void

   Range:		none
*/

void
KillWidgets(nptr)

Node
   *nptr;  /* Node whose subtree is to be destroyed */

{
   Alt
      *aptr;   /* Loop pointer to visit each definition under the node */
   Boolean
      vertical = True;  /* Display orientation */
   Node
      *qptr;   /* Loop pointer to visit each node in a definition */
   
   /*
    * If the node exists, visit in a postfix order all of its children
    * and destroy their widgets
    */
   if (nptr != NULL)
   {	
      for (aptr = nptr->alts; aptr != NULL; aptr = aptr->next)
      {
	 for (qptr = aptr->prod; qptr != NULL; qptr = qptr->next)
	 {
	    KillWidgets(qptr);
	 }
      }
      KillNodeWidgets(nptr, vertical);
   }
   return;
}




/*
   FUNCTION: 		AdjustHorizontal

   Synopsis: 
       Adjust the view shown the user in terms of the horizontal scrollbar
       and data.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
AdjustHorizontal(formw, horizw, nodew, altflag)

Widget
   formw,    /* Form widget containing the data being viewed */
   horizw,   /* The horizontal scrollbar widget being adjusted */
   nodew;    /* The node widget under consideration to be in view */
Boolean
   altflag;  /* Boolean to indicate if it is the definition scrollbar
		or the tree display scrollbar */

{
   Arg
      args[MAXARGS];/* List of arguments */
   Dimension
      buffer,     /* Amount of space to the right of the node widget */
      coverage,   /* Amount of the data from the left margin */
      height,     /* Used for GetDimensions() */
      length,     /* Length of the scrollbar */
      form_width, /* Width of the form widget */
      node_width, /* Width of the node needing to be viewed */
      spacing;    /* Group spacing, used for node_hdist */
   float
      new_top,    /* New top of thumb location */
      shown,      /* Percentage of the form being viewed */
      top;        /* Location where the view begins */
   int
      cnt,        /* Number of arguments */
      node_hdist, /* Horizontal distance to the node */
      hdist,      /* Used for GetDimensions() */
      vdist;      /* Used for GetDimensions() */
   Position
      x,          /* Used for GetDimensions() */
      new_x,      /* The new X coordinate for the form widget */
      node_x,     /* X coordinate for the node */
      y,          /* Used for GetDimensions() */
      form_y;     /* Y coordinate for the form widget */

   /*
    * Get values of the horizontal scrollbar
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNlength,     &length); cnt++;
   XtSetArg(args[cnt], XtNshown,      &shown);  cnt++;
   XtSetArg(args[cnt], XtNtopOfThumb, &top);    cnt++;
   XtGetValues(horizw, args, cnt);
#ifdef DEBUG
   fprintf(stderr, "Horizontal scrollbar:\n");
   fprintf(stderr, "\tlength = [%d], shown = [%7.4f], top = [%7.4f]\n",
	   length, shown, top);
#endif

   /*
    * If the amount shown is less than all of the data, 
    * Recalculate scrollbar and form widget attributes
    */
   if (shown < 1.0)
   {
      /*
       * Get the dimensions of the form and node widgets
       */
      GetDimensions(formw,  True, &x, &form_y, &form_width, &height,
		    &hdist, &vdist);
      GetDimensions(nodew,  True, &node_x, &y, &node_width, &height,
		    &node_hdist, &vdist);
      /*
         GetDimensions assumes Form, and uses horizDistance; Group doesn't
         use horizDistance - its spacing is dynamic.  To approximate
         (correctly, for nodeframe), use spacing.  However, in the
         general case, this method must be changed.
      */
      if (XtIsSubclass( formw, groupWidgetClass ))  {
          XtVaGetValues( formw, XtNspacing, &spacing, NULL );
          node_hdist = spacing;
      }

#ifdef DEBUG
      fprintf(stderr, "\tform: width = [%4d], y = [%4d]\n",
	      form_width, form_y);
      fprintf(stderr, "\tnode: width = [%4d], x = [%4d], hdist = [%4d]\n",
	      node_width, node_x, node_hdist);
#endif

      /*
       * If the length is the same as the form width,
       * Reset the scrollbar to show everything
       * Else get new values
       */
      if (length == form_width)
      {
	 new_top = (float)    0.0;
	 new_x   = (Position) 0;
	 shown   = (float)    1.0;
      }
      else
      {
	 /*
	  * If it is the definitions scrollbar 
	  * Add extra buffer space to include the occurrence widget
	  */
	 if (altflag == True)
	 {
	    buffer = node_hdist + 40; /* enough space for occurw & */
				      /* typew */
	 }
	 else
	 {
	    buffer = 2 * DEFAULTSPACE;
	 }

	 /*
	  * Calculate the coverage
	  * If the coverage is more than the length,
	  * Calculate the new form widget X as (length - coverage)
	  * This is usually negative to move the data left of
	  * the viewport.
	  */
	 coverage = node_x + node_width + buffer;
	 if (coverage > length)
	 {
	    new_x = length - coverage;
	 }
	 else
	 {
	    new_x = 0;
	 }

	 /*
	  * The new top of the scrollbar thumb is the percentage
	  * of the distance that the form widget is being moved
	  * divided by the total form widget width
	  * Because shown is correct (scrollbar does that for us),
	  * make sure that new_top won't move shown too much
	  */
	 new_top = (float) (0 - new_x) / form_width;;
	 if ((new_top + shown) > 1.0)
	 {
	    new_top    = (float) 1.0 - shown;
	 }
      }
#ifdef DEBUG
      fprintf(stderr, "\tnew_top = [%7.4f], new_x = [%d], shown = [%7.4f]\n",
	      new_top, new_x, shown);
#endif

      /*
       * Set the new values of the scrollbar
       * Move the form widget
       */
      XawScrollbarSetThumb(horizw, new_top, shown);
      XtMoveWidget(formw, new_x, form_y);
   }
}




/*
   FUNCTION: 		AdjustVertical

   Synopsis: 
       Adjust the vertical scrollbar of the tree display so that
       the widget that is the focus of the interface is one widget 
       above the bottom edge of the viewport.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
AdjustVertical(formw, vertw, nodew)

Widget
   formw, /* The form widget containing the tree display */
   vertw, /* The viewport containing the form widget */
   nodew; /* A node under consideration by the user in the form widget */

{
   Arg
      args[MAXARGS];/* List of arguments */
   Dimension
      coverage,    /* Amount of data from the top of the form to 
		      the bottom of the node to be shown */
      length,      /* Length of the scrollbar */
      form_height, /* Height of the form widget */
      node_height, /* Height of the node widget */
      width;       /* Used for GetDimensions() */
   float
      new_top,     /* New top of thumb location */
      shown,       /* Percentage of the form being viewed */
      top;         /* Location where the view begins */
   int
      cnt,         /* Number of arguments */
      hdist,       /* Used for GetDimensions() */
      node_vdist,  /* Vertical distance to the node */
      vdist;       /* Used for GetDimensions() */
   Position
      x,           /* Used for GetDimensions() */
      form_x,      /* X coordinate of the form widget */
      y,           /* Used for GetDimensions() */
      new_y,       /* New Y coordinate for the form widget */
      node_y;      /* Y coordinate of the node */

   /*
    * Get values of the vertical scrollbar
    */
   cnt = 0;
   XtSetArg(args[cnt], XtNlength,     &length); cnt++;
   XtSetArg(args[cnt], XtNshown,      &shown);  cnt++;
   XtSetArg(args[cnt], XtNtopOfThumb, &top);    cnt++;
   XtGetValues(vertw, args, cnt);
#ifdef DEBUG
   fprintf(stderr, "Vertical scrollbar:\n");
   fprintf(stderr, "\tlength = [%d], shown = [%7.4f], top = [%7.4f]\n",
	   length, shown, top);
#endif

   /*
    * If the amount shown is less than all of the data, 
    * Recalculate scrollbar and form widget attributes
    */
   if (shown < 1.0)
   {
      /*
       * Get the dimensions of the form and node widgets
       */
      GetDimensions(formw,  True, &form_x, &y, &width, &form_height,
		    &hdist, &vdist);
      GetDimensions(nodew,  True, &x, &node_y, &width, &node_height,
		    &hdist, &node_vdist);

#ifdef DEBUG
      fprintf(stderr, "\tform: height = [%4d], x = [%4d]\n",
	      form_height, form_x);
      fprintf(stderr, "\tnode: height = [%4d], y = [%4d], vdist = [%d]\n",
	      node_height, node_y, node_vdist);
#endif

      /*
       * If the length is the same as the form height,
       * Reset the scrollbar to show everything
       * Else get new values
       */
      if (length == form_height)
      {
	 new_top = (float)    0.0;
	 new_y   = (Position) 0;
	 shown   = (float)    1.0;
      }
      else
      {
	 /*
	  * Calculate the coverage
	  * If the coverage is more than the length,
	  * Calculate the new form widget Y as (length - coverage)
	  * This is usually negative to move the data up of
	  * the viewport.
	  */
	 coverage = node_y + (2 * node_height) + node_vdist;
	 if (coverage > length)
	 {
	    new_y = length - coverage;
	 }
	 else
	 {
	    new_y = 0;
	 }

	 /*
	  * The new top of the scrollbar thumb is the percentage
	  * of the distance that the form widget is being moved
	  * divided by the total form widget height
	  * Because shown is correct (scrollbar does that for us),
	  * make sure that new_top won't move shown too much
	  */
	 new_top = (float) (0 - new_y) / form_height;;
	 if ((new_top + shown) > 1.0)
	 {
	    new_top = (float) 1.0 - shown;
	 }

      }
#ifdef DEBUG
      fprintf(stderr, "\tnew_top = [%7.4f], new_y = [%d], shown = [%7.4f]\n",
	      new_top, new_y, shown);
#endif

      /*
       * Set the new values of the scrollbar
       * Move the form widget
       */
      XawScrollbarSetThumb(vertw, new_top, shown);
      XtMoveWidget(formw, form_x, new_y);
   }
   return;
}




/*
   FUNCTION: 		AdjustScrollBars

   Synopsis: 
       Adjust all of the scrollbars for the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON

   Globals used:
       none

   Globals Modified:
       none

   Return value:	static void

   Range:		none
*/

static void
AdjustScrollBars(nptr)

Node
   *nptr;      /* The node under consideration by the user */

{
   Widget
      formw,   /* The form widget to pass to the 
		  individual adjust functions */
      horizw,  /* Horizontal scrollbar widget */
      vertw,   /* Vertical scrollbar widget */
      vieww;   /* Viewport widget */

   if (nptr != NULL)
   {
      /*
       * Adjust the altframe horizontal scrollbar
       * Adjust the tree display horizontal and vertical scrollbars
       */
      if (nptr->nodew != NULL)
      {
	 formw  = XtParent(nptr->nodew);
	 vieww  = XtParent(formw);
	 horizw = XtNameToWidget(vieww, "horizontal");
	 AdjustHorizontal(formw, horizw, nptr->nodew, True);
      }
#ifndef R5
      if (nptr->treew != NULL)
      {
	 formw  = XtParent(nptr->treew);
	 vieww  = XtParent(formw);
	 horizw = XtNameToWidget(vieww, "horizontal");
	 vertw  = XtNameToWidget(vieww, "vertical");
	 AdjustHorizontal(formw, horizw, nptr->treew, False);
	 AdjustVertical(formw, vertw, nptr->treew);
      }
#endif
   }
   return;
}




/*
   FUNCTION: 		UpdateInterface

   Synopsis: 
       Update the interface based on a movement or change by the user.

   Initial purpose:
       Same as above.

   Present purpose:
       Same as above.

   Creator:	Conleth O'Connell
   Reviewer:	

   Modifications:
       DATE       WHO    REASON
       91/02/28   cso    Added the focus setting and a check for a
                          NULL widget.
       91/05/23   cso    Removed Chosen_Symbol.

   Globals used:
       none

   Globals Modified:
       Chosennode   Points to the new pointer

   Return value:	void

   Range:		none
*/

void
UpdateInterface(oldptr, newptr)

Node
   *oldptr, /* The old position in the data structures */
   *newptr; /* The new position in the data structures */

{
   Dimension width, height, bwidth;   
   XtWidgetGeometry pref;
   Widget nodeview, nodeframe;

   nodeview  = XtNameToWidget( Frame, "*nodeview" );
   nodeframe = XtNameToWidget( Frame, "*nodeframe" );
   
   GroupDoLayout( nodeframe, False );  XtUnmanageChild(nodeframe);

   /*
    * Hide the definition containing the the old pointer
    * Update the tree display
    * Update the compass
    * Update auxiliary information
    * Display the definition containing the new pointer
    * Adjust the scrollbars
    * If the new pointer has a node widget
    * Highlight the symbol in the node widget
    * Set the keyboard focus to that widget for typing
    * If the new pointer is nonNULL, set Chosennode
    */
   HideSubtree(oldptr);
   UpdateTree(oldptr, newptr);

   UpdateCompass(newptr);

#ifdef AUX
   UpdateAux(newptr);
#endif
   DisplaySubtree(newptr);

   if (newptr->nodew != NULL)
   {
       XtVaSetValues( newptr->nodew,
           XtNdisplayCaret, True,
           XtNborderWidth,  2,
           NULL );
       XtSetKeyboardFocus(Frame, newptr->nodew);
   }

   /*
       Query nodeframe's preferred size, then set width to max of the
       viewport's width and nodeframe's preferred width.
   */

   XtQueryGeometry( nodeframe, NULL, &pref );
   XtVaGetValues( nodeview,
       XtNwidth,  &width,
       XtNheight, &height,
       NULL );
   XtVaSetValues( nodeframe,
       XtNwidth,  Max( width,  pref.width ),
       NULL );

   GroupDoLayout( nodeframe, True );  XtManageChild(nodeframe);

   AdjustScrollBars(newptr);
   if (*newptr->symbol != '\0')
   {
       Chosennode = newptr;
   }
   else
   {
       Chosennode = NULL;
   }
   return;
}
