#include   "flexdef.h"
#include   "parse.h"
/******************************************************************
**
**   zzlex()
**
**  The zzlex() routine returns the token which represents the next
**  character in the regular expression.  A single character may be 
**  represented by several different tokens depending on its context.
**  For example, the character '^' returns a different token if it
**  is the first character in the expression than if it were any other 
**  character in the expression.  The token associated with a character 
**  is often determined by the mode that this routine is in.
**
**  Modes used:
**    NORMAL_MODE:  This is the mode in which this routine starts.
**    QUOTE_MODE:   A quotation mark has been encountered.  Remain
**                    in this mode until a closing quote is found.
**    NUM_MODE:     A '{' has been encountered, expect a number to
**                    follow.  Remain in this mode until the '}' is
**                    encountered.
**    CCL_MODE:     A '[' was encountered.  Remain in this mode until
**                    the ']' is encountered.
**
**   Variables:
**      x:      Index into the regular expression.
**      state:  Used to mark the beginning and ending of the 
**                regular expression.
**      mode:   Used to indicate how to interpret characters which
**                have more that one token associated with them.
**      y:      Loop counter.
**      token   The token value to be returned.
**      cclval  Value associated with a character class.
**  
******************************************************************/
int   zzlex()
{
static int   x = 0, state = 1;
static int   mode = NORMAL_MODE;
int         y, token, cclval;

/*
**  If we're in state 0, we've already processed on regular expression
**    and we're ready to start on another.  Reset the index and the
**    mode.  Go to state 1.
*/

   if (state == 0)
     {
      x = 0;
      mode = NORMAL_MODE;
      state = 1;
     }

/*
**  If we're in state 2, we hit the end of the string on the last
**    call, so return the EOF token and go to state 0 to get ready
**    to process the next regular expression.
*/

   else if (state == 2)
     {
      state = 0;         /*  So we'll reset the variables next time */
      return(EOF);
     }
   

/*
**  If we're in the NORMAL_MODE and the '/' is encountered, it indicates
**    trailing context.  We should ignore trailing context, so skip
**    the remainder of the regular expression.
*/

   if ((mode == NORMAL_MODE) && (rex[x] == '/'))
     {
#ifdef DEBUG
      fprintf(stderr, "Trailing context encountered.  It will be ignored.\n");
#endif
      while (rex[x] != '\0')
        {
         x++;
        }
     }


/*
**  Only the end-of-string marker, closing quote, and backslash have
**    special meaning when we are in the QUOTE_MODE.  If we are in
**    the QUOTE_MODE and encounter a different character, treat it
**    as a CHAR.
*/
   if ((mode == QUOTE_MODE) && 
       (rex[x] != '\0') && (rex[x] != '"') && (rex[x] != '\\'))
     {
      token = CHAR;
      zzlval = (int) rex[x];
     }

/*
**  Operands may have special meaning.  Treat each operand separately.
*/
   else
   switch (rex[x])
     {
/*
**  The '^' only has special meaning if it is the first character in
**    the expression or if it is the first character in a character
**    class definition.  If the '^' is the first character in the
**    regular expression, treat it as if it were a '\n' and set
**    the bol_marker to true.  If the '^' is the first character
**    in a character class definition, return the token '^' to
**    the parser.  If neither of these cases is true, treat the
**    '^' as a normal character.
*/
      case '^':
         if (x==0)
           {
            bol_needed = true;
            token = CHAR;
            zzlval = '\n';
           }
         else if (rex[x-1] == '[')
           {
            token = (int) '^';
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '"' only has special meaning in the NORMAL_MODE and the 
**    QUOTE_MODE.  It toggles between these two modes.  If neither
**    mode is currently active, it should be treated as a normal
**    character.
*/

      case '"':
         if ( mode == NORMAL_MODE)
           {
            token = (int) '"';
            mode = QUOTE_MODE;
           }
         else if (mode == QUOTE_MODE)
           {
            token = (int) '"';
            mode = NORMAL_MODE;
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '{' only has special meaning when we are in the
**    NORMAL_MODE.  If we are not in the NORMAL_MODE, or
**    if we are in the NORMAL_MODE but the '{' is the first
**    character in the regular expression, then the '{' should
**    be treated as a normal character.
*/

      case '{':
         if (( mode == NORMAL_MODE) && (x != 0))
           {
            token = (int) '{';
            mode = NUM_MODE;
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '$' only has special meaning if it is the last character in
**    the regular expression.  If it is not the last character in the
**    regular expression, it should be treated as a normal character.
*/

      case '$':
         if (rex[x+1] == '\0')
           {
            eol_needed = true;
            token = CHAR;
            zzlval = (int) '\n';
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '\0' indicates the end of the regular expression.  Return a 
**    '\n' because if '\0' appeared in parse.y, yacc would think it's
**    an invalid escape sequence.  Also go into state 2 to show that
**    the entire regular expression has been processed.
*/

      case '\0':
         state = 2;
         token = (int) '\n';
         break;

/*
**  The '[' should be treated as a special character unless it appears
**    as part of the substring '[]' or '[\]'.  If this is the case, 
**    treat the '[' as a normal character.  If this is not the case, 
**    initialize the character class that it represents.
*/

      case '[':

/*
**  Treat it as a normal character.
*/
         if ( (rex[x+1] == ']') || 
              ((rex[x+1] == '\\') && (rex[x+2] == ']')) )
           {
            token = CHAR;
            zzlval = (int) rex[x];
            break;
           }

/*
**  It defines a character class.  Copy the character class into
**    nmstr, see if the character class has been encountered before,
**    if it has, return the number associated with it, if not
**    begin to define the new character class.
*/

         /*  Copy the class into nmstr[]  */

         y = x;

reloop:
         while ((rex[y] != ']') && (rex[y] != '\0'))
           {
            nmstr[y-x] = rex[y];
            y++;
           }

         if (rex[y-1] == '\\')
           {
            y++;
            goto reloop;
           }

         nmstr[y-x] = ']';
         nmstr[y-x+1] = '\0';
         

         /* check to see if we've already encountered this ccl */
         /* Yes, we want to do the assignment! */

         if ( (cclval = ccllookup( (Char *) nmstr )) )
             {
             zzlval = cclval;
             ++cclreuse;
             token = PREVCCL;
             x = y;
             }
         else
             {
             /* we fudge a bit.  We know that this ccl will
              * soon be numbered as lastccl + 1 by cclinit
              */
             cclinstal( (Char *) nmstr, lastccl + 1 );

             token = (int) '[';
             mode = CCL_MODE;
             }
         break;


/*
**  The ',' only has special meaning if we are in the NUM_MODE.  If we
**    are in NUM_MODE, then we've got something like "a{3, 10}".  If
**    we're not in NUM_MODE, treat the ',' as a normal character.
*/

      case ',':
         if (mode == NUM_MODE)
           {
            token = (int) ',';
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '-' only has special meaning the the CCL_MODE, and further,
**    if we are in the CCL_MODE, it only has special meaning if it
**    trails a '[' or precedes a ']'.  If this is not the case, 
**    treat the '-' as a normal character.
*/

       case '-':
         if (mode == CCL_MODE)
           {
            if ((rex[x-1] != '[') && (rex[x+1] != ']'))
              {
               token = (int) '-';
              }
            else
              {
               token = CHAR;
               zzlval = (int) rex[x];
              }
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The ']' only has special meaning if we are in the CCL_MODE.  If 
**    this is not the case, treat the ']' as a normal character.
*/

      case ']':
         if (mode == CCL_MODE)
           {
            token = (int) ']';
            mode = NORMAL_MODE;
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '}' only has special meaning if we are in the NUM_MODE.  If 
**    this is not the case, treat the '}' as a normal character.
*/

      case '}':
         if (mode == NUM_MODE)
           {
            token = (int) '}';
            mode = NORMAL_MODE;
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  The '\' character always indicates an escape sequence.  A letter
**    may follow to give it special meaning.  An octal value may
**    follow to indicate an octal character.  A hex value may
**    follow to indicate a hax character.
*/

      case '\\':
         switch (rex[x+1])
           {
           /* case 'a': zzlval = (int) '\a';
                      break;*/
            case 'b': zzlval = (int) '\b';
                      break;
            case 'f': zzlval = (int) '\f';
                      break;
            case 'n': zzlval = (int) '\n';
                      break;
            case 'r': zzlval = (int) '\r';
                      break;
            case 't': zzlval = (int) '\t';
                      break;
            case 'v': zzlval = (int) '\v';
                      break;

            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
                      x++;
                      zzlval = 0;
                      while ((isdigit(rex[x])) && (rex[x] != '8')
                                               && (rex[x] != '9'))
                        {
                         zzlval = zzlval * 8 + (rex[x]) - (int) ('0');
                         x++;
                        }

                      if (zzlval >= DEFAULT_CSIZE)
                        {
                         synerr("Illegal octal value specified");
                        }

                      x--;
                      break;

            case 'x':
                      x++;
                      x++;
                      zzlval = 0;
                      while (    (isdigit(rex[x]))
                              || (toupper(rex[x]) == 'A')
                              || (toupper(rex[x]) == 'B')
                              || (toupper(rex[x]) == 'C')
                              || (toupper(rex[x]) == 'D')
                              || (toupper(rex[x]) == 'E')
                              || (toupper(rex[x]) == 'F') )
                                     
                        {
                         if (isdigit(rex[x]))
                           {
                            zzlval = zzlval * 16 + (rex[x]) - (int) ('0');
                           }
                         else
                           {
                            zzlval = zzlval * 16 + toupper(rex[x]) 
                                                   - (int) ('A') + 10;
                           }
                            x++;
                        }

                      if (zzlval >= DEFAULT_CSIZE)
                        {
                         synerr("Illegal octal value specified");
                        }

                      x--;
                      break;

            default:  zzlval = (int) rex[x+1];
                      break;
           }
         x = x + 1;      
         token = CHAR;
         break;

      case '|': case '*': case '+': case '?': case '.': case '(': case ')':
         if (mode == NORMAL_MODE)
           {
            token = (int) rex[x];
           }
         else
           {
            token = CHAR;
            zzlval = (int) rex[x];
           }
         break;

/*
**  Digits have special meaning only when in the NUM_MODE.  If this
**    is not the case, treat them as a normal character.
*/

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
         if (mode == NUM_MODE)
           {
            zzlval = 0;
            while (isdigit(rex[x]))
              {
               zzlval = zzlval * 10 + (int) rex[x] - (int) '0';
               x++;
              }
            x--;
            token = NUMBER;
           }
         else
           {
            zzlval = (int) rex[x];
            token = CHAR;
           }
         break;

/*
**  All other characters should be always be treated as normal characters.
*/

      default:
         token = CHAR;
         zzlval = (int) rex[x];
         
     }
   
   x++;
   return(token);   

}
